/*******************************************************************************
 **
 ** $Id: peacedata.C 173 2008-12-12 20:40:59Z AndrewLahiff $
 **
 ******************************************************************************
 **
 ** Copyright (C) 2007-2008 Andrew Lahiff.
 **
 ** This file is part of PEACEmoments CAA.
 **
 ** PEACEmoments CAA is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** PEACEmoments CAA is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with PEACEmoments CAA.  If not, see <http://www.gnu.org/licenses/>.
 **
 *******************************************************************************/

#include "peacemoments_caa.h"
#include <math.h>

const double mass      = 9.1095e-31;
const double evjoule   = 1.602177e-19;
const double kBC       = 1.3807e-23;
const double TRDEG2RAD = 0.0174532925199432958;
const double pi        = 4.0*atan(1.0);

//=============================================================================
//
// PeaceData class
//
//=============================================================================

//-----------------------------------------------------------------------------
//
// constructor
//
//-----------------------------------------------------------------------------

PeaceData::PeaceData(bool in_usingLEEA, bool in_usingHEEA, int numEnActLEEA, int numEnActHEEA,
                     int inA_LEEA, int inP_LEEA, int inE_LEEA, int inNoEf_LEEA,
		     int inA_levelLow_LEEA, int inA_levelHigh_LEEA,
                     QString inDATASET_ID_LEEA,
	             int inA_HEEA, int inP_HEEA, int inE_HEEA, int inNoEf_HEEA,
		     int inA_levelLow_HEEA, int inA_levelHigh_HEEA,
                     QString inDATASET_ID_HEEA)
{

// which sensors are being used

   usingLEEA = in_usingLEEA;
   usingHEEA = in_usingHEEA;
   
// sweep modes

   sweepModeLEEA = -1;
   sweepModeHEEA = -1;   

// energy levels
 
   energyLevelLowLEEA = inA_levelLow_LEEA;
   energyLevelHighLEEA = inA_levelHigh_LEEA;
   energyLevelLowHEEA = inA_levelLow_HEEA;
   energyLevelHighHEEA = inA_levelHigh_HEEA;

// data origin, distribution type, energy summation

   dataOriginLEEA = -1;
   dataOriginHEEA = -1;
   distTypeLEEA = -1;
   distTypeHEEA = -1;
   iEnergySummationLEEA = 0;
   iEnergySummationHEEA = 0;

// DATASET_ID for each sensor (contains product name)

   dataSetIdLEEA = inDATASET_ID_LEEA;
   dataSetIdHEEA = inDATASET_ID_HEEA;

// for PITCH_FULL* / PITCH_3D* only

   if (dataSetIdLEEA.contains("PITCH_FULL") || dataSetIdHEEA.contains("PITCH_FULL"))
   {
      numPaForMoms = 13;
   }
   else numPaForMoms = 12;

// setup dimensions of the LEEA & HEEA data

   dimLEEA.numA = inA_LEEA;     				// LEEA - number of azimuth bins
   dimLEEA.numP = inP_LEEA;     				// LEEA - number of polar bins
   dimLEEA.numE = inE_LEEA;     				// LEEA - number of energy bins
   dimLEEA.numEfly = inE_LEEA - inNoEf_LEEA; 	                // LEEA - number of flybacks
                                                                //        (we'll set this correctly later)
   dimHEEA.numA = inA_HEEA;     				// HEEA - number of azimuth bins
   dimHEEA.numP = inP_HEEA;     				// HEEA - number of polar bins
   dimHEEA.numE = inE_HEEA;     				// HEEA - number of energy bins
   dimHEEA.numEfly = inE_HEEA - inNoEf_HEEA; 	                // HEEA - number of flybacks
                                                                //        (we'll set this correctly later)

// IMPORTANT NOTE: "number of flybacks" is not actually correct. For 3DX energy windows, PITCH_FULL etc this will
// be completely wrong.

   //cout << "[debug] bins = " << dimLEEA.numA  << " : " << dimLEEA.numP << " : " << dimLEEA.numE << " : " << dimLEEA.numEfly << " :: " << dimHEEA.numA << " : " << dimHEEA.numP << " : " << dimHEEA.numE << " : " << dimHEEA.numEfly << " act = " << numEnActLEEA << " : " << numEnActHEEA << endl;

// setup data storage arrays for LEEA and HEEA

   int arraySizeData = 64*12*64; // maximum possible size
   if (usingLEEA) dataStoreLEEA = new float[arraySizeData];
   if (usingHEEA) dataStoreHEEA = new float[arraySizeData];

// setup energy bin & potential storage arrays for LEEA and HEEA

   actualArraySizeLEEA = numEnActLEEA;
   actualArraySizeHEEA = numEnActHEEA;

   if (usingLEEA)
   {
      energyLEEA = new float[numEnActLEEA];
      energyLowLEEA = new float[numEnActLEEA];
      energyHighLEEA = new float[numEnActLEEA];
      energyUsedLEEA = new int[numEnActLEEA];
      potentialLEEA = new float[dimLEEA.numA];
   }
   if (usingHEEA)
   {
      energyHEEA = new float[numEnActHEEA];
      energyLowHEEA = new float[numEnActHEEA];
      energyHighHEEA = new float[numEnActHEEA];
      energyUsedHEEA = new int[numEnActHEEA];
      potentialHEEA = new float[dimHEEA.numA];
   }
   
// user-defined restricted integration limits

   energyRestrictLow = 0;
   energyRestrictHigh = 0;
   polarLimitLow = 0;
   polarLimitHigh = 0;
   azimuthLimitLow = 0;
   azimuthLimitHigh = 0;
   
// sensors required for different energy regions

   useT  = false;
   useB  = false;
   useLO = false;
   useHO = false; 
   
   senT  = _HEEA;
   senB  = _LEEA;
   senLO = _LEEA;
   senHO = _HEEA;  

// intialize potential - we store one value per azimuth

   if (usingLEEA)
   {
      for (int i=0; i<dimLEEA.numA; i++)
      {
         *(potentialLEEA+i) = 0.0;
      }
   }
   if (usingHEEA)
   {
      for (int i=0; i<dimHEEA.numA; i++)
      {
         *(potentialHEEA+i) = 0.0;
      }
   }

// current coordinate system of relative moments: 0 = sensor, 1 = S/C, 2 = GSE, 3 = GSM
// for 3D products, integration is done in the PEACE sensor coordinate system

   currentCoordinateSystem = _SENSOR;

// intialize GSE spin-axis angles, SPOS & FGM B-field

   spinAxisLatitudeGSE = FILL_VALUE_PM;
   spinAxisLongitudeGSE = FILL_VALUE_PM;
   spos = FILL_VALUE_PM;
   
   for (int i=0; i<3; i++)
   {
      fgmB[i] = FILL_VALUE_PM;
   }

   magFieldIsFill = false;

// initialize other

   potentialOffset1 = 0.0;
   potentialOffset2 = 0.0;
   potentialOffset2units = 1; // bins

// initialize moments

   setAllMomentsToFills();

// sensors used

   sensorUsedTopRegion = -1;
   sensorUsedOverlapRegion = -1;
   sensorUsedBottomRegion = -1;

}

//-----------------------------------------------------------------------------
//
// destructor
//
//-----------------------------------------------------------------------------

PeaceData::~PeaceData()
{

// delete PEACE data storage arrays

   if (usingLEEA) delete [] dataStoreLEEA;
   if (usingHEEA) delete [] dataStoreHEEA;

// delete energy & potential storage arrays

   if (usingLEEA)
   {
      delete [] energyLEEA;
      delete [] energyLowLEEA;
      delete [] energyHighLEEA;
      delete [] energyUsedLEEA;
      delete [] potentialLEEA;
   }
   if (usingHEEA)
   {
      delete [] energyHEEA;
      delete [] energyLowHEEA;
      delete [] energyHighHEEA;
      delete [] energyUsedHEEA;
      delete [] potentialHEEA;
   }
}

//-----------------------------------------------------------------------------
//
// setup energy integration limits for the different regions
//
//-----------------------------------------------------------------------------

int PeaceData::setEnergyRegionLimits(int sensorCombination)
{

// firstly we'll set the number of "flybacks" - note that these are not really
// flybacks, but are the number of leading fills

   if (usingLEEA)
   {
      int icount = 0;
      for (int i=0; i<10; i++)
      {
         if (energyLEEA[i] == -1) icount++;
	 else break;
      }
      dimLEEA.numEfly = icount;
   }

   if (usingHEEA)
   {
      int icount = 0;
      for (int i=0; i<10; i++)
      {
         if (energyHEEA[i] == -1) icount++;
	 else break;
      }
      dimHEEA.numEfly = icount;
   }

// sensor combination

   if (sensorCombination == _LEEA)
   {
      useB = true;
      usingHEEA = false;
   }
   else if (sensorCombination == _HEEA)
   {
      useT = true;
      usingLEEA = false;
   }
   else if (sensorCombination == _TH1H2B)
   {
      useT = true;
      useB = true;
      useHO = true;  
   }
   else if (sensorCombination == _TL1L2B)
   {
      useT = true;
      useB = true;
      useLO = true;
   }
   else if (sensorCombination == _TL1L2)
   {
      useT = true;
      useLO = true;
   }
   else if (sensorCombination == _TH1H2)
   {
      useT = true;
      useHO = true;
   }
   else if (sensorCombination == _L1L2B)
   {
      useLO = true;
      useB = true;
   }
   else if (sensorCombination == _H1H2B)
   {
      useHO = true;
      useB = true;
   }
   else if (sensorCombination == _OVERLAP_TWO_SEC)
   {
      useT = true;
      useB = true;  
   }

//cout << "using lEEEA, HEEA = " << usingLEEA << " : " << usingHEEA << endl;

// distribution setup for LEEA

   if (usingLEEA)
   {

// set distribution

      if (dataOriginLEEA == 0) distTypeLEEA = _3DF;
      else if (dataOriginLEEA == 3) distTypeLEEA = _3DR;
      else if (dataOriginLEEA == 1 || dataOriginLEEA == 2) distTypeLEEA = _3DXN;
      else if (dataOriginLEEA == 11 || dataOriginLEEA == 12) distTypeLEEA = _3DXE;
      else if (dataOriginLEEA == 21 || dataOriginLEEA == 22) distTypeLEEA = _3DXP;
      else distTypeLEEA = _PAD;

      if (dataSetIdLEEA.contains("3DR")) distTypeLEEA = _3DR;

// set energy summation information

      if (distTypeLEEA == _3DR)
      {
         if (sweepModeLEEA == 2) iEnergySummationLEEA = 0;
         else if (sweepModeLEEA == 1) iEnergySummationLEEA = 4;
         else iEnergySummationLEEA = 2;
      }
      else if (distTypeLEEA == _3DXE)
      {
         iEnergySummationLEEA = 2;
      }
      else
      {
         iEnergySummationLEEA = 0;
      }

// setup used energy bins array

      if (distTypeLEEA == _PAD)
      {
         switch (sweepModeLEEA)
         {
            case 1:
               dimLEEA.numEfly = 4;
               break;
            case 2:
               dimLEEA.numEfly = 1;
               break;
            case 3:
               dimLEEA.numEfly = 2;
               break;
         }
      }

//cout << "LEEA used bins = " << dimLEEA.numEfly << " to " << dimLEEA.numE << endl;

      int eSkip, eNum, eVal;
      energyLevels(sweepModeLEEA, energyLevelHighLEEA, &eSkip, &eNum, &eVal, distTypeLEEA);
      for (int i=dimLEEA.numEfly; i<dimLEEA.numE; i++)
      {
         energyUsedLEEA[i] = eVal;
	 eVal = eVal - eSkip;
  	 //cout << "LEEA used energy bins: [" << i << "] " << energyUsedLEEA[i] << endl;
      }

   }

// distribution setup for HEEA

   if (usingHEEA)
   {

// set distribution

      if (dataOriginHEEA == 0) distTypeHEEA = _3DF;
      else if (dataOriginHEEA == 3) distTypeHEEA = _3DR;
      else if (dataOriginHEEA == 1 || dataOriginHEEA == 2) distTypeHEEA = _3DXN;
      else if (dataOriginHEEA == 11 || dataOriginHEEA == 12) distTypeHEEA = _3DXE;
      else if (dataOriginHEEA == 21 || dataOriginHEEA == 22) distTypeHEEA = _3DXP;
      else distTypeHEEA = _PAD;

      if (dataSetIdHEEA.contains("3DR")) distTypeHEEA = _3DR;

// set energy summation information

      if (distTypeHEEA == _3DR)
      {
         if (sweepModeHEEA == 2) iEnergySummationHEEA = 0;
         else if (sweepModeHEEA == 1) iEnergySummationHEEA = 4;
         else iEnergySummationHEEA = 2;
      }
      else if (distTypeHEEA == _3DXE)
      {
         iEnergySummationHEEA = 2;
      }
      else
      {
         iEnergySummationHEEA = 0;
      }

// setup used energy bins array

      if (distTypeHEEA == _PAD)
      {
         switch (sweepModeHEEA)
         {
            case 1:
               dimHEEA.numEfly = 4;
               break;
            case 2:
               dimHEEA.numEfly = 1;
               break;
            case 3:
               dimHEEA.numEfly = 2;
               break;
         }
      }

//cout << "HEEA used bins = " << dimHEEA.numEfly << " to " << dimHEEA.numE << endl;

      int eSkip, eNum, eVal;
      energyLevels(sweepModeHEEA, energyLevelHighHEEA, &eSkip, &eNum, &eVal, distTypeHEEA);
      for (int i=dimHEEA.numEfly; i<dimHEEA.numE; i++)
      {
         energyUsedHEEA[i] = eVal;
         eVal = eVal - eSkip;
         //cout << "HEEA used energy bins: [" << i << "] " << energyUsedHEEA[i] << endl;
      }
   }

// return value initialization

   int returnValue = -1;

// initialization - set limits to default values

   eBinLimits[0].low = dimHEEA.numEfly;
   eBinLimits[0].high = dimHEEA.numE;
   eBinLimits[1].low = dimLEEA.numEfly;
   eBinLimits[1].high = dimLEEA.numE;
   eBinLimits[2].low = dimLEEA.numEfly;
   eBinLimits[2].high = dimLEEA.numE;
   eBinLimits[3].low = dimHEEA.numEfly;
   eBinLimits[3].high = dimHEEA.numE;

// some checking - make sure we have energy lower/upper level information when
// necessary

   if (usingLEEA && (energyLevelLowLEEA == -1 || energyLevelHighLEEA == -1))
      usingLEEA = false;

   if (usingHEEA && (energyLevelLowHEEA == -1 || energyLevelHighHEEA == -1))
      usingHEEA = false;

// determine energy limits based on (i) sensors being used, (ii) user's sensor
// combination selection, (iii) lower/upper energy levels for each sensor. Note
// that we need to be careful when one of the two sensors has energy summation
// and the other doesn't.
//
// For PITCH_FULL overlap 2-sec snapshot moments, the data has already been
// ordered into top, overlap and bottom regions, so we don't need to do 
// anything.

   if (sensorCombination == _OVERLAP_TWO_SEC)
   {

// don't do anything - just use default values set above (!)

   }
   else if (usingLEEA && !usingHEEA)
   {
      useB  = true;
      useLO = false;
      useHO = false;
      useT  = false;

      senB  = _LEEA;

      returnValue = 4;
   }
   else if (usingHEEA && !usingLEEA)
   {
      useB  = false;
      useLO = false;
      useHO = false;
      useT  = true;

      senT  = _HEEA;

      returnValue = 4;
   }
   else if (!usingLEEA && !usingHEEA)
   {
      useB  = false;
      useLO = false;
      useHO = false;
      useT  = false;

      returnValue = -1;
   }
   else if (usingLEEA && usingHEEA)
   {
      returnValue = 1;

// select what sensors are used for the T, B, LO and HO energy regions

      int sensorOverlapType = -1;

      if (energyLevelLowLEEA < energyLevelLowHEEA &&
          energyLevelHighLEEA < energyLevelHighHEEA) // LEEA & HEEA partially overlap, LEEA lower energies
      {
         senT  = _HEEA;
         senLO = _LEEA;
         senHO = _HEEA;
         senB  = _LEEA;
	 sensorOverlapType = 0;
      }
      else if (energyLevelLowHEEA < energyLevelLowLEEA &&
               energyLevelHighLEEA < energyLevelHighHEEA) // HEEA fully overlaps LEEA
      {
         senT  = _HEEA;
	 senLO = _LEEA;
	 senHO = _HEEA;
	 senB  = _HEEA;
	 sensorOverlapType = 1;
      }
      else if (energyLevelLowLEEA < energyLevelLowHEEA &
               energyLevelHighHEEA < energyLevelHighLEEA) // LEEA fully overlaps HEEA
      {
         senT  = _LEEA;
	 senLO = _LEEA;
	 senHO = _HEEA;
	 senB  = _LEEA;
	 sensorOverlapType = 2;
      }
      else if (energyLevelLowHEEA < energyLevelLowLEEA &&
               energyLevelHighHEEA < energyLevelHighLEEA) // LEEA & HEEA partially overlap, HEEA lower energies
      {
         senT  = _HEEA;
	 senLO = _LEEA;
	 senHO = _HEEA;
	 senB  = _HEEA;
	 sensorOverlapType = 3;
      }
      else if (energyLevelLowLEEA == energyLevelLowHEEA &&
               energyLevelHighLEEA == energyLevelHighHEEA) // LEEA & HEEA cover same energies
      {
         senT  = _HEEA;
	 senLO = _LEEA;
	 senHO = _HEEA;
	 senB  = _LEEA;
	 sensorOverlapType = 4;
      }

      //cout << "sensor overlap type = " << sensorOverlapType << " : " << senT << " : " << senLO << " : " << senHO << " : " << senB << " energy sum = " << iEnergySummationLEEA << " : " << iEnergySummationHEEA << endl;

// determine integration limits for each energy region

         int eOff1 = 0, eOff2 = 0;
         if (iEnergySummationLEEA == 2 && iEnergySummationHEEA == 0)
         {
            eOff2 = 2; 
         }
         else if (iEnergySummationLEEA == 0 && iEnergySummationHEEA == 2)
         {
            eOff2 = -2; 
         }
	// cout << "eof 1/2 = " << eOff1 << " , " << eOff2 << "  sum L/H = " << iEnergySummationLEEA << " " << iEnergySummationHEEA << " nums L/H = " << dimLEEA.numE << " , " << dimHEEA.numE << endl;

	 int EL[4] = {0};
         bool iFound = false;
	 for (int j = dimLEEA.numEfly; j < dimLEEA.numE; j++)
         {
	    for (int i = dimHEEA.numEfly; i < dimHEEA.numE; i++)
            {
               if (energyUsedLEEA[j] == energyUsedHEEA[i] - eOff1 && !iFound)
               {
                  iFound = true;
                  EL[0] = i;
                  EL[1] = j;
               }
            }
         }

         iFound = false; 
	 for (int i = dimHEEA.numE - 1; i >= dimHEEA.numEfly; i--)
         {
	    for (int j = dimLEEA.numE - 1; j >= dimLEEA.numEfly; j--)
            {
               if (energyUsedHEEA[i] == energyUsedLEEA[j] - eOff2 && !iFound) 
               { 
                  iFound = true;
                  EL[2] = i + 1;
                  EL[3] = j + 1; 
               }
            }
         }

         if (sensorOverlapType == 0) 		// LEEA & HEEA partially overlap, LEEA lower energies
         {   
            eBinLimits[0].low = dimHEEA.numEfly;
            eBinLimits[0].high = EL[0];
	    eBinLimits[1].low  = EL[3];
	    eBinLimits[1].high  = dimLEEA.numE;
	    eBinLimits[2].low = EL[1];
	    eBinLimits[2].high = EL[3];
	    eBinLimits[3].low = EL[0];
	    eBinLimits[3].high = EL[2];
	 }
	 else if (sensorOverlapType == 1)	// HEEA fully overlaps LEEA (T,B,L,H)
         {
            eBinLimits[0].low = dimHEEA.numEfly;
            eBinLimits[0].high = EL[0];
	    eBinLimits[1].low = EL[2];
	    eBinLimits[1].high = dimHEEA.numE;
	    eBinLimits[2].low = EL[1];
	    eBinLimits[2].high = EL[3];
	    eBinLimits[3].low = EL[0];
	    eBinLimits[3].high = EL[2];
	 }
	 else if (sensorOverlapType == 2) 	// LEEA fully overlaps HEEA
	 {
            eBinLimits[0].low = dimLEEA.numEfly;
            eBinLimits[0].high = EL[1];
            eBinLimits[1].low = EL[3];
            eBinLimits[1].high = dimLEEA.numE;
            eBinLimits[2].low = dimHEEA.numEfly;
            eBinLimits[2].high = dimHEEA.numE;
            eBinLimits[3].low = EL[1];
            eBinLimits[3].high = EL[3];
	 }
	 else if (sensorOverlapType == 3)	// LEEA & HEEA partially overlap, HEEA lower energies
	 {

	 }
	 else if (sensorOverlapType == 4)	// LEEA & HEEA cover the same energy range
	 {
            eBinLimits[0].low = dimHEEA.numEfly;
	    eBinLimits[0].high = dimHEEA.numE;
	    eBinLimits[1].low = dimLEEA.numEfly;
	    eBinLimits[1].high = dimLEEA.numE;
	    eBinLimits[2].low = -1;
	    eBinLimits[2].high = -1;
	    eBinLimits[3].low = -1;
	    eBinLimits[3].high = -1;

// if the user specified LEEA or TL1L2B, select LEEA. If the user specified HEEA or
// TH1H2B, select HEEA.

	    if (sensorCombination == _LEEA || sensorCombination == _TL1L2B)
	    {
               useB = true;
               useLO = false;
	       useHO = false;
	       useT = false;
	    }
	    else if (sensorCombination == _HEEA || sensorCombination == _TH1H2B)
	    {
               useB = false;
	       useLO = false;
	       useHO = false;
	       useT = true;
	    }
	    else
	    {
               useB = false;
	       useLO = false;
	       useHO = false;
	       useT = false;
	    }
	 }
	 else
	 {
            useB = false;
	    useLO = false;
	    useHO = false;
	    useT = false;
         }

	 //cout << "FOUND : " << EL[0] << " : " << EL[1] << " : " << EL[2] << " : " << EL[3] << endl;

	// cout << "HO = " << eBinLimits[3].low << " to " << eBinLimits[3].high << endl;

     // cout << "LIMITS = " << eBinLimits[0].low << " : " << eBinLimits[0].high << " / " << eBinLimits[1].low << " : " << eBinLimits[1].high << " / " << eBinLimits[2].low << " : " << eBinLimits[2].high << " / " << eBinLimits[3].low << " : " << eBinLimits[3].high << " / " <<endl;
      //cout << "RAW = " << EL[0] << " : " << EL[1] << " : " << EL[2] << " : " << EL[3] << " limits L = " << dimLEEA.numEfly << " : " << dimLEEA.numE << " // " << dimHEEA.numEfly << " : " << dimHEEA.numE << endl;

   }

// set variables which specify what sensors are used for the top, overlap and bottom regions

   if (useT) sensorUsedTopRegion = senT;
   else sensorUsedTopRegion = -1;

   return(returnValue);
}

//-----------------------------------------------------------------------------
//
// carry out moments integration
//
//-----------------------------------------------------------------------------

int PeaceData::integrate(int sensorCombination, int usePotentialCorrection, int whichAzimuth,
                         bool testingInt_)
{
   testingInt = testingInt_;

   int returnT = _SUCCESS, returnB = _SUCCESS, 
       returnLO = _SUCCESS, returnHO = _SUCCESS;

// initialize total standard moments

   sm_n = 0.0;
   for (int i=0; i<3; i++)
   {
      sm_nu[i]  = 0.0;
      sm_Q[i]   = 0.0;
      sm_P[i]   = 0.0;
      sm_P[i+3] = 0.0;
   }

// set energy integration limits

   setEnergyRegionLimits(sensorCombination);

// check for complete set of pitch angles if necessary

   if (sensorCombination == _OVERLAP_TWO_SEC)
   {
      int check = checkData(whichAzimuth);
      if (check != _SUCCESS) return(_FAILED);
   }

// perform integration for top region if required

   if (useT)
   {
      returnT = integratePartial(0, senT, usePotentialCorrection, whichAzimuth, sensorCombination);
   }

// perform integration for bottom region if required
   
   if (useB)
   {
      returnB = integratePartial(1, senB, usePotentialCorrection, whichAzimuth, sensorCombination);
   }

// perform integration for LEEA overlap region if required
   
   if (useLO)
   {
      returnLO = integratePartial(2, senLO, usePotentialCorrection, whichAzimuth, sensorCombination);
   }

// perform integation for HEEA overlap region if required
   
   if (useHO)
   {
      returnHO = integratePartial(3, senHO, usePotentialCorrection, whichAzimuth, sensorCombination);
   }

   if (returnT == _SUCCESS && returnB == _SUCCESS &&
       returnLO == _SUCCESS && returnHO == _SUCCESS) return(_SUCCESS);
   else return(_FAILED);

}

//-----------------------------------------------------------------------------
//
// check for missing bins
// 
//-----------------------------------------------------------------------------

int PeaceData::checkData(int aziTry)
{

// set the lower & upper energy limits. Note that for PITCH_FULL* & PITCH_3D*
// data, we use the top and bottom regions, not the traditional "overlap" regions
// (the data in the CDF/CEF input file is already arranged into top, overlap and
// bottom regions)

   int e_st_leea = eBinLimits[1].low;
   int e_sp_leea = eBinLimits[1].high;
   int e_st_heea = eBinLimits[0].low;
   int e_sp_heea = eBinLimits[0].high;

   int pmaxL = dimLEEA.numP;
   int pmaxH = dimHEEA.numP;
   int emaxL = actualArraySizeLEEA;
   int emaxH = actualArraySizeHEEA;

//   cout << "DATA_CHECK (en lo to hi) " << e_st_leea << " to " << e_sp_leea << endl;

   //cout  << "DATA_CHECK_FOR_AZIMUTH " << aziTry << endl;
   //cout <<  "----------------------" << endl;
   int count_e = 0, count_ne = 0;
   for (int jl=e_st_leea; jl<e_sp_leea; jl++)
   {

      bool bj = false;
      if (energyLEEA[jl] > 2070.0 && energyLEEA[jl] < 2080.0) bj = true;

      int jh_u = -1;
      for (int jh=e_st_heea; jh<e_sp_heea; jh++)
      {
         if (energyUsedLEEA[jl] == energyUsedHEEA[jh]) jh_u = jh;
      }
      if (jh_u > -1)
      {
         int count = 0, countL = 0, countH = 0;
         for (int pa=0; pa<pmaxL; pa++)
         {
            if (*(dataStoreLEEA + pmaxL*emaxL*aziTry + emaxL*pa + jl) >= 0.0)
            {
               count++; countL++;
               //if (bj) cout << "PSD (L) = " << *(dataStoreLEEA + pmaxL*emaxL*aziTry + emaxL*pa + jl) << " from " << jl << " : " << pa << " :: " << pmaxL << " : " << emaxL << endl;
            }
            if (*(dataStoreHEEA + pmaxH*emaxH*aziTry + emaxH*pa + jh_u) >= 0.0)
            {
	       count++; countH++;
	       //if (bj) cout << "PSD (H) = " << *(dataStoreHEEA + pmaxH*emaxH*aziTry + emaxH*pa + jh_u) << " from " << jl << " : " << pa << " :: " << pmaxH << " : " << emaxH << endl;
	    }
	    //cout << "PSD L/H " << pa << " : " << *(dataStoreLEEA + pmaxL*emaxL*aziTry + emaxL*pa + jl)  << " : " << *(dataStoreHEEA + pmaxH*emaxH*aziTry + emaxH*pa + jh_u) << endl;
      }
      //if (bj) cout << "energy = " << jl << " : COUNT =  " << count << " ( " << energyLEEA[jl] << " ) counts L,H = " << countL << " & " << countH << endl;
         if (count == numPaForMoms) count_e++;
         count_ne++;
      }
   }

 //  if (bj) cout << "toal with 12/13 = " << count_e << " from " << count_ne << " and zimuth = " << aziTry << endl;

   if (count_e == count_ne) return(_SUCCESS);
   else return(_FAILED);
}

//-----------------------------------------------------------------------------
//
// carry out moments integration for a particular energy region
//
//-----------------------------------------------------------------------------

// region : 0 = top, 1 = bottom, 2 = LEEA overlap, 3 = HEEA overlap

int PeaceData::integratePartial(int region, int sensor, int applyPotentialCorrection,
                                int whichAzimuth, int sensorCombination)
{

   int doNotIncludeResults = 0;

// decide whether to use restricted integration limits

   int useEnergyLimits = 0;
   int usePolarLimits = 0;
   int useAzimuthLimits = 0;

   if (energyRestrictLow >= 0 && energyRestrictHigh > 0) useEnergyLimits = 1;
   if (polarLimitLow >= 0 && polarLimitHigh > 0) usePolarLimits = 1;
   if (azimuthLimitLow >= 0 && azimuthLimitHigh > 0) useAzimuthLimits = 1;

// determine energy limits

   double energyUsedLowest = 2.0e+07, energyUsedHighest = -2.0e+07;

// prepare for integration

   struct PeaceDataDim dimThis;

   int phi_st = 0, phi_sp = 0;
   int pol_st = 0, pol_sp = 0;
   int en_st = 0, en_sp = 0;

   double en1, en2, enl, enh, currentPotential = 0.0;

   double V2SUM, V3SUM, V4SUM, V5SUM, TVSUM1, TVSUM2, TVSUM3, TVSUM4;
   double TVSUM5, TVSUM6, TVSUM7, TVSUM8;
   double ECji, piece, th1, th2, ph1, ph2;

   float *energyThis, *energyThisLow, *energyThisHigh, *dataThis;
   float *potentialThis;

// set sensor & all details as appropriate for this sensor

   int  distType = 0, numActualE = 0;

   if (sensor == _LEEA)
   {
      dimThis.numA   = dimLEEA.numA;
      dimThis.numP   = dimLEEA.numP;
      dimThis.numE   = dimLEEA.numE;
      energyThis     = energyLEEA;
      energyThisLow  = energyLowLEEA;
      energyThisHigh = energyHighLEEA;
      dataThis       = dataStoreLEEA;
      distType       = distTypeLEEA;
      potentialThis  = potentialLEEA;
      numActualE     = actualArraySizeLEEA;
   }
   else
   {
      dimThis.numA   = dimHEEA.numA;
      dimThis.numP   = dimHEEA.numP;
      dimThis.numE   = dimHEEA.numE;
      energyThis     = energyHEEA;
      energyThisLow  = energyLowHEEA;
      energyThisHigh = energyHighHEEA;
      dataThis       = dataStoreHEEA;
      distType       = distTypeHEEA;
      potentialThis  = potentialHEEA;
      numActualE     = actualArraySizeHEEA;
   }

// azimuth fill & zero table counters

   int fillTableAzi[64] = {0}, zeroTableAzi[64] = {0};
   int fillAziMax, zeroAziMax;

// set the appropriate energy limits for the region under consideration

   en_st = eBinLimits[region].low;
   en_sp = eBinLimits[region].high;

   if (en_st < 0 || en_sp < 0 || en_st > numActualE || en_sp > numActualE) return(_FAILED);

// set the polar limits, including possibility of user-defined integration
// limits

   pol_st = 0;
   pol_sp = dimThis.numP;

   if (polarRestrictStatus == _ON)
   {
      if (polarRestrictLow >= pol_st && polarRestrictLow <= pol_sp &&
          polarRestrictHigh >= pol_st && polarRestrictHigh <= pol_sp &&
          polarRestrictLow < polarRestrictHigh)
      {
         pol_st = (int)polarRestrictLow;
         pol_sp = (int)polarRestrictHigh;
      }
   }

// set the azimuthal limits

   phi_st = 0;
   phi_sp = dimThis.numA;

// initialize the current contribution to standard moments

   double smp_n = 0, smp_nu[3] = {0}, smp_P[6] = {0}, smp_Q[3] = {0};

// phi integration angles

   double anglesPhi[64][2] = {0};
   
   if (sensorCombination == _OVERLAP_TWO_SEC)
   {

// for pitch angle data we assume gyroptropy

      for (int i=0; i<dimThis.numA; i++)
      {
         anglesPhi[i][0] = 0.0;
         anglesPhi[i][1] = 360.0;
      }
   }
   else
   {
      double areaAzi = 360.0/(double)dimThis.numA;
      for (int i=0; i<dimThis.numA; i++)
      {
         anglesPhi[i][0] = (double)i*areaAzi;
         anglesPhi[i][1] = (double)(i+1)*areaAzi;
      }

// adjust azimuthal angles if necessary for HEEA sensor - we use a coordinate
// system centred on the LEEA sensor, HEEA is located 180 degrees from LEEA

      if (sensor == _HEEA)
      {
         for (int i=0; i<dimThis.numA; i++)
         {
            anglesPhi[i][0] += 180.0;
            anglesPhi[i][1] += 180.0;
            if (anglesPhi[i][0] >= 360.0) anglesPhi[i][0] -= 360.0;
            if (anglesPhi[i][1] > 360.0) anglesPhi[i][1] -= 360.0;
         }
      }
   }

// polar integration angles

   double anglesPol[13][2] = {0};

   double areaPol = 180.0/(double)dimThis.numP;
   for (int i=0; i<dimThis.numP; i++)
   {
      anglesPol[i][0] = (double)i*areaPol;
      anglesPol[i][1] = (double)(i+1)*areaPol;
   }

// set the azimuthal offset for PITCH_FULL* & PITCH_3D* products; we need to
// calculate moments from just a single azimuth

   if (sensorCombination == _OVERLAP_TWO_SEC)
   {
      phi_st = whichAzimuth;
      phi_sp = whichAzimuth + 1;
   }

//----------------------
// azimuthal integration
//----------------------

   for (int k = phi_st; k < phi_sp; k++)
   {
      TVSUM1 = 0.0;
      TVSUM2 = 0.0;
      TVSUM3 = 0.0;
      TVSUM4 = 0.0;
      TVSUM5 = 0.0;
      TVSUM6 = 0.0;
      TVSUM7 = 0.0;            
      TVSUM8 = 0.0;

      ph1 = anglesPhi[k][0];
      ph2 = anglesPhi[k][1];

// determine the S/C potential
      
      if (applyPotentialCorrection == 1)
      {
         if (fabs(currentPotential) > MAX_VALUE_PM) doNotIncludeResults = 1; // check for fill values
         currentPotential = potentialOffset1 - *(potentialThis + k);
      }
      else currentPotential = 0.0;      
      
// determine the energy bin containing the S/C potential
// - remember that the energies are ordered from high to low (this is the opposite to PEACEmoments
//   IDFS, where energies are ordered from low to high)

      int energyBinContainingPotential = en_sp;
      if (currentPotential >= (*(energyThis + en_sp - 1) - *(energyThisLow + en_sp - 1)))
      {

// when S/C potential above lowest PEACE energy, just find the
// PEACE energy bin containing the potential

         for (int j=en_st; j<en_sp; j++)
	 {
            if (currentPotential >= (*(energyThis + j) - *(energyThisLow + j)) &&
                currentPotential < *(energyThis + j) + *(energyThisHigh + j))
	    {
               energyBinContainingPotential = j;
	    }
	 }
      }
      else
      {

// we need to handle the case when the lowest PEACE energy is above
// the S/C potential. We still want to be able to remove energy bins
// above the potential.



      }

//------------------
// polar integration
//------------------

      for (int i = pol_st; i < pol_sp; i++)
      {
         V2SUM = 0.0;
         V3SUM = 0.0;
         V4SUM = 0.0;
         V5SUM = 0.0;

         th1 = anglesPol[i][0];
         th2 = anglesPol[i][1];

//-------------------
// energy integration
//-------------------

         for (int j = en_st; j < en_sp; j++)
         {
            bool enDont = false;

            piece = *(dataThis+dimThis.numP*numActualE*k+numActualE*i+j);

// check for fills in each azimuth

            if (fabs(piece) > MAX_VALUE_PM) fillTableAzi[k]++;

// check for zeroes in each azimuth

            if (fabs(piece) == 0.0) zeroTableAzi[k]++;

// set the energy lower & upper limits

            enl = *(energyThis + j) - *(energyThisLow + j);
            enh = *(energyThis + j) + *(energyThisHigh + j);

// remove acceleration due to S/C potential

            en1 = enl - currentPotential;
            en2 = enh - currentPotential;	    

// take into account S/C potential offset 2, i.e. low energy cutoff

            if (potentialOffset2 != 0)
            {
               if (potentialOffset2units == _EV)
               {
                  if (enl < currentPotential + potentialOffset2) enDont = true;
               }
               else
               {
                  if (energyBinContainingPotential - potentialOffset2 > j)
                  {
                  }
                  else
                  {
                     enDont = true;
                  }
               }
               //cout << "POT_OFFSET = " << energyBinContainingPotential << " from " << enDont << endl;
            }
	    
// handle user-defined restricted energy integration

            double interpFactorE = 1.0;    // energy interpolation factor
	    double enEl = enl, enEh = enh; // energy bin limits

	    if (energyRestrictScpType == 1)
	    {
	       enEl = en1; // user has specfied that they are looking at a S/C potential
	       enEh = en2; //  corrected spectrogram
	    }

	    if (energyRestrictStatus == _ON && energyRestrictType == 0 && 
	        enEl >= energyRestrictLow && enEh < energyRestrictHigh)
            {
	    }
	    else if (energyRestrictStatus == _ON && energyRestrictType == 1 &&
	             enEh >= energyRestrictLow && enEl < energyRestrictHigh)
	    {
	       if (energyRestrictLow >= enEl && energyRestrictLow < enEh)
	       {
	       
// user's lower energy integration limit is contained in the current energy bin	       
	       
	          interpFactorE = (enEh - energyRestrictLow)/(enEh - enEl);
	       
	       }
	       else if (energyRestrictHigh >= enEl && energyRestrictHigh < enEh)
	       {
	       
// user's upper energy integration limit is containined in the current energy bin
        	       
	          interpFactorE = (energyRestrictHigh - enEl)/(enh - enEl);
	       
	       }
	    }     
	    else if (energyRestrictStatus == _OFF)
	    {
            }	    
	    else enDont = true;	    	    

// only continue if lower energy limit is positive, i.e. energy bin is entirely above
// S/C potential

            if (en1 > 0.0 && !enDont)
            {
     
// convert energy to velocity

               en1 = sqrt(2.0*evjoule*en1/mass);
               en2 = sqrt(2.0*evjoule*en2/mass);

// determine lowest & highest energies
               
               if (en1 < energyUsedLowest) energyUsedLowest = en1;
               if (en2 > energyUsedHighest) energyUsedHighest = en2;

// get PEACE data, convert units as appropriate
 
	       ECji = piece*interpFactorE/1.0e+18;
               if (testingInt) ECji = testingIntValue/1.0e+18;

// check for fill values

               if (fabs(piece) > MAX_VALUE_PM) fillTableAzi[k] += 1;

// check for zero values

               if (fabs(piece) == 0.0) zeroTableAzi[k] += 1;

// add contributions from this energy bin if necessary

	       if (ECji > 0.0 && fabs(piece) < MAX_VALUE_PM)
	       {
                  V2SUM += ECji*VN_PM(en2, en1, 2);
	          V3SUM += ECji*VN_PM(en2, en1, 3);
	          V4SUM += ECji*VN_PM(en2, en1, 4);
	          V5SUM += ECji*VN_PM(en2, en1, 5);
	       }
            }

	 } // <- end velocity integration

         TVSUM1 += T_PM(th2, th1, 2)*V2SUM;
         TVSUM2 += T_PM(th2, th1, 4)*V3SUM;
	 TVSUM3 += T_PM(th2, th1, 6)*V3SUM;
	 TVSUM4 += T_PM(th2, th1, 7)*V4SUM;
	 TVSUM5 += T_PM(th2, th1,10)*V4SUM;
	 TVSUM6 += T_PM(th2, th1, 9)*V4SUM;
	 TVSUM7 += T_PM(th2, th1, 4)*V5SUM;
	 TVSUM8 += T_PM(th2, th1, 6)*V5SUM;

      } // <- end polar integration

      smp_n     += T_PM(ph2, ph1, 1)*TVSUM1;
      smp_nu[0] += T_PM(ph2, ph1, 3)*TVSUM2;
      smp_nu[1] += T_PM(ph2, ph1, 2)*TVSUM2;
      smp_nu[2] += T_PM(ph2, ph1, 1)*TVSUM3;
      smp_P[0]  += T_PM(ph2, ph1, 5)*TVSUM4;
      smp_P[1]  += T_PM(ph2, ph1, 4)*TVSUM4;
      smp_P[2]  += T_PM(ph2, ph1, 1)*TVSUM5;
      smp_P[3]  += T_PM(ph2, ph1, 6)*TVSUM4;
      smp_P[4]  += T_PM(ph2, ph1, 2)*TVSUM6;
      smp_P[5]  += T_PM(ph2, ph1, 3)*TVSUM6;
      smp_Q[0]  += T_PM(ph2, ph1, 3)*TVSUM7;
      smp_Q[1]  += T_PM(ph2, ph1, 2)*TVSUM7;
      smp_Q[2]  += T_PM(ph2, ph1, 1)*TVSUM8;

   } // <- end azimuth integration

// integration testing

   if (testingInt && testingIntValue > 0)
   {

      double exact = testingIntValue*(4.0/3.0)*pi/1.0e+18*
                     (energyUsedHighest*energyUsedHighest*energyUsedHighest - 
                      energyUsedLowest*energyUsedLowest*energyUsedLowest);
 
       std::cout << "__INTEGRATION_TEST exact,numerical = " << exact << " , " << smp_n << std::endl;
   }

// check if some azimuths only contain fill values. We don't want to
// include spins like this as we don't have a complete set of azimuths.

   fillAziMax = dimThis.numP*(en_sp - en_st);
   for (int k=0; k<dimThis.numA; k++)
   {
      if (fillTableAzi[k] == fillAziMax) doNotIncludeResults = 1;
   }

// check if some azimuths only contain zeros

   zeroAziMax = dimThis.numP*(en_sp - en_st);
   int iZeroAzCounter = 0;
   for (int k = 0; k < dimThis.numA; k++)
   {
      if (zeroTableAzi[k] == zeroAziMax) iZeroAzCounter++;
   }

   iZeroAzCounter = dimThis.numA - iZeroAzCounter;

// apply mass corrections where necessary & add this partial contribution
// to total standard moments

   if (doNotIncludeResults == 0)
   {
      sm_n     += smp_n;
      sm_nu[0] += smp_nu[0];
      sm_nu[1] += smp_nu[1];
      sm_nu[2] += smp_nu[2];
      sm_P[0]  += smp_P[0]*mass;
      sm_P[1]  += smp_P[1]*mass;
      sm_P[2]  += smp_P[2]*mass;
      sm_P[3]  += smp_P[3]*mass;
      sm_P[4]  += smp_P[4]*mass;
      sm_P[5]  += smp_P[5]*mass;
      sm_Q[0]  += smp_Q[0]*mass/2.0;
      sm_Q[1]  += smp_Q[1]*mass/2.0;
      sm_Q[2]  += smp_Q[2]*mass/2.0;
   }

   if (doNotIncludeResults == 0) return(_SUCCESS);
   else return(_FAILED);
}

//-----------------------------------------------------------------------------
//
// convert standard moments to moments relative to the bulk flow
//
//-----------------------------------------------------------------------------

void PeaceData::convertToRelativeMoments()
{
   densityIsFill = false, velocityIsFill = false, pressureIsFill = false;

//--------
// density
//--------

   n = sm_n;
   if (fabs(n) > MAX_VALUE_PM) densityIsFill = true;

//---------
// velocity
//---------

   for (int i=0; i<3; i++)
   {
      v[i] = (n > 0.0 && fabs(sm_nu[i]) < MAX_VALUE_PM) ? sm_nu[i]/n : FILL_VALUE_PM;
      if (fabs(v[i]) > MAX_VALUE_PM) velocityIsFill = true;
   }

   if (!velocityIsFill) v_mag = sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);

//------------
// temperature
//------------

   if (fabs(sm_P[0]) < MAX_VALUE_PM && fabs(sm_P[1]) < MAX_VALUE_PM &&
       fabs(sm_P[2]) < MAX_VALUE_PM && fabs(sm_P[3]) < MAX_VALUE_PM)
   {
   }
   else pressureIsFill = true;

   double u2 = v[0]*v[0] + v[1]*v[1] + v[2]*v[2];

   if (n > 0.0 && !densityIsFill && !velocityIsFill && !pressureIsFill)
   {
      T = (1.0/(3.0*kBC))*((sm_P[0] + sm_P[1] + sm_P[2])/n - mass*u2);
   }
   else T = FILL_VALUE_PM;

//----------------
// pressure tensor
//----------------

   if (n > 0.0 && !densityIsFill && !velocityIsFill && !pressureIsFill)
   {
      for (int i=0; i<6; i++)
      {
         int i1 = 0, i2 = 0;
         vectorToTensorIndex(i, &i1, &i2);
         P[i1][i2] = sm_P[i] - mass*n*v[i1]*v[i2];
         if (i1 != i2) P[i2][i1] = P[i1][i2];
      }
   }
   else
   {
      for (int i=0; i<6; i++)
      {
         int i1 = 0, i2 = 0; 
	 vectorToTensorIndex(i, &i1, &i2);
         P[i1][i2] = FILL_VALUE_PM;
         if (i1 != i2) P[i2][i1] = P[i1][i2];
      }
   }

//----------
// heat flow
//----------

   double uP[3];
   uP[0] = v[0]*sm_P[0] + v[1]*sm_P[3] + v[2]*sm_P[5];
   uP[1] = v[0]*sm_P[3] + v[1]*sm_P[1] + v[2]*sm_P[4];
   uP[2] = v[0]*sm_P[5] + v[1]*sm_P[4] + v[2]*sm_P[2];

   if (n > 0.0 && !densityIsFill && !velocityIsFill && !pressureIsFill)
   {
      for (int i=0; i<3; i++)
      {
         H[i] = sm_Q[i] - v[i]*(sm_P[0] + sm_P[1] + sm_P[2])/2.0
                - uP[i] + mass*n*u2*v[i];
      }
   }
   else
   {
      for (int i=0; i<3; i++)
      {
         H[i] = FILL_VALUE_PM;
      }
   }

//------------
// energy flux
//------------

   if (n > 0.0 && !densityIsFill && !velocityIsFill)
   {
      for (int i=0; i<3; i++)
      {
         Q[i] = (fabs(sm_Q[i]) < MAX_VALUE_PM) ? sm_Q[i] : FILL_VALUE_PM;
      }
   }
   else
   {
      for (int i=0; i<3; i++)
      {  
         Q[i] = FILL_VALUE_PM;
      }
   }

}

//-----------------------------------------------------------------------------
//
// rotate to S/C or GSE coordinates
//
//-----------------------------------------------------------------------------

int PeaceData::applyRotation(int coordinateSystem)
{

   if (coordinateSystem == _BFA)
   {
      currentCoordinateSystem = _BFA;
      return(_SUCCESS);
   }

   int returnValue = 0;

// don't even bother trying to rotate to another coordinate system if we just have
// fill values

   if (densityIsFill || velocityIsFill || pressureIsFill) return(1);

//----------------------------------------------------------
// apply rotation from PEACE to S/C coordinates if necessary
//----------------------------------------------------------

   if (coordinateSystem == _SPACECRAFT || coordinateSystem == _GSE ||
       coordinateSystem == _SR2 || coordinateSystem == _ISR2)
   {

// define rotation matrix from PEACE to S/C coordinates

      double TRm[3][3] = {0};
      TRm[0][2] = 1.0;
      TRm[1][0] = sqrt(3.0)/2.0;
      TRm[1][1] = 0.5;
      TRm[2][0] = -0.5;
      TRm[2][1] = sqrt(3.0)/2.0;

// rotate velocity, energy flux & heat flux

      double vR[3] = {0}, QR[3] = {0}, HR[3] = {0};
      for (int i=0; i<3; i++)
      {
         for (int j=0; j<3; j++)
         {
            vR[i] += TRm[i][j]*v[j];
            QR[i] += TRm[i][j]*Q[j];
            HR[i] += TRm[i][j]*H[j];
         }
      }

      for (int i=0; i<3; i++)
      {
         v[i] = vR[i];
         Q[i] = QR[i];
         H[i] = HR[i];
      }

// rotate pressure tensor

      double PR[3][3] = {0};
      for (int i=0; i<3; i++)
      {
         for (int j=0; j<3; j++)
         {
            for (int it=0; it<3; it++)
            {
               for (int jt=0; jt<3; jt++)
               {
                  PR[i][j] += TRm[i][it]*TRm[j][jt]*P[it][jt];
               }
            }
         }
      }

      for (int i=0; i<3; i++)
      {
         for (int j=0; j<3; j++)
         {
            P[i][j] = PR[i][j];
         }
      }

// set the current coordinate system flag

      currentCoordinateSystem = _SPACECRAFT;
   }

//--------------------------------------------------------
// apply rotation from S/C to GSE coordinates if necessary
//--------------------------------------------------------

   if ((coordinateSystem == _GSE || coordinateSystem == _SR2 || coordinateSystem == _ISR2)
        && fabs(spos) < MAX_VALUE_PM &&
        fabs(spinAxisLatitudeGSE) < MAX_VALUE_PM && fabs(spinAxisLongitudeGSE) < MAX_VALUE_PM)
   {
      double TRm[3][3] = {0};

// determine the required rotation matrix 

      sc2gse(spinAxisLatitudeGSE, spinAxisLongitudeGSE, TRm, spos);

// rotate velocity, energy flux & heat flux

      double vR[3] = {0}, QR[3] = {0}, HR[3] = {0};
      for (int i=0; i<3; i++)
      {  
         for (int j=0; j<3; j++)
         {
            vR[i] += TRm[i][j]*v[j];
            QR[i] += TRm[i][j]*Q[j];
            HR[i] += TRm[i][j]*H[j];
         }
      }
        
      for (int i=0; i<3; i++)
      {
         v[i] = vR[i];
         Q[i] = QR[i];
         H[i] = HR[i];
      }

// rotate pressure tensor

      double PR[3][3] = {0};
      for (int i=0; i<3; i++)
      {
         for (int j=0; j<3; j++)
         {
            for (int it=0; it<3; it++)
            {
               for (int jt=0; jt<3; jt++)
               {
                  PR[i][j] += TRm[i][it]*TRm[j][jt]*P[it][jt];
               }  
            }  
         }  
      }  
          
      for (int i=0; i<3; i++)
      {
         for (int j=0; j<3; j++)
         {
            P[i][j] = PR[i][j];
         }  
      }  

// set the current coordinate system flag

      currentCoordinateSystem = _GSE;
   }
   else if (coordinateSystem == _GSE || coordinateSystem == _ISR2 ||
            coordinateSystem == _SR2)
   {

// we couldn't perform the required coordinate system rotation, so set
// vectors/tensors to fills

      setAllVectorsToFills();
      returnValue = -1;
   }

//---------------------------
// rotate to SR2 if necessary
//---------------------------

   if (coordinateSystem == _SR2 || coordinateSystem == _ISR2)
   {

// determine the rotation matrix

      double TR1[3][3] = {0};
      double RR[3] = {0};

      double theta = TRDEG2RAD*spinAxisLongitudeGSE;
      double phi = TRDEG2RAD*(90.0 - spinAxisLatitudeGSE);
         
      RR[0] = cos(theta)*sin(phi);
      RR[1] = sin(theta)*sin(phi);
      RR[2] = cos(phi);
         
      double alpha = 1.0/sqrt(RR[1]*RR[1] + RR[2]*RR[2]);

      TR1[0][0] = alpha*(RR[0]*RR[0] + RR[2]*RR[2]);
      TR1[0][1] = -alpha*RR[0]*RR[1];
      TR1[0][2] = -alpha*RR[0]*RR[2];
      
      TR1[1][0] = 0;
      TR1[1][1] = alpha*RR[2];
      TR1[1][2] = -alpha*RR[1];
         
      TR1[2][0] = RR[0]; 
      TR1[2][1] = RR[1];
      TR1[2][2] = RR[2];

// rotate vectors, including B-field, as the B-field of course must be in the
// same coordinate system as PEACE moments

      double vR[3] = {0}, QR[3] = {0}, HR[3] = {0}, BR[3] = {0};
      for (int i=0; i<3; i++)
      {
         for (int j=0; j<3; j++)
         {
            vR[i] += TR1[i][j]*v[j];
            QR[i] += TR1[i][j]*Q[j];
            HR[i] += TR1[i][j]*H[j];
            BR[i] += TR1[i][j]*fgmB[j];
         }
      }

      for (int i=0; i<3; i++)
      {
         v[i] = vR[i];
         Q[i] = QR[i];
         H[i] = HR[i];
         fgmB[i] = BR[i];
      }

// rotate tensors

      double PR[3][3] = {0};
      for (int i=0; i<3; i++)
      {  
         for (int j=0; j<3; j++)
         {  
            for (int it=0; it<3; it++)
            {  
               for (int jt=0; jt<3; jt++)
               {  
                  PR[i][j] += TR1[i][it]*TR1[j][jt]*P[it][jt];
               }
            }
         }
      }
          
      for (int i=0; i<3; i++)
      {  
         for (int j=0; j<3; j++)
         {  
            P[i][j] = PR[i][j];
         }
      }

// set the current coordinate system flag

      currentCoordinateSystem = _SR2;

   }
   else if (coordinateSystem == _SR2 || coordinateSystem == _ISR2)
   {
      
// we couldn't perform the required coordinate system rotation, so set
// vectors/tensors to fills

      setAllVectorsToFills();
      returnValue = -1;
   }  

//----------------------------
// rotate to ISR2 if necessary
//----------------------------

   if (coordinateSystem == _ISR2)
   {

// determine the rotation matrix

      double TR2[3][3] = {0};
      
      TR2[0][0] = 1;
      TR2[1][1] = -1;
      TR2[2][2] = -1;

// rotate vectors (including B-field of course)

      double vR[3] = {0}, QR[3] = {0}, HR[3] = {0}, BR[3] = {0};
      for (int i=0; i<3; i++)
      {
         for (int j=0; j<3; j++)
         {
            vR[i] += TR2[i][j]*v[j];
            QR[i] += TR2[i][j]*Q[j];
            HR[i] += TR2[i][j]*H[j];
            BR[i] += TR2[i][j]*fgmB[j];
         }
      }

      for (int i=0; i<3; i++)
      {
         v[i] = vR[i];
         Q[i] = QR[i];
         H[i] = HR[i];
         fgmB[i] = BR[i];
      }

// rotate tensors

      double PR[3][3] = {0};
      for (int i=0; i<3; i++)
      {  
         for (int j=0; j<3; j++)
         {  
            for (int it=0; it<3; it++)
            {  
               for (int jt=0; jt<3; jt++)
               {  
                  PR[i][j] += TR2[i][it]*TR2[j][jt]*P[it][jt];
               }
            }
         }
      }
          
      for (int i=0; i<3; i++)
      {  
         for (int j=0; j<3; j++)
         {  
            P[i][j] = PR[i][j];
         }
      }

// set the current coordinate system flag

      currentCoordinateSystem = _ISR2;

   }
   else if (coordinateSystem == _ISR2)
   {

// we couldn't perform the required coordinate system rotation, so set
// vectors/tensors to fills

      setAllVectorsToFills();
      returnValue = -1;
   }


// generate para/perp components if possible

   if (generateParaPerp() < 0) returnValue = -1;

   return(returnValue);
}

//-----------------------------------------------------------------------------
//
// adjust units if necessary
//
//-----------------------------------------------------------------------------

void PeaceData::adjustUnits(void)
{

// convert density from m^-3 to cm^-3
	
   if (!densityIsFill) n = n/(1000000.0);
   
// convert velocity from m s^-1 to km s^-1

   if (!velocityIsFill)
   {
      for (int i=0; i<3; i++)
      {
         if (fabs(v[i]) < MAX_VALUE_PM) v[i] = v[i]/1000.0;
	 if (!magFieldIsFill)
	 {
	    if (fabs(vpara[i]) < MAX_VALUE_PM) vpara[i] = vpara[i]/1000.0;
	    if (fabs(vperp[i]) < MAX_VALUE_PM) vperp[i] = vperp[i]/1000.0;
	 }
      }
      if (!magFieldIsFill)
      {
         if (fabs(vpara_mag) < MAX_VALUE_PM) vpara_mag = vpara_mag/1000.0;
         if (fabs(vperp_mag) < MAX_VALUE_PM) vperp_mag = vperp_mag/1000.0;
      }
   }

// convert temperature to MK

   if (fabs(T) < MAX_VALUE_PM) T = T/1.0e+6;
   if (fabs(Tpara) < MAX_VALUE_PM) Tpara = Tpara/1.0e+6;
   if (fabs(Tperp) < MAX_VALUE_PM) Tperp = Tperp/1.0e+6;

// convert pressure to nPa

   for (int i=0; i<3; i++)
   {
      for (int j=0; j<3; j++)
      {
         if (fabs(P[i][j]) < MAX_VALUE_PM) P[i][j] = P[i][j]/1.0e+9;
      }
   }

// convert energy flux to u...

   if (!velocityIsFill && !pressureIsFill)
   {     
      for (int i=0; i<3; i++)
      {  
         if (fabs(Q[i]) < MAX_VALUE_PM) Q[i] = Q[i]/1.0e-06;
         if (!magFieldIsFill)
         {
            if (fabs(Qpara[i]) < MAX_VALUE_PM) Qpara[i] = Qpara[i]/1.0e-06;
            if (fabs(Qperp[i]) < MAX_VALUE_PM) Qperp[i] = Qperp[i]/1.0e-06;
         }
      }  
      if (!magFieldIsFill)
      {
         if (fabs(Qpara_mag) < MAX_VALUE_PM) Qpara_mag = Qpara_mag/1.0e-06;
         if (fabs(Qperp_mag) < MAX_VALUE_PM) Qperp_mag = Qperp_mag/1.0e-06;
      }  
   }

// convert heat flux to uJ m^-2 s^-1

   if (!velocityIsFill && !pressureIsFill)
   {
      for (int i=0; i<3; i++)
      {
         if (fabs(H[i]) < MAX_VALUE_PM) H[i] = H[i]/1.0e-06;
         if (!magFieldIsFill)
         {
            if (fabs(Hpara[i]) < MAX_VALUE_PM) Hpara[i] = Hpara[i]/1.0e-06;
            if (fabs(Hperp[i]) < MAX_VALUE_PM) Hperp[i] = Hperp[i]/1.0e-06;
         }  
      }  
      if (!magFieldIsFill)
      {
         if (fabs(Hpara_mag) < MAX_VALUE_PM) Hpara_mag = Hpara_mag/1.0e-06;
         if (fabs(Hperp_mag) < MAX_VALUE_PM) Hperp_mag = Hperp_mag/1.0e-06;
      }  
   }

   return;
}

//-----------------------------------------------------------------------------
//
// set all moments to fill values
//
//-----------------------------------------------------------------------------

void PeaceData::setAllMomentsToFills(void)
{
   n = FILL_VALUE_PM;
   T = FILL_VALUE_PM;
   setAllVectorsToFills();
   setAllParaPerpToFills();
   v_mag = FILL_VALUE_PM;
   j_mag = FILL_VALUE_PM;
   ce = FILL_VALUE_PM;
}

//-----------------------------------------------------------------------------
//
// set all vector/tensor moments to fill values
//
//-----------------------------------------------------------------------------

void PeaceData::setAllVectorsToFills(void)
{
   for (int i=0; i<3; i++)
   {
      v[i] = FILL_VALUE_PM;
      Q[i] = FILL_VALUE_PM;
      H[i] = FILL_VALUE_PM;
      j[i] = FILL_VALUE_PM;
      for (int j=0; j<3; j++)
      {
         P[i][j] = FILL_VALUE_PM;
      }
   }
}

//-----------------------------------------------------------------------------
//
// set all para/perp moments to fill values
//
//-----------------------------------------------------------------------------

void PeaceData::setAllParaPerpToFills(void)
{
   vpara_mag = FILL_VALUE_PM;
   vperp_mag = FILL_VALUE_PM;
   Qpara_mag = FILL_VALUE_PM;
   Qperp_mag = FILL_VALUE_PM;
   Hpara_mag = FILL_VALUE_PM;
   Hperp_mag = FILL_VALUE_PM;
   Tpara     = FILL_VALUE_PM;
   Tperp     = FILL_VALUE_PM;
   Tperpx    = FILL_VALUE_PM;
   Tperpy    = FILL_VALUE_PM;
   Ppara     = FILL_VALUE_PM;
   Pperp     = FILL_VALUE_PM;
   Pperpx    = FILL_VALUE_PM;
   Pperpy    = FILL_VALUE_PM;

   for (int i=0; i<3; i++)
   {
      vpara[i] = FILL_VALUE_PM;
      vperp[i] = FILL_VALUE_PM;
      Qpara[i] = FILL_VALUE_PM;
      Qperp[i] = FILL_VALUE_PM;
      Hpara[i] = FILL_VALUE_PM;
      Hperp[i] = FILL_VALUE_PM;
      jpara[i] = FILL_VALUE_PM;
      jperp[i] = FILL_VALUE_PM;
   }
}

//-----------------------------------------------------------------------------
//
// generate components parallel & perpendicular to B-field
//
//-----------------------------------------------------------------------------

int PeaceData::generateParaPerp(void)
{

   int returnValue = 0;
   
// check if the current coordinate system of the relative moments is GSE

   if (currentCoordinateSystem == _SENSOR || currentCoordinateSystem == _SPACECRAFT)
   {
      setAllParaPerpToFills();
      return(-1); // incorrect coordinate system
   }
   
// calculate parallel & perpendicular projections if we have proper B-field values

   if (fabs(fgmB[0]) < MAX_VALUE_PM && fabs(fgmB[1]) < MAX_VALUE_PM &&
       fabs(fgmB[2]) < MAX_VALUE_PM)
   {

//---------
// velocity
//---------

      double line1, line2;
      line1 = fgmB[0]*v[0] + fgmB[1]*v[1] + fgmB[2]*v[2];
      line2 = fgmB[0]*fgmB[0] + fgmB[1]*fgmB[1] + fgmB[2]*fgmB[2];

      if (fabs(line1) < MAX_VALUE_PM && fabs(line2) < MAX_VALUE_PM && line2 > 0.0)
      {
         for (int i=0; i<3; i++)
         {
            vpara[i] = line1*fgmB[i]/line2;
            vperp[i] = v[i] - vpara[i];
         }
         vpara_mag = sqrt(vpara[0]*vpara[0] + vpara[1]*vpara[1] + vpara[2]*vpara[2]);
         if (line1 < 0.0) vpara_mag = -vpara_mag; // change sign for anti-parallel
         vperp_mag = sqrt(vperp[0]*vperp[0] + vperp[1]*vperp[1] + vperp[2]*vperp[2]);
      }
      else
      {
         for (int i=0; i<3; i++)
         {
            vpara[i] = FILL_VALUE_PM;
            vperp[i] = FILL_VALUE_PM;
         }
         vpara_mag = FILL_VALUE_PM;
         vperp_mag = FILL_VALUE_PM;
      }

//------------
// energy flux
//------------

      line1 = fgmB[0]*Q[0] + fgmB[1]*Q[1] + fgmB[2]*Q[2];
      line2 = fgmB[0]*fgmB[0] + fgmB[1]*fgmB[1] + fgmB[2]*fgmB[2];

      if (fabs(line1) < MAX_VALUE_PM && fabs(line2) < MAX_VALUE_PM && line2 > 0.0)
      {
         for (int i=0; i<3; i++)
         {
            Qpara[i] = line1*fgmB[i]/line2;
            Qperp[i] = Q[i] - Qpara[i];
         }
         Qpara_mag = sqrt(Qpara[0]*Qpara[0] + Qpara[1]*Qpara[1] + Qpara[2]*Qpara[2]);
         if (line1 < 0.0) Qpara_mag = -Qpara_mag; // change sign for anti-parallel
         Qperp_mag = sqrt(Qperp[0]*Qperp[0] + Qperp[1]*Qperp[1] + Qperp[2]*Qperp[2]);
      }  
      else
      {  
         for (int i=0; i<3; i++)
         {
            Qpara[i] = FILL_VALUE_PM;
            Qperp[i] = FILL_VALUE_PM;
         }
         Qpara_mag = FILL_VALUE_PM;
         Qperp_mag = FILL_VALUE_PM;
      }

//----------
// heat flux
//----------

      line1 = fgmB[0]*H[0] + fgmB[1]*H[1] + fgmB[2]*H[2];
      line2 = fgmB[0]*fgmB[0] + fgmB[1]*fgmB[1] + fgmB[2]*fgmB[2];

      if (fabs(line1) < MAX_VALUE_PM && fabs(line2) < MAX_VALUE_PM && line2 > 0.0)
      {
         for (int i=0; i<3; i++)
         {
            Hpara[i] = line1*fgmB[i]/line2;
            Hperp[i] = H[i] - Hpara[i];
         }
         Hpara_mag = sqrt(Hpara[0]*Hpara[0] + Hpara[1]*Hpara[1] + Hpara[2]*Hpara[2]);
         if (line1 < 0.0) Hpara_mag = -Hpara_mag; // change sign for anti-parallel
         Hperp_mag = sqrt(Hperp[0]*Hperp[0] + Hperp[1]*Hperp[1] + Hperp[2]*Hperp[2]);
      }  
      else
      {  
         for (int i=0; i<3; i++)
         {
            Hpara[i] = FILL_VALUE_PM;
            Hperp[i] = FILL_VALUE_PM;
         }
         Hpara_mag = FILL_VALUE_PM;
         Hperp_mag = FILL_VALUE_PM;
      }

//-----------------------
// pressure & temperature
//-----------------------

      double TRm[3][3] = {0}, Pb[3][3] = {0};
      bool pressureIsFill = false;

// calculate the rotation matrix - the rotation is from the GSE coordinate
// system to a B-field aligned coordinate system where the z-axis is 
// parallel to the B-field

      gse2bfa(fgmB[0], fgmB[1], fgmB[2], TRm);

// rotate the pressure tensor

      for (int i=0; i<3; i++)
      {
         for (int j=0; j<3; j++)
         {
            for (int it=0; it<3; it++)
            {
               for (int jt=0; jt<3; jt++)
               {
                  if (fabs(P[it][jt]) > MAX_VALUE_PM) pressureIsFill = true;
                  Pb[i][j] += TRm[i][it]*TRm[j][jt]*P[it][jt];
               }
            }
         }
      }

      if (Pb[0][0] < 0.0) Pb[0][0] = -Pb[0][0];
      if (Pb[1][1] < 0.0) Pb[1][1] = -Pb[1][1];
      if (Pb[2][2] < 0.0) Pb[2][2] = -Pb[2][2];

// parallel & perpendicular components of the pressure

      if (!pressureIsFill)
      {
         Ppara  = Pb[2][2];
         Pperp  = (Pb[0][0] + Pb[1][1])/2.0;
         Pperpx = Pb[0][0];
         Pperpy = Pb[1][1];
      }
      else
      {
         Ppara  = FILL_VALUE_PM;
         Pperp  = FILL_VALUE_PM;
         Pperpx = FILL_VALUE_PM;
         Pperpy = FILL_VALUE_PM;
      }

// parallel & perpendicular components of the temperature

      if (n > 0.0 && !pressureIsFill)
      {
         Tpara  = Ppara/(n*kBC);
         Tperp  = Pperp/(n*kBC);
         Tperpx = Pperpx/(n*kBC);
         Tperpy = Pperpy/(n*kBC);
      }
      else
      {
         Tpara  = FILL_VALUE_PM;
         Tperp  = FILL_VALUE_PM;
         Tperpx = FILL_VALUE_PM;
         Tperpy = FILL_VALUE_PM;
      }

   }
   else
   {

// the FGM B-field contains fill values, so we don't try to calculate any 
// parallel or perpendicular projections. Set everything to fill values.

      setAllParaPerpToFills();
      magFieldIsFill = true;

      returnValue = -2;
   }

   return(returnValue);

}

//-----------------------------------------------------------------------------
//
// return the current coordinate system
//
//-----------------------------------------------------------------------------

int PeaceData::currentCoordSystem(void)
{
   return(currentCoordinateSystem);
}

//-----------------------------------------------------------------------------
//
// convert pressure in vector format to tensor format
//
//-----------------------------------------------------------------------------

void PeaceData::vectorToTensorIndex(int in, int *i1, int *i2)
{

// the pressure tensor is symmetric, so we sometimes store it as a vector
// with 6 components. The remaining off-diagonal components can easily be
// determined because of the symmetry. This function determines the indicies
// for a 3x3 matrix corresponding to the vector index. The order of the 
// components in the vector is as follows: Pxx, Pyy, Pzz, Pxy, Pyz, Pxz.

   switch(in)
   {
      case 0:
      *i1 = 0;
      *i2 = 0;
      break;

      case 1: 
      *i1 = 1;
      *i2 = 1;
      break;

      case 2:
      *i1 = 2;
      *i2 = 2;
      break;

      case 3:
      *i1 = 0;
      *i2 = 1;
      break;

      case 4:
      *i1 = 1;
      *i2 = 2;
      break;

      case 5:
      *i1 = 0;
      *i2 = 2;
      break;
   }
}

//-----------------------------------------------------------------------------
//
// set restricted energy integration parameters
//
//-----------------------------------------------------------------------------

void PeaceData::setRestrictedEnergyIntegration(int eS, int eT, int scpT, double eL, double eU)
{
   energyRestrictScpType = scpT;
   energyRestrictStatus = eS;
   energyRestrictType = eT;
   energyRestrictLow = eL;
   energyRestrictHigh = eU;
}

//-----------------------------------------------------------------------------
//
// set restricted polar integration parameters
//
//-----------------------------------------------------------------------------

void PeaceData::setRestrictedPolarIntegration(int pS, double pL, double pU)
{
   polarRestrictStatus = pS;
   polarRestrictLow = pL;
   polarRestrictHigh = pU;
}

//-----------------------------------------------------------------------------
//
// calculate currents
//
//-----------------------------------------------------------------------------

int PeaceData::calculateCurrents()
{
   if (!densityIsFill && !velocityIsFill)
   {
      for (int i=0; i<3; i++)
      {
         j[i] = (fabs(v[i]) < MAX_VALUE_PM) ? -evjoule*n*v[i] : FILL_VALUE_PM;
         jpara[i] = (fabs(vpara[i]) < MAX_VALUE_PM) ? -evjoule*n*vpara[i] : FILL_VALUE_PM;
         jperp[i] = (fabs(vperp[i]) < MAX_VALUE_PM) ? -evjoule*n*vperp[i] : FILL_VALUE_PM;
      }
      j_mag = (fabs(v_mag) < MAX_VALUE_PM) ? evjoule*n*v_mag : FILL_VALUE_PM;
      jpara_mag = (fabs(vpara_mag) < MAX_VALUE_PM) ? -evjoule*n*vpara_mag : FILL_VALUE_PM;
      jperp_mag = (fabs(vperp_mag) < MAX_VALUE_PM) ? evjoule*n*vperp_mag : FILL_VALUE_PM;
   }
   else
   {
      for (int i=0; i<3; i++)
      {
         j[i] = FILL_VALUE_PM;
         jpara[i] = FILL_VALUE_PM;
         jperp[i] = FILL_VALUE_PM;
      }
      j_mag = FILL_VALUE_PM;
      jpara_mag = FILL_VALUE_PM;
      jperp_mag = FILL_VALUE_PM;
   }
   return _SUCCESS;
}

//-----------------------------------------------------------------------------
//
// calculate magnitude of a vector
//
//-----------------------------------------------------------------------------

double PeaceData::toMagnitude(double v[])
{
   double result;
   if (fabs(v[0]) < MAX_VALUE_PM && fabs(v[1]) < MAX_VALUE_PM && fabs(v[2]) < MAX_VALUE_PM)
   {
      result = sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
   }
   else result = FILL_VALUE_PM;
   return(result);
}

//-----------------------------------------------------------------------------
//    
// return number of flybacks
//       
//-----------------------------------------------------------------------------
      
int PeaceData::getFlyBacksLEEA()
{     
   return(dimLEEA.numEfly);
}     

int PeaceData::getFlyBacksHEEA()
{
   return(dimHEEA.numEfly);
}

//-----------------------------------------------------------------------------
//
// calculate characteristic energy
// 
//-----------------------------------------------------------------------------
   
int PeaceData::calculateCharacteristicEnergy()
{  

   double R[3][3] = {0};
   R[0][0] = 1.0;
   R[1][1] = 1.0;
   R[2][2] = 1.0;

   // generate rotation matrix for rotation from GSE to B-field aligned

   if (currentCoordinateSystem == _GSE)
   {

      // check if we have a B-field or not - we need it!

      if (fgmB[0] == FILL_VALUE_PM && fgmB[1] == FILL_VALUE_PM && fgmB[2] == FILL_VALUE_PM)
      {
         return _MAG_FIELD_REQ_FOR_CE;
      }

      // calculate rotation matrix

      double mag_xyz = sqrt(fgmB[0]*fgmB[0] + fgmB[1]*fgmB[1] + fgmB[2]*fgmB[2]);

      double bxx = fgmB[0]/mag_xyz;
      double byy = fgmB[1]/mag_xyz;
      double bzz = fgmB[2]/mag_xyz;
         
      double mag_yz = sqrt(fgmB[1]*fgmB[1] + fgmB[2]*fgmB[2]);
          
      R[0][0] = mag_yz;
      R[0][1] = -bxx*byy/mag_yz;
      R[0][2] = -bxx*bzz/mag_yz;
      R[1][0] = 0.0;
      R[1][1] = bzz/mag_yz;
      R[1][2] = -byy/mag_yz;
      R[2][0] = bxx;
      R[2][1] = byy;
      R[2][2] = bzz;
   }
   else if (currentCoordinateSystem != _BFA)
   {
      return _INCORRECT_COORD_SYSTEM_FOR_CE;
   }

   // rotate energy flux vector & number flux if necessary

   double QR[3] = {0}, NF[3] = {0};

   
   for (int i=0; i<3; ++i)
   {
      for (int j=0; j<3; ++j)
      {
         QR[i] += R[i][j]*Q[j];
         NF[i] += R[i][j]*v[j]*n;
      }
   }

   // calculate characteristic energy

   if (fabs(NF[2]) > 0.0 && fabs(QR[2]) < MAX_VALUE_PM && fabs(v[2]) < MAX_VALUE_PM && fabs(NF[3]) < MAX_VALUE_PM &&
       n > 0.0)
   {
      ce = QR[2]/NF[2]/evjoule;
   }
   else
   {
      ce = FILL_VALUE_PM;
   }

   return _SUCCESS;
   
}

