/*******************************************************************************
 **
 ** $Id: peacemoments_caa.h 176 2009-04-17 23:39:24Z AndrewLahiff $
 **
 ******************************************************************************
 **
 ** Copyright (C) 2007-2008 Andrew Lahiff.
 **
 ** This file is part of PEACEmoments CAA.
 **
 ** PEACEmoments CAA is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** PEACEmoments CAA is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with PEACEmoments CAA.  If not, see <http://www.gnu.org/licenses/>.
 **
 *******************************************************************************/
 
#ifndef PEACEMOMENTS_CAA_H_
#define PEACEMOMENTS_CAA_H_

using namespace std;

//=============================================================================
//
// headers
//
//=============================================================================

#include <iostream>
#include <fstream>
#include <iomanip>
#include <math.h>
#include <stdio.h>
#include <string>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#include "peacemoments.h"
#include "peacemoments_q.h"
#include "readfile.h"
#include "ret_codes.h"

//=============================================================================
//
// constants
//
//=============================================================================

// version number

#define VERSION_PM "0.7.9"

// constants

#define MAX_VALUE_PM 3.0e+38
#define FILL_VALUE_PM -3.4e+38
#define INT_FILL_VALUE_PM -123456789
#define CLOSE_ENOUGH 1.0/86400.0

// potential offset 2 units

#define _BINS 0
#define _EV 1

// units

#define _PSD 0
#define _DEF 1
#define _DNF 2

// sensor combinations

#define _HEEA 1
#define _LEEA 0
#define _TL1L2B 2
#define _TH1H2B 3
#define _TL1H1B 4
#define _TL2H2B 5
#define _TL1L2 6
#define _TH1H2 7
#define _L1L2 8
#define _H1H2 9
#define _L1L2B 10
#define _H1H2B 11
#define _OVERLAP_TWO_SEC 12

#define _LEEA_HEEA 20
#define _ONE_SENSOR 30
#define _TWO_SENSOR 31

// distribution types

#define _PAD 0
#define _SPINPAD 1
#define _3DR 2
#define _3DF 3
#define _3DXN 4
#define _3DXP 5
#define _3DXE 6

// coordinate systems

#define _SENSOR 0
#define _SPACECRAFT 1
#define _GSE 2
#define _BFA 3
#define _GSM 4
#define _SR2 5
#define _ISR2 6

// required parts from CEF/CDF files

#define _time_tags 0
#define _time_tags_DeltaLower 1
#define _time_tags_DeltaUpper 2
#define _time_tags_HalfInterval 14
#define _Angle_SR2phi 16
#define _Mode_Sensor 3
#define _Mode_SweepMode 4
#define _Mode_RealSize 5
#define _Mode_RealBinNum 6
#define _Mode_DataOrigin 7
#define _Mode_EnergyLevelRange 8
#define _Mode_SPOS 17
#define _Status_Quality 25
#define _Status_CountsStats 18
#define _Status_EFW 19
#define _Status_WHI 20
#define _Status_DataStream 21
#define _Status_Eclipse 48
#define _Energy 9
#define _Energy_DeltaLower 10
#define _Energy_DeltaUpper 11
#define _Azimuth 22
#define _Azimuth_DeltaLower 23
#define _Azimuth_DeltaUpper 24
#define _Data 12
#define _BackgroundLevel 49
#define _Spacecraft_potential 13
#define _MagneticField 15
#define _Data_Overlap_LEEA 26
#define _Data_Overlap_HEEA 27
#define _Mode_SweepMode_LEEA 28
#define _Mode_SweepMode_HEEA 29
#define _Mode_RealSize_Overlap_LEEA 30
#define _Mode_RealSize_Overlap_HEEA 31
#define _Mode_RealBinNum_Overlap_LEEA 32
#define _Mode_RealBinNum_Overlap_HEEA 33
#define _Energy_Overlap_LEEA 34
#define _Energy_DeltaLower_Overlap_LEEA 35
#define _Energy_DeltaUpper_Overlap_LEEA 36
#define _Energy_Overlap_HEEA 37
#define _Energy_DeltaLower_Overlap_HEEA 38
#define _Energy_DeltaUpper_Overlap_HEEA 39
#define _Azimuth_Overlap_LEEA 40
#define _Azimuth_Overlap_LEEA_DeltaLower 41
#define _Azimuth_Overlap_LEEA_DeltaUpper 42
#define _Azimuth_Overlap_HEEA 43
#define _Azimuth_Overlap_HEEA_DeltaLower 44
#define _Azimuth_Overlap_HEEA_DeltaUpper 45
#define _Mode_EnergyLevelRange_LEEA 46
#define _Mode_EnergyLevelRange_HEEA 47

// file types

#define _POTENTIAL_SPIN 100
#define _POTENTIAL_HIGH 101
#define _BFIELD_SPIN 102
#define _BFIELD_HIGH 103
#define _GSE_SPINAXIS_SP 104

// currently doing

#define _CURRENTLY_DOING_MOMENTS 1
#define _CURRENTLY_DOING_MOMENTS_PLOT 2
#define _CURRENTLY_DOING_SPECTROGRAM_PLOT 3

// misc

#define _OFF 0
#define _ON 1
#define _FILE 2

//=============================================================================
//
// structures
//
//=============================================================================

struct dtp
{
   double datetime;
   double potential;
};

struct PeaceDataDim
{
   int numA;
   int numP;
   int numE;
   int numEfly;
};

struct energyBinLimits
{
   int low;
   int high;
};

struct aBinsNum
{
   int numA;
   int numP;
};

struct sWML
{
   bool n;
   bool v;
   bool q;
   bool h;
   bool t;
   bool p;
   bool vpara;
   bool vperp;
   bool tpara;
   bool tperp;
   bool qpara;
   bool qperp;
   bool hpara;
   bool hperp;

   bool j;
   bool jpara;
   bool jperp;

   bool v_mag;
   bool q_mag;
   bool h_mag;

   bool vpara_mag;
   bool vperp_mag;
   bool qpara_mag;
   bool qperp_mag;
   bool hpara_mag;
   bool hperp_mag;

   bool j_mag;
   bool jpara_mag;
   bool jperp_mag;

   bool ce;

   bool b;
   bool scp;
};

struct sPML
{
   bool on;
   bool n;
   bool v;
   bool q;
   bool h;
   bool t;
   bool p;
   bool vpara;
   bool vperp;
   bool tpara;
   bool tperp;
   bool qpara;
   bool qperp;
   bool hpara;
   bool hperp;
 
   bool j;
   bool jpara;
   bool jperp;

   bool v_mag;
   bool q_mag;
   bool h_mag;

   bool vpara_mag;
   bool vperp_mag;
   bool qpara_mag;
   bool qperp_mag;
   bool hpara_mag;
   bool hperp_mag;

   bool j_mag;
   bool jpara_mag;
   bool jperp_mag;

   bool ce;
  
   bool b;
   bool scp;
   
   QString plotTitle;

   double plotHeight;
   double plotWidth;
   
   double plotHeightSpectro;
   double plotWidthSpectro; 

   bool saveToFile;
   QString plotFileName;

   bool includeSpectrogram;
   double spectroEmin;
   double spectroEmax;
   double spectroCmin;
   double spectroCmax;

}; 

struct momCalcParams
{
   bool potentialCorrectionOnOff;
   double potentialOffset1;
   double potentialOffset2;
   int potentialOffset2Units;
   int sensorCombination;
   int coordinateSystem;

   double energyRestrictLow;
   double energyRestrictHigh;
   int energyRestrictType;
   int energyRestrictScpType;
   int energyRestrictStatus;

   double polarRestrictLow;
   double polarRestrictHigh;
   int polarRestrictStatus;

   bool filterEFW;
   bool filterWHI;
   bool filterStats;
   bool filterDataStream;
   bool filterQuality;
};

//=============================================================================
//
// PeaceData class
//
//=============================================================================

class PeaceData
{
   public:
      PeaceData(bool, bool, int, int,
                int, int, int, int, int, int, QString, int, int, int, int, int, int, QString);
      ~PeaceData();
      int setEnergyRegionLimits(int);		// preparation for moments calculation
      int integratePartial(int, int, int, int, int);	// calculate moments for a single region
      int integrate(int, int, int =-1, bool = false);	// calculate moments
      int applyRotation(int);			// rotate to S/C or GSE coordinates
      void convertToRelativeMoments();		// standard to relative moments
      int generateParaPerp();			// generate parallel & perpendicular components
      void adjustUnits();			// convert units
      int currentCoordSystem();			// return current coordinate system
      void setAllMomentsToFills();		// set all moments to fill values
      void setAllVectorsToFills();		// set all vector/tensor moments to fill values
      void setAllParaPerpToFills();		// set all parap/perp moments to fill values
      void setRestrictedEnergyIntegration(int, int, int, double, double);
      void setRestrictedPolarIntegration(int, double, double);
      int calculateCurrents(); 			// calculate currents
      int calculateMagnitudes(); 		// calculate magnitudes of vectors
      int calculateCharacteristicEnergy();      // calculate characteristic energy
      int checkData(int);                       // check for missing bins

      float *dataStoreLEEA;                	// PEACE LEEA data
      float *dataStoreHEEA;              	// PEACE HEEA data

      int energyLevelLowLEEA;			// energy lower level for LEEA
      int energyLevelHighLEEA; 			// energy upper level for LEEA
      int energyLevelLowHEEA;              	// energy lower level for HEEA
      int energyLevelHighHEEA;            	// energy upper level for HEEA

      int *energyUsedLEEA;			// energy bins used for LEEA
      int *energyUsedHEEA;			// energy bins used for HEEA

      float *energyLEEA;                   	// energy lower/upper limits for LEEA bins
      float *energyLowLEEA;                	// energy lower/upper limits for LEEA bins    
      float *energyHighLEEA;              	// energy lower/upper limits for LEEA bins     
      float *energyHEEA;                  	// energy lower/upper limits for HEEA bins
      float *energyLowHEEA;              	// energy lower/upper limits for HEEA bins
      float *energyHighHEEA;                    // energy lower/upper limits for HEEA bins
      
      int sweepModeLEEA;                        // LEEA sweep mode
      int sweepModeHEEA;                        // HEEA sweep mode
      
      double beginDateTimeSpin;                 // begin date/time of spin in decimal date/time
      double endDateTimeSpin;                   // end date/time of spin in decimal date/time
      
      float *potentialLEEA;                     // spacecraft potential for each azimuth for LEEA
      float *potentialHEEA;                     // spacecraft potential for each azimuth for HEEA      
      float potentialSpinAv; 			// spin-averaged potential
 
      double potentialOffset1;			// spacecraft potential offset 1
      double potentialOffset2;			// spacecraft potential offset 2
      int potentialOffset2units;		// spacecraft potential offset 2 unit
      
      double spinAxisLatitudeGSE;               // spin-axis latitude in GSE
      double spinAxisLongitudeGSE;              // spin-axis longitude in GSE
      double fgmB[3];                           // FGM spin-averaged B-field      
      
      double spos;				// rephased sun-pulse offset

      double n, v[3], Q[3], H[3], P[3][3], T;	// moments relative to the bulk flow
      double v_mag, q_mag, h_mag;		// magnitudes

      double vpara_mag, vpara[3];		// velocity parallel to B-field
      double vperp_mag, vperp[3];		// velocity perpendicular to B-field
      double Qpara_mag, Qpara[3];		// energy flux parallel to B-field
      double Qperp_mag, Qperp[3];		// energy flux perpendicular to B-field
      double Hpara_mag, Hpara[3];		// heat flux parallel to B-field
      double Hperp_mag, Hperp[3];		// heat flux perpendicular to B-field
      double Tpara, Tperp, Tperpx, Tperpy;	// parallel & perpendicular temperature
      double Ppara, Pperp, Pperpx, Pperpy;	// parallel & perpendicular pressure
     
      double j[3], jpara[3], jperp[3];		// currents vectors
      double j_mag, jpara_mag, jperp_mag;	// currents magnitudes

      double ce;                                // characteristic energy
      
      double energyLimitLow, energyLimitHigh;	// user-defined lower/upper energy limits
      double polarLimitLow, polarLimitHigh;	// user-defined lower/upper polar limits
      double azimuthLimitLow, azimuthLimitHigh;	// user-defined lower/upper azimuth limits

      int dataOriginLEEA, dataOriginHEEA;   	// data origin for LEEA and HEEA

      int sensorUsedTopRegion;                  // sensor used for top energy region
      int sensorUsedOverlapRegion;              // sensor used for overlap energy region
      int sensorUsedBottomRegion;               // sensor used for bottom energy region

      int padTimingMaster;                      // what sensor is being used for the source of timing
                                                // information

      int getFlyBacksLEEA();  
      int getFlyBacksHEEA();

      double testingIntValue; 

   private:
      bool usingLEEA, usingHEEA;                // which sensors are being used
      struct PeaceDataDim dimLEEA;              // number of bins for LEEA
      struct PeaceDataDim dimHEEA;              // number of bins for HEEA
      int actualArraySizeLEEA;			// actual array size for LEEA
      int actualArraySizeHEEA;			// actual array size for HEEA
      
      QString dataSetIdLEEA;			// LEEA DATASET_ID
      QString dataSetIdHEEA;			// HEEA DATASET_ID

      int distTypeLEEA, distTypeHEEA;		// distribution types for LEEA & HEEA

      int iEnergySummationLEEA;			// energy summation used for LEEA
      int iEnergySummationHEEA;			// energy summation used for HEEA

      struct energyBinLimits eBinLimits[4];	// energy bin limits for different regions

      bool   useT;				// require top region
      bool   useB;				// require bottom region
      bool   useLO;				// require overlap region LEEA
      bool   useHO;				// require overlap region from HEEA
      
      int    senT;				// sensor used for top
      int    senB;				// sensor used for bottom
      int    senLO;				// sensor used for overlap (LEEA)
      int    senHO;				// sensor used for overlap (HEEA)

      double sm_n, sm_nu[3], sm_P[6], sm_Q[3];  // standard moments (total sum over all regions)

      int currentCoordinateSystem;              // current coordinate system of relative moments

      int energyRestrictStatus;                 // restricted energy integration status (on, off, file)
      int energyRestrictScpType;      		// restricted energy integration energy S/C potential type
      int energyRestrictType;                   // restricted energy integration type
      double energyRestrictLow;                 // restricted energy integration range lower limit
      double energyRestrictHigh;                // restricted energy integration range upper limit

      int polarRestrictStatus;                  // restricted polar integration status (on, off)
      double polarRestrictLow;                  // restricted polar integration range lower limit
      double polarRestrictHigh;                 // restricted polar integration range upper limit

      int numPaForMoms;                         // how many PA's to expect from PITCH_* products

      bool densityIsFill, velocityIsFill, pressureIsFill, magFieldIsFill;

      void vectorToTensorIndex(int, int *, int *);	// convert pressure vector to tensor indecies

      double toMagnitude(double []);		// convert vector to magnitude

      bool testingInt;                          // integration testing
};

//=============================================================================
//
// WriteCef class
// 
//=============================================================================

class WriteCef
{
   public:
      WriteCef(int);
      int open(QString);
      void close();
      int beginData();
      int endData();
      int endCurrentDataRecord();
      int writeDataVariable(int, double *);
      int writeDataVariable(int, float *);
      int writeDataVariable(int, int *);
      int writeDataVariable(QString);
      int writeMetaVersionNumber(int);
      int writeMetaDatasetVersion(void);
      int writeMetaFileTimeSpan(QString, QString);
      int writeMetaLogicalFileId(QString);
      int writeMetaSetupPotential(double, double, int, bool);
      int writeMetaSetupRestrInt(int, int, int, double, double, int, double, double);
      int writeMetaFileUsed(QString, QString);
      int writeFileName(QString);
      int writeInitial(void);
      int writeMetaStatic(void);
      int writeGenerationDate(void);
      int writeTimeTagsMeta();
      int writeDensityMeta();
      int writeVelocityMeta();
      int writeVelocityParaMeta();
      int writeVelocityPerpMeta();
      int writeTempMeta();
      int writeTempParaMeta();
      int writeTempPerpMeta();
      int writeHeatFluxMeta();
      int writeHeatFluxParaMeta();
      int writeHeatFluxPerpMeta();
      int writeEnergyFluxMeta();
      int writeEnergyFluxParaMeta();
      int writeEnergyFluxPerpMeta();
      int writePressureMeta();
      int writeMagFieldMeta();
      int writePotentialMeta();
      int writeUsefulInfoHeader();
      int writeCurrentMeta();
      int writeCurrentParaMeta();
      int writeCurrentPerpMeta();
      int writeCurrentMagMeta();
      int writeCurrentParaMagMeta();
      int writeCurrentPerpMagMeta();
      int writeVelocityMagMeta();
      int writeVelocityParaMagMeta();
      int writeVelocityPerpMagMeta();
      int writeEnergyFluxMagMeta();
      int writeEnergyFluxParaMagMeta();
      int writeEnergyFluxPerpMagMeta();
      int writeHeatFluxMagMeta();
      int writeHeatFluxMagParaMeta();
      int writeHeatFluxMagPerpMeta();
      int writeCharacteristicEnergyMeta();
      int writeMetaMomentsSetup(QString, QString, int);
      int writeObservatory(QString);
      int writeExperiment();
      int writeMetaStatusVariables();

   private:
      QString fileName, spacing, fieldNameComp;
      char str_F[50], str_CS[50], str_C[50];
      QFile file;
      QTextStream out;
};

//=============================================================================
//
// PeacemomentsParser class
//
//=============================================================================

class PeacemomentsParser : public QXmlDefaultHandler
{
   public:
      PeacemomentsParser();
      bool startDocument();
      bool endElement(QString &, const QString &, const QString &);
      bool startElement(const QString &, const QString &, const QString &,
                        const QXmlAttributes &);
			
      QString filePeace1, filePeace2, filePotential, fileAuxGSE, fileFGM;
      QString setupSensor, setupCoord;
      QString potOffset1, potOffset2;
      int potOffset2units;
      int filterQua, filterWHI, filterEFW, filterSta, filterDS;
      QString outputDir, outputFile;
      int outInclN, outInclV, outInclQ, outInclH, outInclP, outInclT,
          outInclVpara, outInclVperp, outInclQpara, outInclQperp,
	  outInclHpara, outInclHperp, outInclTpara, outInclTperp,
          outInclJ, outInclJpara, outInclJperp,
          outInclScP, outInclB, outInclCE;
      int onoff;

      QString int_en_low, int_en_high;
      int int_en_state, int_en_type, int_en_scp_type;
      QString int_pl_low, int_pl_high;
      int int_pl_state;

      int dataD_SCN, dataD_PEACE1, dataD_PEACE2, dataD_POT, dataD_AUX, dataD_FGM;
      QString timeBegin, timeEnd;
      int yearBegin, yearEnd, doyBegin, doyEnd, dayBegin, dayEnd, monthBegin, monthEnd;
			
   private:
   			
};

//=============================================================================
//
// functions
//
//=============================================================================

double T_PM(double, double, int);
double VN_PM(double, double, int);
int maxDaysInYear(int);
double timeToDec(int, int, int);
void gse2bfa(double, double, double, double [3][3]);
int doytoday(int, int, int);
int daytodoy(int, int, int);
int date_extr(char [], int);
double qTimeToDec(QString);
double time_to_tis(char []);
void getPeaceLocations(ReadFile *, int []);
void getMagFieldLocations(ReadFile *, int []);
void getPotentialLocations(ReadFile *, int []);
void getAuxGSELocations(ReadFile *, int []);
int checkForFileType(QString);
int returnSpacecraftNumber(QString);
bool timeDateOverlapCheck(QString, QString, QString, QString);
int putPeaceDataIntoLEEA(ReadFile *, PeaceData *, int *, bool);
int putPeaceDataIntoHEEA(ReadFile *, PeaceData *, int *, bool);
int putPeaceDataIntoLEEAo(ReadFile *, PeaceData *, int *);
int putPeaceDataIntoHEEAo(ReadFile *, PeaceData *, int *);
void energyLevels(int, int, int *, int *, int *, int);
int getSensorCombination(char []);
int getCoordinateSystem(char []);
void sc2gse(double, double, double [3][3], double);
void getSpinStartEnd(QString, double, double, double *, double *);
int statusHandler(int, char *);
void setAzimuthTimes(double, double, int, double *);
void generateMergedAzimuthList(double *, double *, int, int, double *, int *);
int downloadDataPrep(QString, QString, QString, QString, QString, QString,
                     QString, QString, QString, QString *);                   
QString sensorCombinationToStr(int);
QDate toQDate(QString);
void coordinateSystemToStr(int, char *, char *, char *);
double interpolateLinear(double, double, double, double, double);
int decompress(const QString &, const QString &);
int extract_files(QString, QStringList *);
int quickCheckFileType(char *);

void generateDateTime(QString, float, float, int, float, float, float, QString *, float *, float *,
                      int);

#endif /*PEACEMOMENTS_CAA_H_*/
