/*******************************************************************************
 **
 ** $Id: time.C 149 2008-11-28 20:51:29Z AndrewLahiff $
 **
 ******************************************************************************
 **
 ** Copyright (C) 2007-2008 Andrew Lahiff.
 **
 ** This file is part of PEACEmoments CAA.
 **
 ** PEACEmoments CAA is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** PEACEmoments CAA is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with PEACEmoments CAA.  If not, see <http://www.gnu.org/licenses/>.
 **
 *******************************************************************************/

#include "peacemoments_caa.h"

//-----------------------------------------------------------------------------
//
// return the number of days in a year
//
//-----------------------------------------------------------------------------

int maxDaysInYear(int year)
{
   if ((year % 4) == 0)
   {
      if (((year % 100) == 0) && ((year % 400) != 0))
      {
         return(365);
      }
      else
      {  
         return(366);
      }
   }
   else
   {  
      return(365);
   }  
}

//-----------------------------------------------------------------------
//
// convert time in year, days, milliseconds to decimal time since 2000
//
//-----------------------------------------------------------------------

double timeToDec(int year, int doy, int time)
{
   double mySum = 0;   
   for (int i = 2000; i < year; i++)
   {
      mySum = mySum + (double)maxDaysInYear(i);
   }
   mySum = mySum + doy + (double)time/86400000.0;
   return(mySum);
}

//-----------------------------------------------------------------------
//
// convert ISO date/time as a QString to decimal time
//
//-----------------------------------------------------------------------

double qTimeToDec(QString in)
{
   QRegExp tdRx("\\d\\d\\d\\d-\\d\\d-\\d\\dT\\d\\d:\\d\\d:\\d\\d.\\d\\d\\d");

   if (tdRx.indexIn(in) == 0)
   {
      QString date = in.mid(0, 10);
      QString time = in.mid(11, 12);

      int YY = date_extr((char *)(date.toStdString().c_str()), 0);
      int MM = date_extr((char *)(date.toStdString().c_str()), 1);
      int DD = date_extr((char *)(date.toStdString().c_str()), 2);
      int doy = daytodoy(YY, MM, DD);

      int tims = (int)(1000.0*time_to_tis((char *)(time.toStdString()).c_str()) + 0.5);

      return(timeToDec(YY, doy, tims));
   }
   else
   {
      return(_FAILED);
   }
}

//-----------------------------------------------------------------------
//
// extract year, month or day from a date string
//
//-----------------------------------------------------------------------

int date_extr(char date[], int iwhich)
{
   int selected, YY, MM, DD;
   
   string dateStr = date;
   YY = atoi(((dateStr.substr(0, 4)).c_str()));
   MM = atoi(((dateStr.substr(5, 2)).c_str()));
   DD = atoi(((dateStr.substr(8, 2)).c_str()));

   if (iwhich == 1)
   {
      selected = MM;
   }
   else if (iwhich == 2)
   {
      selected = DD;
   }
   else
   {
      selected = YY;
   }
   
   return(selected);
}

//-----------------------------------------------------------------------
//
// convert time string to time in seconds
//
//-----------------------------------------------------------------------

double time_to_tis(char timein[])
{
   double result, d_hr, d_mn, d_sc;
   string timeStr = timein;

   d_hr = atof(((timeStr.substr(0, 2)).c_str()));
   d_mn = atof(((timeStr.substr(3, 2)).c_str()));
   d_sc = atof(((timeStr.substr(6, 6)).c_str()));

   result = d_sc + d_mn*60.0 + d_hr*3600.0;

   return(result);
}

//-----------------------------------------------------------------------------
//
// check if two time intervals overlap or not
//
//-----------------------------------------------------------------------------

bool timeDateOverlapCheck(QString start1, QString end1, QString start2, QString end2)
{
   double start1Dec = qTimeToDec(start1);
   double start2Dec = qTimeToDec(start2);
   double end1Dec = qTimeToDec(end1);
   double end2Dec = qTimeToDec(end2);

   bool returnValue = false;

   if ((start1Dec >= start2Dec && start1Dec < end2Dec) ||
       (start2Dec >= start1Dec && start2Dec < end1Dec))
   {
      returnValue = true;
   }
   return(true); // temporary until FILE_TIME_SPAN is sorted out by the CAA team
   return(returnValue);
}

//-----------------------------------------------------------------------------
//
// caclulate PEACE spin start & end times (in decimal time)
//
//-----------------------------------------------------------------------------

void getSpinStartEnd(QString timeDatePeace, double time_tags_DeltaLower, double time_tags_DeltaUpper,
		     double *spinStart, double *spinEnd)
{
   double centreTime = qTimeToDec(timeDatePeace);
   if (centreTime != _FAILED)
   {
      *spinStart = centreTime - time_tags_DeltaLower/86400.0;
      *spinEnd = centreTime + time_tags_DeltaUpper/86400.0;
   }
   else
   {
      *spinStart = _FAILED;
      *spinEnd = _FAILED;
   }
}

//-----------------------------------------------------------------------------
//
// convert QString date in YYYY-MM-DD format to a QDate
// 
//-----------------------------------------------------------------------------

QDate toQDate(QString in)
{
    int beginYr = in.mid(0, 4).toInt();
    int beginMn = in.mid(5, 2).toInt();
    int beginDy = in.mid(8, 2).toInt();
    return(QDate(beginYr, beginMn, beginDy));
}

//-----------------------------------------------------------------------
//
// generate list of decimal times of individual sweeps (azimuths)
//
//-----------------------------------------------------------------------

void setAzimuthTimes(double timeBegin, double timeEnd, int numBins, double *array)
{
   for (int i=0; i<numBins; i++)
   {
      *(array+i) = timeBegin + (double)i*(timeEnd - timeBegin)/((double)numBins);
   }
   *(array+numBins) = timeEnd;
}

//-----------------------------------------------------------------------
//
// generate merged list of azimuth times
//
//-----------------------------------------------------------------------

void generateMergedAzimuthList(double *array1, double *array2, int bins1, int bins2,
                               double *array3, int *arrayList)
{
   int usedA[128];
   double arrayA[128];

// if only one sensor used, just return the first list

   if (bins2 == 0)
   {
      for (int i=0; i<bins1; i++)
      {
         *(array3+i) = *(array1+i);
         *(arrayList+i) = 1;
      }
      return;
   }

// firstly generate unsorted merged list

   int j=0;
   for (int i=0; i<bins1; i++)
   {
      arrayA[j] = *(array1+i);
      usedA[j] = 1;
      j++;
   }
   for (int i=0; i<bins2; i++)
   {
      arrayA[j] = *(array2+i);
      usedA[j] = 2;
      j++;
   }

// now generated time-ordered merged list

   for (int i=0; i<j; i++)
   {
      double min = 2.5e+30;
      int kk = 0;
      for (int k=0; k<j; k++) 
      {
         if (arrayA[k] < min)
         {
            min = arrayA[k];
            kk = k;
         }
      }
      *(array3+i) = arrayA[kk];
      *(arrayList+i) = usedA[kk];
      arrayA[kk] = 2.8e+30;
   }
}

//-----------------------------------------------------------------------------
//
// generate timestamps for the PITCH_FULL product
//
//-----------------------------------------------------------------------------

// convert time-in-seconds to string in HH:MM:SS.ZZZ format

void MilliToStr (char str[], double secs)
{
    float V;
    short H, M, S;
    int milliseconds = (int)(1000.0*secs+0.5); // convert secs to millisecs

    V = milliseconds / 3600000;
    H = (short) V;
    milliseconds -= 3600000 * H;
    V = milliseconds / 60000;
    M = (short) V;
    milliseconds -= 60000 * M;
    V = milliseconds / 1000;
    S = (short) V;
    milliseconds -= 1000 * S;
    sprintf (str, "%02d:%02d:%02d.%03d", H, M, S, milliseconds);
}

// convert SR2 azimuthal angle to timestamp (time in milliseconds)

double procAzi(double tis, double in, int spos, double spinDuration, int sensor)
{
   double SUNAZ = 26.2 - ((double)spos)*360.0/(16.0*1024.0);
   in += SUNAZ + 30.0;
   if (sensor == 0) in += 180.0;
   if (in > 360.0) in -= 360.0;
   if (in < 0.0) in += 360.0;
   in = in*spinDuration/360.0 + tis;
   return(in);
}

// generate timestamp along with lower/upper deltas from azimuthal info
// for PITCH_FULL moments

void generateDateTime(QString in, float inLow, float inUpp, int spos,
                      float aziCentre, float aziLow, float aziUpp,
                      QString *out, float *outLow, float *outUpp,
                      int sensor)
{

   // extract time-in-seconds from HH:MM:SS.ZZZ string

   QString time = in.mid(11, 12);
   double tis = time_to_tis((char *)(time.toStdString()).c_str());

   // convert azimuthal angle to time

   double aziProcBegin = procAzi(tis - inLow, aziCentre - aziLow, spos, inLow + inUpp, sensor);
   double aziProcEnd   = procAzi(tis - inLow, aziCentre + aziUpp, spos, inLow + inUpp, sensor);
   double aziProcCentre= (aziProcEnd + aziProcBegin)/2.0;

   // flyback correction

   double flyTime = (aziProcEnd - aziProcBegin)/15.0;
   aziProcCentre += flyTime;

/*
     aziProcBegin += flyTime;
     aziProcEnd += flyTime;
     char a[15], b[15];
     MilliToStr(a, aziProcBegin);
     MilliToStr(b, aziProcEnd);
     cout << "DATA_RECORD " << a << " " << b << endl;
*/

   // generate the lower/upper deltas

   *outLow = aziProcCentre - aziProcBegin;
   *outUpp = aziProcEnd - aziProcCentre;

   // finally, generate the centre time in HH:MM:SS.ZZZ format

   char timeNew[15];
   MilliToStr(timeNew, aziProcCentre);
//   cout << "     azi = " << aziCentre << " : " << timeNew << " spos = " << spos << endl;
   *out = QString("%1T%2Z").arg(in.mid(0, 10)).arg(timeNew);
}
