/*******************************************************************************
 **
 ** $Id: window.C 166 2008-11-30 09:42:37Z AndrewLahiff $
 **
 ******************************************************************************
 **
 ** Copyright (C) 2007-2008 Andrew Lahiff.
 **
 ** This file is part of PEACEmoments CAA.
 **
 ** PEACEmoments CAA is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** PEACEmoments CAA is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with PEACEmoments CAA.  If not, see <http://www.gnu.org/licenses/>.
 **
 *******************************************************************************/

#include "peacemoments_caa.h"
#include "peacemoments_gui.h"
#include "window.h"
#include "thread.h"

#define FILE_WIDTH 400        // width of input file QLineEdit
#define SELECTOR_WIDTH_1 100  // width of calculate etc buttons
#define SELECTOR_WIDTH_2 80
#define SELECTOR_WIDTH_3 120
#define SELECTOR_WIDTH_4 95
#define POTENTIAL_WIDTH 40
#define FILE_BUTTON_WIDTH 30 

//=============================================================================
//
// Window class - the main GUI
//
//=============================================================================

Window::Window(char fileToLoad[])
{ 

//   setPalette(QPalette(QColor("#c0c0c0")));
//   setAutoFillBackground(true);

//-----------------------------------------------------------------------------
//
// setup the tabbed pages
//         
//-----------------------------------------------------------------------------

   tabWidget = new QTabWidget;
   filesTabPage = new FilesTab();
   dataTabPage = new DataTab();
   
   tabWidget->addTab(filesTabPage, "Input files");
   tabWidget->addTab(dataTabPage, "Download data");
   
   tabWidget->setTabToolTip(0, "Specify input CDF or CEF files");
   tabWidget->setTabToolTip(1, "Download all required data from the CAA");

   dataTabPage->setEnabled(true);

//-----------------------------------------------------------------------------
// 
// spacecraft potential correction
//         
//-----------------------------------------------------------------------------

   potentialGroup = new QGroupBox("Spacecraft potential correction / photoelectron removal");
   potentialGroup->setCheckable(true);
   potentialGroup->setChecked(true);
   potentialGroup->setToolTip("Spacecraft potential correction on/off");

   potentialOffset1 = new QLineEdit();
   potentialOffset1->setAlignment(Qt::AlignRight);
   potentialOffset1->insert("1.0");
   potentialOffset1->setFixedWidth(POTENTIAL_WIDTH);
   potentialOffset1->setToolTip("The difference between the EFW probe potential and the spacecraft potential; usually 1-2 eV");
   
   QLabel* potentialOffset1Label = new QLabel("Offset");
   potentialOffset1Label->setBuddy(potentialOffset1);
   potentialOffset1Label->setToolTip("The difference between the EFW probe potential and the spacecraft potential; usually 1-2 eV");

   QLabel *potentialOffset1Units = new QLabel("eV");
   potentialOffset1Units->setBuddy(potentialOffset1);

   potentialOffset2 = new QLineEdit();
   potentialOffset2->setAlignment(Qt::AlignRight);
   potentialOffset2->insert("0");
   potentialOffset2->setFixedWidth(POTENTIAL_WIDTH);
   potentialOffset2->setToolTip("Increase the low energy cutoff of the energy integration relative to the spacecraft potential");

   QLabel* potentialOffset2Label = new QLabel("Low energy cutoff offset");
   potentialOffset2Label->setBuddy(potentialOffset2);
   potentialOffset2Label->setToolTip("Increases the low energy cutoff of the energy integration relative to the spacecraft potential");

   potentialOffset2Units = new QComboBox;
   potentialOffset2Units->addItem("bins");
   potentialOffset2Units->addItem("eV");
   potentialOffset2Units->setToolTip("Select units for the low energy cutoff offset");
   potentialOffset2Units->setFixedWidth(60);

   QHBoxLayout *potBoxLayout = new QHBoxLayout;
   potBoxLayout->addWidget(potentialOffset1Label);
   potBoxLayout->addWidget(potentialOffset1);
   potBoxLayout->addWidget(potentialOffset1Units);
   potBoxLayout->addSpacing(7);
   potBoxLayout->addWidget(potentialOffset2Label);
   potBoxLayout->addWidget(potentialOffset2);
   potBoxLayout->addWidget(potentialOffset2Units);
   potentialGroup->setLayout(potBoxLayout);

//-----------------------------------------------------------------------------
//
// "Files" group (Calculate, Spectrogram, Load, Save, Quit)
//
//-----------------------------------------------------------------------------

   QGroupBox *fileGroup = new QGroupBox();
   
   generateMomentsAction = new QPushButton("&Calculate");
   generateMomentsAction->setAutoDefault(true);
   generateMomentsAction->setToolTip("Calculate moments");
   generateMomentsAction->setFixedWidth(SELECTOR_WIDTH_1);

   connect(generateMomentsAction, SIGNAL(clicked()),
           this, SLOT(generateMoments()));

   integrationAction = new QPushButton("&Integration");
   integrationAction->setToolTip("Integration settings");
   integrationAction->setFixedWidth(SELECTOR_WIDTH_1);
   
   connect(integrationAction, SIGNAL(clicked()),
           this, SLOT(displayIntegration()));

   outputAction = new QPushButton("&Output");
   outputAction->setToolTip("Output settings");
   outputAction->setFixedWidth(SELECTOR_WIDTH_1);

   connect(outputAction, SIGNAL(clicked()),
           this, SLOT(displayOutput()));

   spectroViewerAction = new QPushButton("&Plots");
   spectroViewerAction->setAutoDefault(false);
   spectroViewerAction->setFixedWidth(SELECTOR_WIDTH_1);
   spectroViewerAction->setToolTip("Plot settings");

   connect(spectroViewerAction, SIGNAL(clicked()),
           this, SLOT(displayPlot()));

   loadCPMAction = new QPushButton("&Load");
   loadCPMAction->setAutoDefault(false);
   loadCPMAction->setToolTip("Load settings from a file");
   loadCPMAction->setFixedWidth(SELECTOR_WIDTH_2);

   connect(loadCPMAction, SIGNAL(clicked()),
           this, SLOT(loadCPM()));

   saveCPMAction = new QPushButton("&Save");
   saveCPMAction->setAutoDefault(false);
   saveCPMAction->setToolTip("Save settings to a file");
   saveCPMAction->setFixedWidth(SELECTOR_WIDTH_2);

   connect(saveCPMAction, SIGNAL(clicked()),
           this, SLOT(saveCPM()));

   quitAction = new QPushButton("&Quit");
   quitAction->setAutoDefault(false);
   quitAction->setToolTip("Quit");
   quitAction->setFixedWidth(SELECTOR_WIDTH_2);

   connect(quitAction, SIGNAL(clicked()), qApp, SLOT(quit()));

   QHBoxLayout *fileLayout = new QHBoxLayout;
   fileLayout->addWidget(generateMomentsAction);
   fileLayout->addStretch(1);
   fileLayout->addWidget(integrationAction);
   fileLayout->addStretch(1);
   fileLayout->addWidget(outputAction);
   fileLayout->addStretch(1);
   fileLayout->addWidget(spectroViewerAction);
   fileLayout->addStretch(1);
   fileLayout->addWidget(loadCPMAction); 
   fileLayout->addStretch(1);
   fileLayout->addWidget(saveCPMAction);
   fileLayout->addStretch(1);
   fileLayout->addWidget(quitAction);
   
   fileGroup->setLayout(fileLayout);

//-----------------------------------------------------------------------------
//
// sensor combination, coordinate system
//
//-----------------------------------------------------------------------------

   QGroupBox *setupGroup = new QGroupBox("Sensor setup / coordinate system");

   sensorCombiSelect = new QComboBox;
   sensorCombiSelect->setMaxCount(11);
   sensorCombiSelect->addItem("LEEA");
   sensorCombiSelect->addItem("HEEA");
   sensorCombiSelect->addItem("TH1H2B");
   sensorCombiSelect->addItem("TL1L2B");
   sensorCombiSelect->addItem("L1L2B"); 
   sensorCombiSelect->addItem("H1H2B");
   sensorCombiSelect->addItem("TL1L2");
   sensorCombiSelect->addItem("TH1H2");
   sensorCombiSelect->addItem("H1H2");
   sensorCombiSelect->addItem("L1L2");
   sensorCombiSelect->addItem("Overlap 2-sec");
   sensorCombiSelect->setToolTip("Select the required sensor combination");
   sensorCombiSelect->setFixedWidth(SELECTOR_WIDTH_3);

   coordSelect = new QComboBox;
   coordSelect->setMaxCount(6);
   coordSelect->addItem("GSE");
   coordSelect->addItem("Spacecraft");
   coordSelect->addItem("Sensor");
   coordSelect->addItem("SR2");
   coordSelect->addItem("ISR2"); 
   coordSelect->addItem("B-field aligned");
   coordSelect->setToolTip("Select the required coordinate system");
   coordSelect->setFixedWidth(SELECTOR_WIDTH_3);
 
   QHBoxLayout *setupLayout = new QHBoxLayout;
   setupLayout->addWidget(sensorCombiSelect);
   setupLayout->addStretch(1);
   setupLayout->addWidget(coordSelect);
   setupGroup->setLayout(setupLayout);

//-----------------------------------------------------------------------------
//
// combine all the components to make the main GUI
//
//-----------------------------------------------------------------------------

   QWidget *w = new QWidget;
   setCentralWidget(w);
   setWindowTitle(QString("PEACEmoments CAA %1").arg(VERSION_PM));

   QHBoxLayout *layoutH1 = new QHBoxLayout;
   layoutH1->addWidget(potentialGroup);
   layoutH1->addStretch(1);
   layoutH1->addWidget(setupGroup);

   QVBoxLayout *layout = new QVBoxLayout;
   layout->addWidget(tabWidget);
   layout->addLayout(layoutH1);
   layout->addWidget(fileGroup);
   w->setLayout(layout);

   statusBar()->showMessage("");

   progressBar = new QProgressBar;
   progressBar->setFixedWidth(250);   
   progressBar->setFixedHeight(statusBar()->height()/2);

   statusBar()->addPermanentWidget(progressBar);

//-----------------------------------------------------------------------------
//
// About window - for Mac only
//
//-----------------------------------------------------------------------------

#ifdef Q_WS_MAC
   aboutAct = new QAction(tr("&About"), this);
   connect(aboutAct, SIGNAL(triggered()), this, SLOT(about()));
   helpMenu = menuBar()->addMenu(tr("&Help"));
   helpMenu->addAction(aboutAct);
#endif

//-----------------------------------------------------------------------------
//
// initialize input files
//
//-----------------------------------------------------------------------------

   userFilePeace1 = new QString("");
   userFilePeace2 = new QString("");
   userFilePotential = new QString("");
   userFileAuxGSE = new QString("");
   userFileMagField = new QString("");
   userFileOut = new QString("moments.cef");
   userDirOut = new QString("");

//-----------------------------------------------------------------------------
//
// initialize content of moments output files & moments plots
// 
//-----------------------------------------------------------------------------

   myWML = new struct sWML; // content of moments output files
   myPML = new struct sPML; // content of moments plots

// default moments generation

   myWML->n = true;
   myWML->v = true;
   myWML->t = true;
   myWML->q = false;
   myWML->p = false;
   myWML->h = false;
   myWML->vpara = false;
   myWML->vperp = false;
   myWML->qpara = false;
   myWML->qperp = false;
   myWML->hpara = false;
   myWML->hperp = false;
   myWML->tpara = false;
   myWML->tperp = false;
   
   myWML->j = false;
   myWML->jpara = false;
   myWML->jperp = false;
   
   myWML->b = false;

// default plotting

   myPML->on = true;
   myPML->n = true;
   myPML->v = true;
   myPML->t = true;
   myPML->q = false;
   myPML->p = false;
   myPML->h = false;
   myPML->vpara = false;
   myPML->vperp = false;
   myPML->qpara = false;
   myPML->qperp = false;
   myPML->hpara = false;
   myPML->hperp = false;
   myPML->tpara = false;
   myPML->tperp = false;

   myPML->plotWidth = 680;
   myPML->plotHeight = 700;

   myPML->plotTitle = "Moments Plot";

   myPML->saveToFile = false;
   myPML->plotFileName = "moments.png";

   myPML->includeSpectrogram = true;
   myPML->spectroEmin = 1.0;
   myPML->spectroEmax = 30000.0;
   myPML->spectroCmin = 1.0e-06;
   myPML->spectroCmax = 1.0e-03;
   
   myPML->plotWidthSpectro = 800;
   myPML->plotHeightSpectro = 400;

//-----------------------------------------------------------------------------
//
// storage of moments calculation parameters
//
//-----------------------------------------------------------------------------

   myParams = new struct momCalcParams;

   myParams->potentialOffset1 = 1.0;
   myParams->potentialOffset2 = 0.0;
   myParams->potentialOffset2Units = 0;

//-----------------------------------------------------------------------------
//
// moments calculation thread
//
//-----------------------------------------------------------------------------

   qRegisterMetaType<QImage>("QImage");

   calculationThread = new calculateMomentsThread();

// respond to signals from the calculation/plotting thread

   connect(calculationThread, SIGNAL(returnedMessage(int, QString, int)),
           this, SLOT(displayMessage(int, QString, int)));

   connect(calculationThread, SIGNAL(finished()),
           this, SLOT(slotUpdateStatusBar()));

   connect(calculationThread, SIGNAL(started()),
           this, SLOT(slotUpdateStatusBar()));

   connect(calculationThread, SIGNAL(currentlyDoing(int)),
           this, SLOT(slotUpdateStatusBarProgress(int)));

   connect(calculationThread, SIGNAL(thePlot(const QImage &)),
           this, SLOT(displayMomentsPlotImage(const QImage &)));
	   
   connect(calculationThread->peaceMoments, SIGNAL(progress(int)),
           progressBar, SLOT(setValue(int)));
   
//-----------------------------------------------------------------------------
//
// spectrogram generation thread
//
//-----------------------------------------------------------------------------
                  
   spectrogramThread = new generateSpectrogramThread();               
                     
   connect(spectrogramThread, SIGNAL(thePlot(const QImage &)),
           this, SLOT(displaySpectrogramPlotImage(const QImage &)));         

   connect(spectrogramThread, SIGNAL(started()),
           this, SLOT(slotUpdateStatusBarSpectro()));

   connect(spectrogramThread, SIGNAL(returnedStatus(int)),
           this, SLOT(displayMessageSpectro(int)));
           
//-----------------------------------------------------------------------------
//
// other windows
//
//-----------------------------------------------------------------------------

   integrationWindow = new IntegrationWindow();  // integration settings window
   outputWindow = new OutputWindow();            // output settings window
   plotWindow = new PlotWindow();                // plot settings window
   
   momentsWindow = new PlotMoments((int)myPML->plotWidth, (int)myPML->plotHeight);
   spectrogramWindow = new PlotSpectrogram((int)myPML->plotWidthSpectro, (int)myPML->plotHeightSpectro);

   outputWindow->outputFileEdit->setText(*userFileOut);

// set style sheet for the windows

   int uFS = defaultFontSize - 2;
   QString style1 = "";
   style1 = style1 + QString("QPushButton {font-size: %1pt; background-color: #b0b0b0; color: #000000; border-width: 1px; border-color: #909090; border-style: outset; border-radius: 2px; padding: 2px}").arg(uFS);
   style1 = style1 + QString("QPushButton:pressed {background-color: #a0a0a0}");
   style1 = style1 + QString("QPushButton:hover {background-color: #d0d0d0}");
   style1 = style1 + QString("QCheckBox {font-size: %1pt;}").arg(uFS);
   style1 = style1 + QString("QGroupBox {font-size: %1pt; background-color: #c0c0c0;}").arg(uFS);
   style1 = style1 + QString("QLineEdit {font-size: %1pt;}").arg(uFS);
   style1 = style1 + QString("QLabel {font-size: %1pt;}").arg(uFS);
   style1 = style1 + QString("QComboBox {font-size: %1pt; padding: 2px; border-radius: 2px; background-color: #f0f0f0; selection-background-color: #3333ff}").arg(uFS);
   
//   outputWindow->setStyleSheet(style1);
//   plotWindow->setStyleSheet(style1);
//   integrationWindow->setStyleSheet(style1);

//-----------------------------------------------------------------------------
//
// other signals
//
//-----------------------------------------------------------------------------

// redirection signals from plots

   connect(plotWindow, SIGNAL(remakeMomentsPlot(QString)),
           this, SLOT(generateMomentsPlotOnly(QString)));
           
   connect(plotWindow, SIGNAL(makeSpectrogramPlot(QString)),
           this, SLOT(makeSpectrogram(QString)));

// HTTP downloads

   connect(dataTabPage->http, SIGNAL(requestFinished(int, bool)),
           this, SLOT(httpRequestFinished(int, bool)));

   connect(dataTabPage->http, SIGNAL(requestStarted(int)),
           this, SLOT(httpRequestStarted(int)));

   connect(dataTabPage->http, SIGNAL(done(bool)),
           this, SLOT(downloadHandler(bool)));

   connect(dataTabPage, SIGNAL(downloadAborted()),
           this, SLOT(downloadAborted()));

// update output directory storage for datatab page

   connect(outputWindow->outputDirectoryEdit, SIGNAL(textChanged(QString)),
           this, SLOT(slotModifiedOutputDir(QString)));

//----------------------------------------------------------------------------
//
// load user default settings if possible
//
//---------------------------------------------------------------------------

   loadSettings();

//-----------------------------------------------------------------------------
//
// load file if one was specified on the command line
//
//-----------------------------------------------------------------------------

   QString QfileToLoad = QString(fileToLoad);
   if (!QfileToLoad.isEmpty())
   {
      loadCPMonly(QfileToLoad);
   }
   
}

//-----------------------------------------------------------------------------
//
// destructor
//
//-----------------------------------------------------------------------------

Window::~Window()
{
/*   delete calculateMomentsThread;
   delete generateSpectrogramThread;
   delete integrationWindow;
   delete outputWindow;
   delete plotWindow;
   delete momentsWindow;
   delete spectrogramWindow;*/
}

//-----------------------------------------------------------------------------
//    
// Mac about window
// 
//-----------------------------------------------------------------------------
   
void Window::about()
{  
   QMessageBox::about(this, tr("PEACEmoments CAA"),
                      tr("<b>PEACEmoments CAA</b><br>"
                      "<i>- Cluster PEACE Electron Moments from CAA Data</i><br><br>"
                      "by A. D. Lahiff<br><br>"
                      "Version %1<br><br>"
                      "www.mssl.ucl.ac.uk/missions/CAA/software/peacemoments").arg(VERSION_PM));
}

//------------------------------------------------------------------------------
//
// load XML configuration file - get file from GUI first
//
//------------------------------------------------------------------------------

void Window::loadCPM()
{

// get filename from user

   QString loadFileText = QFileDialog::getOpenFileName(this, "", "", "PEACEmoments files (*.xml)");
   
   if (!loadFileText.isEmpty()) loadCPMonly(loadFileText);

}

//------------------------------------------------------------------------------
//
// load XML configuration file - actual
//
//------------------------------------------------------------------------------

void Window::loadCPMonly(QString loadFileText)
{

// continue if we have a filename from the user

   if (!loadFileText.isEmpty())
   {
      QFile fileLoad(loadFileText);
      if (!fileLoad.open(QIODevice::ReadOnly)) return;
      QXmlInputSource source(&fileLoad);
      QXmlSimpleReader reader;
      
// parse the XML file

      PeacemomentsParser handler;
      reader.setContentHandler(&handler);
      reader.setErrorHandler(&handler);
      reader.parse(source);
      
      fileLoad.close();

// set the GUI

// - files
      filesTabPage->editFilePeace1->setText(handler.filePeace1);
      filesTabPage->editFilePeace2->setText(handler.filePeace2);  
      filesTabPage->editFilePotential->setText(handler.filePotential);
      filesTabPage->editFileAuxGSE->setText(handler.fileAuxGSE);            
      filesTabPage->editFileMagField->setText(handler.fileFGM);

// - output settings      
      outputWindow->outputDirectoryEdit->setText(handler.outputDir);
      outputWindow->outputFileEdit->setText(handler.outputFile);
      
      outputWindow->filterQUA->setChecked(handler.filterQua);
      outputWindow->filterWHI->setChecked(handler.filterWHI);
      outputWindow->filterEFW->setChecked(handler.filterEFW);
      outputWindow->filterStats->setChecked(handler.filterSta);
      outputWindow->filterDataStream->setChecked(handler.filterDS);
      
      outputWindow->inclOutN->setChecked(handler.outInclN);
      outputWindow->inclOutV->setChecked(handler.outInclV);
      outputWindow->inclOutQ->setChecked(handler.outInclQ);
      outputWindow->inclOutH->setChecked(handler.outInclH);
      outputWindow->inclOutVpara->setChecked(handler.outInclVpara);
      outputWindow->inclOutVperp->setChecked(handler.outInclVperp);
      outputWindow->inclOutQpara->setChecked(handler.outInclQpara);
      outputWindow->inclOutQperp->setChecked(handler.outInclQperp);
      outputWindow->inclOutHpara->setChecked(handler.outInclHpara);
      outputWindow->inclOutHperp->setChecked(handler.outInclHperp);
      outputWindow->inclOutP->setChecked(handler.outInclP);
      outputWindow->inclOutT->setChecked(handler.outInclT);
      outputWindow->inclOutTpara->setChecked(handler.outInclTpara);
      outputWindow->inclOutTperp->setChecked(handler.outInclTperp);
      outputWindow->inclOutJ->setChecked(handler.outInclJ);
      outputWindow->inclOutJpara->setChecked(handler.outInclJpara);
      outputWindow->inclOutJperp->setChecked(handler.outInclJperp);
      outputWindow->inclOutCE->setChecked(handler.outInclCE);
      outputWindow->inclOutB->setChecked(handler.outInclB);
      outputWindow->inclOutScP->setChecked(handler.outInclScP);

// - potential settings      
      if (handler.onoff == 1) potentialGroup->setChecked(Qt::Checked);
      else potentialGroup->setChecked(Qt::Unchecked);
      
      if (handler.potOffset1.toInt() != INT_FILL_VALUE_PM) potentialOffset1->setText(handler.potOffset1);
      if (handler.potOffset2.toInt() != INT_FILL_VALUE_PM) potentialOffset2->setText(handler.potOffset2);
      potentialOffset2Units->setCurrentIndex(handler.potOffset2units);
      
// setup
      if (handler.setupSensor == "LEEA")
      {
         sensorCombiSelect->setCurrentIndex(0);
      }
      else if (handler.setupSensor == "HEEA")
      {
         sensorCombiSelect->setCurrentIndex(1);      
      }
      else if (handler.setupSensor == "TH1H2B")
      {
         sensorCombiSelect->setCurrentIndex(2);      
      }  
      else if (handler.setupSensor == "TL1L2B")
      {
         sensorCombiSelect->setCurrentIndex(3);      
      }  
      else if (handler.setupSensor == "L1L2B")
      {
         sensorCombiSelect->setCurrentIndex(4);      
      }   
      else if (handler.setupSensor == "H2H2B")
      {
         sensorCombiSelect->setCurrentIndex(5);      
      }    
      else if (handler.setupSensor == "TL1L2")
      {
         sensorCombiSelect->setCurrentIndex(6);      
      }  
      else if (handler.setupSensor == "TH1H2")
      {
         sensorCombiSelect->setCurrentIndex(7);      
      }  
      else if (handler.setupSensor == "H1H2")
      {
         sensorCombiSelect->setCurrentIndex(8);      
      }   
      else if (handler.setupSensor == "L1L2")
      {
         sensorCombiSelect->setCurrentIndex(9);      
      }                           
      else if (handler.setupSensor == "Overlap 2-sec")
      {
         sensorCombiSelect->setCurrentIndex(10);
      }
      
      if (handler.setupCoord == "GSE")
      {
         coordSelect->setCurrentIndex(0);
      }
      else if (handler.setupCoord == "Spacecraft")
      {
         coordSelect->setCurrentIndex(1);      
      }
      else if (handler.setupCoord == "Sensor")
      {
         coordSelect->setCurrentIndex(2);      
      }
      else if (handler.setupCoord == "SR2")
      {
         coordSelect->setCurrentIndex(3);      
      }
      else if (handler.setupCoord == "ISR2")
      {
         coordSelect->setCurrentIndex(4);      
      }
      else if (handler.setupCoord == "B-field aligned")
      {
         coordSelect->setCurrentIndex(5);
      }

// restricted energy integration
      if (handler.int_en_low.toInt() != INT_FILL_VALUE_PM) integrationWindow->energyLowEdit->setText(handler.int_en_low);
      if (handler.int_en_high.toInt() != INT_FILL_VALUE_PM) integrationWindow->energyHighEdit->setText(handler.int_en_high);
      if (handler.int_en_state == 1) integrationWindow->energyIntGroup->setChecked(true);
      else if (handler.int_en_state == 0) integrationWindow->energyIntGroup->setChecked(false);
      if (handler.int_en_type == 1) integrationWindow->energyIntType->setChecked(true);
      if (handler.int_en_scp_type == 1) integrationWindow->energyIntSpType->setChecked(true);

// restricted polar integration
      if (handler.int_pl_low.toInt() != INT_FILL_VALUE_PM) integrationWindow->polarLowEdit->setText(handler.int_pl_low);
      if (handler.int_pl_high.toInt() != INT_FILL_VALUE_PM) integrationWindow->polarHighEdit->setText(handler.int_pl_high);
      if (handler.int_pl_state == 1) integrationWindow->polarIntGroup->setChecked(true);
      else if (handler.int_pl_state == 0) integrationWindow->polarIntGroup->setChecked(false);

// time begin
      dataTabPage->stYearSpinBox->setValue(handler.yearBegin);
      dataTabPage->stDoySpinBox->setValue(handler.doyBegin);
      dataTabPage->stTimeEdit->setTime(QTime::fromString(handler.timeBegin, QString("hh:mm:ss")));
      dataTabPage->stDateEdit->setDate(QDate(handler.yearBegin, handler.monthBegin, handler.dayBegin));
     
// time end
      dataTabPage->spYearSpinBox->setValue(handler.yearEnd);
      dataTabPage->spDoySpinBox->setValue(handler.doyEnd);
      dataTabPage->spTimeEdit->setTime(QTime::fromString(handler.timeEnd, QString("hh:mm:ss")));
      dataTabPage->spDateEdit->setDate(QDate(handler.yearEnd, handler.monthEnd, handler.dayEnd));

// data download
      dataTabPage->dataScn->setCurrentIndex(handler.dataD_SCN);
      dataTabPage->dataPeace1->setCurrentIndex(handler.dataD_PEACE1);
      dataTabPage->dataPeace2->setCurrentIndex(handler.dataD_PEACE2);
      dataTabPage->dataPotential->setCurrentIndex(handler.dataD_POT);
      dataTabPage->dataAuxGSE->setCurrentIndex(handler.dataD_AUX);
      dataTabPage->dataMagField->setCurrentIndex(handler.dataD_FGM);
   }
}

//------------------------------------------------------------------------------
//
// save XML configuration file
//
//------------------------------------------------------------------------------

void Window::saveCPM()
{

// get filename from user

   QString saveFileText = QFileDialog::getSaveFileName(this, "", "", "PEACEmoments files (*.xml)");

// continue if we have a filename from the user

   if (!saveFileText.isEmpty())
   {
      if (!saveFileText.endsWith(".xml")) saveFileText = saveFileText + ".xml";

// begin defining the XML file

      QDomDocument doc("PEACEmomentsCaaML");
      QDomElement root = doc.createElement("peacemoments");
      doc.appendChild(root);

// version

      QDomElement currentVersion = doc.createElement("version");
      QDomText infoVersion = doc.createTextNode(QString("%1").arg(VERSION_PM));
      root.appendChild(currentVersion);
      currentVersion.appendChild(infoVersion);

// date & time

      QDomElement dtbDE = doc.createElement("begin_time");
      dtbDE.setAttribute("year", dataTabPage->stYearSpinBox->value());
      dtbDE.setAttribute("month", dataTabPage->stDateEdit->date().month());
      dtbDE.setAttribute("day", dataTabPage->stDateEdit->date().day());
      dtbDE.setAttribute("doy", dataTabPage->stDoySpinBox->value());
      dtbDE.setAttribute("time", dataTabPage->stTimeEdit->text());
      root.appendChild(dtbDE);

      QDomElement dteDE = doc.createElement("end_time");
      dteDE.setAttribute("year", dataTabPage->spYearSpinBox->value());
      dteDE.setAttribute("month", dataTabPage->spDateEdit->date().month());
      dteDE.setAttribute("day", dataTabPage->spDateEdit->date().day());
      dteDE.setAttribute("doy", dataTabPage->spDoySpinBox->value());
      dteDE.setAttribute("time", dataTabPage->spTimeEdit->text());
      root.appendChild(dteDE);

// required data

      QDomElement dataDE = doc.createElement("data");
      dataDE.setAttribute("scn", dataTabPage->dataScn->currentIndex());
      dataDE.setAttribute("peace1", dataTabPage->dataPeace1->currentIndex());
      dataDE.setAttribute("peace2", dataTabPage->dataPeace2->currentIndex());
      dataDE.setAttribute("potential", dataTabPage->dataPotential->currentIndex());
      dataDE.setAttribute("auxGSE", dataTabPage->dataAuxGSE->currentIndex());
      dataDE.setAttribute("fgmB", dataTabPage->dataMagField->currentIndex());
      root.appendChild(dataDE);

// files

      QDomElement filesDE = doc.createElement("files");
      filesDE.setAttribute("peace1", filesTabPage->editFilePeace1->displayText());
      filesDE.setAttribute("peace2", filesTabPage->editFilePeace2->displayText());
      filesDE.setAttribute("potential", filesTabPage->editFilePotential->displayText());
      filesDE.setAttribute("auxGSE", filesTabPage->editFileAuxGSE->displayText());
      filesDE.setAttribute("fgmB", filesTabPage->editFileMagField->displayText());
      root.appendChild(filesDE);

// sensor / coordinate system
      
      QDomElement setupDE = doc.createElement("setup");
      setupDE.setAttribute("sensor", sensorCombiSelect->currentText());
      setupDE.setAttribute("coord", coordSelect->currentText());
      root.appendChild(setupDE);
      
// spacecraft potential settings

      QDomElement scpotDE = doc.createElement("pot_corr");
      scpotDE.setAttribute("offset1", potentialOffset1->displayText());
      scpotDE.setAttribute("offset2", potentialOffset2->displayText());
      scpotDE.setAttribute("units", potentialOffset2Units->currentIndex());
      scpotDE.setAttribute("onoff", potentialGroup->isChecked());
      root.appendChild(scpotDE);    
      
// output settings

      QDomElement outputDE = doc.createElement("output");
      outputDE.setAttribute("out_dir", outputWindow->outputDirectoryEdit->displayText());
      outputDE.setAttribute("mom_filename", outputWindow->outputFileEdit->displayText());
      root.appendChild(outputDE);
      
// output includes
             
      QDomElement outputInclDE = doc.createElement("includes");        
      outputInclDE.setAttribute("n", outputWindow->inclOutN->isChecked());
      outputInclDE.setAttribute("v", outputWindow->inclOutV->isChecked());  
      outputInclDE.setAttribute("q", outputWindow->inclOutQ->isChecked());
      outputInclDE.setAttribute("h", outputWindow->inclOutH->isChecked());  
      outputInclDE.setAttribute("vpara", outputWindow->inclOutVpara->isChecked());
      outputInclDE.setAttribute("vperp", outputWindow->inclOutVperp->isChecked()); 
      outputInclDE.setAttribute("qpara", outputWindow->inclOutQpara->isChecked()); 
      outputInclDE.setAttribute("qperp", outputWindow->inclOutQperp->isChecked());    
      outputInclDE.setAttribute("hpara", outputWindow->inclOutHpara->isChecked()); 
      outputInclDE.setAttribute("hperp", outputWindow->inclOutHperp->isChecked());    
      outputInclDE.setAttribute("p", outputWindow->inclOutP->isChecked()); 
      outputInclDE.setAttribute("t", outputWindow->inclOutT->isChecked());    
      outputInclDE.setAttribute("tpara", outputWindow->inclOutTpara->isChecked()); 
      outputInclDE.setAttribute("tperp", outputWindow->inclOutTperp->isChecked());         
      outputInclDE.setAttribute("j", outputWindow->inclOutJ->isChecked());
      outputInclDE.setAttribute("jpara", outputWindow->inclOutJpara->isChecked());
      outputInclDE.setAttribute("jperp", outputWindow->inclOutJperp->isChecked());      
      outputInclDE.setAttribute("ce", outputWindow->inclOutCE->isChecked());
      outputInclDE.setAttribute("b", outputWindow->inclOutB->isChecked());
      outputInclDE.setAttribute("scp", outputWindow->inclOutScP->isChecked());                         
      root.appendChild(outputInclDE);        

// output filters

      QDomElement outputFiltDE = doc.createElement("filters");
      outputFiltDE.setAttribute("quality", outputWindow->filterQUA->isChecked());
      outputFiltDE.setAttribute("WHI", outputWindow->filterWHI->isChecked()); 
      outputFiltDE.setAttribute("EFW", outputWindow->filterEFW->isChecked());
      outputFiltDE.setAttribute("stats", outputWindow->filterStats->isChecked());
      outputFiltDE.setAttribute("datastream", outputWindow->filterDataStream->isChecked());  
      root.appendChild(outputFiltDE);          

// energy integration

      QDomElement intEnDE = doc.createElement("res_int_en");
      intEnDE.setAttribute("state", integrationWindow->energyIntGroup->isChecked());
      intEnDE.setAttribute("scp_type", integrationWindow->energyIntSpType->isChecked());
      intEnDE.setAttribute("type", integrationWindow->energyIntType->isChecked());
      intEnDE.setAttribute("low", integrationWindow->energyLowEdit->displayText());
      intEnDE.setAttribute("high", integrationWindow->energyHighEdit->displayText());
      root.appendChild(intEnDE);

// polar integration
 
      QDomElement intPlDE = doc.createElement("res_int_pl");
      intPlDE.setAttribute("state", integrationWindow->polarIntGroup->isChecked());
      intPlDE.setAttribute("low", integrationWindow->polarLowEdit->displayText());
      intPlDE.setAttribute("high", integrationWindow->polarHighEdit->displayText());
      root.appendChild(intPlDE);

// write to output file
      
      QFile fileSave(saveFileText);
      if(!fileSave.open(QIODevice::WriteOnly))
      {   
          QStringList message;
          message << "Unable to save file" << "Check write permissions?";
          errorBox(message);
          return;
      }
      
      QTextStream ts(&fileSave);
      ts << doc.toString();
      fileSave.close();
   }


}

//------------------------------------------------------------------------------
//
// calculate moments slot
//                    
//------------------------------------------------------------------------------

void Window::generateMoments()
{
   generateMomentsDo(1, "");
}

//------------------------------------------------------------------------------
//
// calculate moments slot
//
//------------------------------------------------------------------------------

void Window::generateMomentsPlotOnly(QString fileNameSave)
{
   generateMomentsDo(0, fileNameSave);
}

//------------------------------------------------------------------------------
//
// calculate moments
//
//------------------------------------------------------------------------------

void Window::generateMomentsDo(int whatToDo, QString fileNameSave)
{
   if (!calculationThread->isRunning())
   {

// get settings from GUI

      *userFilePeace1 = filesTabPage->editFilePeace1->displayText();
      *userFilePeace2 = filesTabPage->editFilePeace2->displayText();
      *userFilePotential = filesTabPage->editFilePotential->displayText();
      *userFileAuxGSE = filesTabPage->editFileAuxGSE->displayText();
      *userFileMagField = filesTabPage->editFileMagField->displayText();
      
      myParams->potentialCorrectionOnOff = potentialGroup->isChecked();

      myParams->potentialOffset1 = (potentialOffset1->displayText()).toDouble();
      myParams->potentialOffset2 = (potentialOffset2->displayText()).toDouble();

      if (potentialOffset2Units->currentText() == "bins") myParams->potentialOffset2Units = _BINS;
      else myParams->potentialOffset2Units = _EV;

      myParams->sensorCombination
         = getSensorCombination((char *)(sensorCombiSelect->currentText().toStdString()).c_str());
      myParams->coordinateSystem
         = getCoordinateSystem((char *)(coordSelect->currentText().toStdString()).c_str());

// check for invalid settings

     if (userFilePeace1->contains("PITCH_FULL"))
     {
        if (!userFilePeace2->isEmpty())
        {
           QStringList messageList;
           messageList << "PITCH_FULL has been specified" << "Cannot use data from two files";
           errorBox(messageList);
           return;
        }

        if (sensorCombiSelect->currentText() != "Overlap 2-sec")
        {
           QStringList messageList;
           messageList << "PITCH_FULL has been specified" << "Invalid sensor combination";
           errorBox(messageList);
           return;
        } 

        if (coordSelect->currentText() != "B-field aligned")
        {
           QStringList messageList;
           messageList << "PITCH_FULL has been specified" << "Invalid coordinate system";
           errorBox(messageList);
           return;
        }
     }

     if (userFilePeace2->contains("PITCH_FULL"))
     {
        QStringList messageList;
        messageList << "PITCH_FULL has been specified" << "Cannot use data from two files";
        errorBox(messageList);
        return;
     }

// get more settings from GUI

      bool iE = integrationWindow->energyIntGroup->isChecked();
      myParams->energyRestrictStatus = _OFF;
      if (iE) myParams->energyRestrictStatus = _ON;
      myParams->energyRestrictScpType = integrationWindow->energyIntSpType->isChecked();
      myParams->energyRestrictType = integrationWindow->energyIntType->isChecked();
      myParams->energyRestrictLow = integrationWindow->energyLowEdit->displayText().toDouble();
      myParams->energyRestrictHigh = integrationWindow->energyHighEdit->displayText().toDouble();

      bool iP = integrationWindow->polarIntGroup->isChecked();
      myParams->polarRestrictStatus = _OFF;
      if (iP) myParams->polarRestrictStatus = _ON;
      myParams->polarRestrictLow = integrationWindow->polarLowEdit->displayText().toDouble();
      myParams->polarRestrictHigh = integrationWindow->polarHighEdit->displayText().toDouble();

      myParams->filterEFW = outputWindow->filterEFW->isChecked();
      myParams->filterWHI = outputWindow->filterWHI->isChecked();
      myParams->filterStats = outputWindow->filterStats->isChecked();
      myParams->filterDataStream = outputWindow->filterDataStream->isChecked();
      myParams->filterQuality = outputWindow->filterQUA->isChecked();

// get more settings from GUI: what moments to calculate

      if (outputWindow->inclOutN->isChecked()) myWML->n = true;
      else myWML->n = false;
      if (outputWindow->inclOutV->isChecked()) myWML->v = true;
      else myWML->v = false;
      if (outputWindow->inclOutQ->isChecked()) myWML->q = true;
      else myWML->q = false;
      if (outputWindow->inclOutH->isChecked()) myWML->h = true;
      else myWML->h = false;
      if (outputWindow->inclOutP->isChecked()) myWML->p = true;
      else myWML->p = false;
      if (outputWindow->inclOutT->isChecked()) myWML->t = true;
      else myWML->t = false;
      if (outputWindow->inclOutVpara->isChecked()) myWML->vpara = true;
      else myWML->vpara = false;
      if (outputWindow->inclOutVperp->isChecked()) myWML->vperp = true;
      else myWML->vperp = false;
      if (outputWindow->inclOutQpara->isChecked()) myWML->qpara = true;
      else myWML->qpara = false;
      if (outputWindow->inclOutQperp->isChecked()) myWML->qperp = true;
      else myWML->qperp = false;
      if (outputWindow->inclOutHpara->isChecked()) myWML->hpara = true;
      else myWML->hpara = false;
      if (outputWindow->inclOutHperp->isChecked()) myWML->hperp = true;
      else myWML->hperp = false;
      if (outputWindow->inclOutTpara->isChecked()) myWML->tpara = true;
      else myWML->tpara = false;
      if (outputWindow->inclOutTperp->isChecked()) myWML->tperp = true;
      else myWML->tperp = false;
      if (outputWindow->inclOutB->isChecked()) myWML->b = true;
      else myWML->b = false;
      if (outputWindow->inclOutScP->isChecked()) myWML->scp = true;
      else myWML->scp = false;
      if (outputWindow->inclOutJ->isChecked()) myWML->j = true;
      else myWML->j = false;
      if (outputWindow->inclOutJpara->isChecked()) myWML->jpara = true;
      else myWML->jpara = false;
      if (outputWindow->inclOutJperp->isChecked()) myWML->jperp = true;
      else myWML->jperp = false;
      if (outputWindow->inclOutCE->isChecked()) myWML->ce = true;
      else myWML->ce = false;

// what to plot

      if (plotWindow->iPlotS->isChecked()) myPML->includeSpectrogram = true;
      else myPML->includeSpectrogram = false;
      if (plotWindow->iPlotN->isChecked()) myPML->n = true;
      else myPML->n = false;
      if (plotWindow->iPlotV->isChecked()) myPML->v = true;
      else myPML->v = false;
      if (plotWindow->iPlotQ->isChecked()) myPML->q = true;
      else myPML->q = false;
      if (plotWindow->iPlotH->isChecked()) myPML->h = true;
      else myPML->h = false;
      if (plotWindow->iPlotVpara->isChecked()) myPML->vpara = true;
      else myPML->vpara = false;
      if (plotWindow->iPlotVperp->isChecked()) myPML->vperp = true;
      else myPML->vperp = false;
      if (plotWindow->iPlotQpara->isChecked()) myPML->qpara = true;
      else myPML->qpara = false;
      if (plotWindow->iPlotQperp->isChecked()) myPML->qperp = true;
      else myPML->qperp = false;
      if (plotWindow->iPlotHpara->isChecked()) myPML->hpara = true;
      else myPML->hpara = false;
      if (plotWindow->iPlotHperp->isChecked()) myPML->hperp = true;
      else myPML->hperp = false;
      if (plotWindow->iPlotT->isChecked()) myPML->t = true;
      else myPML->t = false;
      if (plotWindow->iPlotTpara->isChecked()) myPML->tpara = true;
      else myPML->tpara = false;
      if (plotWindow->iPlotTperp->isChecked()) myPML->tperp = true;
      else myPML->tperp = false;
      if (plotWindow->iPlotB->isChecked()) myPML->b = true;
      else myPML->b = false;
      if (plotWindow->iPlotScP->isChecked()) myPML->scp = true;
      else myPML->scp = false;
      if (plotWindow->iPlotJ->isChecked()) myPML->j = true;
      else myPML->j = false;
      if (plotWindow->iPlotJpara->isChecked()) myPML->jpara = true;
      else myPML->jpara = false;
      if (plotWindow->iPlotJperp->isChecked()) myPML->jperp = true;
      else myPML->jperp = false;

      if (plotWindow->plotSettingsMoments->isChecked()) myPML->on = true;
      else myPML->on = false;

      if (!fileNameSave.isEmpty())
      {
         myPML->saveToFile = true;
         myPML->plotFileName = fileNameSave;
      }
      else myPML->saveToFile = false;

// pass settings to the moments calculation class

      *userFileOut = outputWindow->outputFileEdit->displayText();
      *userDirOut = outputWindow->outputDirectoryEdit->displayText();

      calculationThread->filePeace1 = userFilePeace1;
      calculationThread->filePeace2 = userFilePeace2;
      calculationThread->filePotential = userFilePotential;
      calculationThread->fileAuxGSE = userFileAuxGSE;
      calculationThread->fileMagField = userFileMagField;
      calculationThread->fileOut = userFileOut;
      calculationThread->dirOut = userDirOut;
      calculationThread->myWML = myWML;
      calculationThread->myPML = myPML;
      calculationThread->myParams = myParams;
      calculationThread->whatToDo = whatToDo;

// start the moments calculation

      generateMomentsAction->setEnabled(false);
      calculationThread->start();
   }
}

//------------------------------------------------------------------------------
//
// update status bar - moments calculation
//
//------------------------------------------------------------------------------

void Window::slotUpdateStatusBar()
{
   if (calculationThread->isRunning())
   {
      statusBar()->showMessage("Calculating moments...");
   }
}

//------------------------------------------------------------------------------
//
// update status bar - making spectrogram
//
//------------------------------------------------------------------------------

void Window::slotUpdateStatusBarSpectro()
{
   if (spectrogramThread->isRunning())
   {
      statusBar()->showMessage("Generating spectrogram...");
   }
}

//------------------------------------------------------------------------------
//
// update status bar
//
//------------------------------------------------------------------------------

void Window::slotUpdateStatusBarProgress(int iMessage)
{
   if (iMessage == _CURRENTLY_DOING_MOMENTS_PLOT)
   {
      statusBar()->showMessage("Generating moments plot...");
   }
}

//------------------------------------------------------------------------------
//
// display a message from the moments calculation routine if necessary
//
//------------------------------------------------------------------------------

void Window::displayMessage(int returnCode, QString message, int numRecords)
{
   if (returnCode != _SUCCESS)
   {
      statusBar()->showMessage("Error", 4000);
      QStringList messageList;
      messageList << message;
      errorBox(messageList);
   }
   else
   {
      QString message;
      if (numRecords > -1)
      {
         message = QString("Done, written %1 records to the output file").arg(numRecords);
      }
      else
      {
         message = QString("Done");
      }
      statusBar()->showMessage(message, 4000);
      progressBar->reset(); // reset the progress bar
   }
   generateMomentsAction->setEnabled(true);
}

//------------------------------------------------------------------------------
//
// display a message from the spectrogram if necessary
//
//------------------------------------------------------------------------------

void Window::displayMessageSpectro(int returnCode)
{
   if (returnCode != _SUCCESS)
   {
      statusBar()->showMessage("Error", 4000);
      
      char messageStr[100];
      statusHandler(returnCode, messageStr);
      
      QStringList messageList;
      messageList << QString(messageStr);
      errorBox(messageList);
   }
   else
   {
      QString message = "Done";
      statusBar()->showMessage(message, 4000);
   }
   plotWindow->replotSpectro->setEnabled(true);
}

//------------------------------------------------------------------------------
//
// display the moments plot
//
//------------------------------------------------------------------------------

void Window::displayMomentsPlotImage(const QImage &image)
{
   momentsWindow->plot = QPixmap::fromImage(image);
   momentsWindow->setMomentsWindowSize(myPML->plotWidth, myPML->plotHeight);
   momentsWindow->show();
   momentsWindow->setMaximumSize(momentsWindow->minimumWidth(),
                                 momentsWindow->minimumHeight());
   momentsWindow->update();
   momentsWindow->raise();
   momentsWindow->activateWindow(); 
}

//------------------------------------------------------------------------------
//
// error message
//
//------------------------------------------------------------------------------

void Window::errorBox(QStringList theMessage)
{  
   QMessageBox msgBox;
   msgBox.setIcon(QMessageBox::Critical);
   msgBox.setText(theMessage.at(0));
   if (theMessage.size() > 1)
   {  
      msgBox.setInformativeText(theMessage.at(1));
   }
   msgBox.exec();
}

//------------------------------------------------------------------------------
//
// display integration window
//
//------------------------------------------------------------------------------

void Window::displayIntegration()
{
   integrationWindow->show();
   integrationWindow->setMaximumSize(integrationWindow->minimumWidth(),
                                     integrationWindow->minimumHeight());
   integrationWindow->raise();
   integrationWindow->activateWindow();
}

//------------------------------------------------------------------------------
//
// display output settings window
//
//------------------------------------------------------------------------------
   
void Window::displayOutput()
{  
   outputWindow->show();
   outputWindow->setMaximumSize(outputWindow->minimumWidth(),
                                outputWindow->minimumHeight());
   outputWindow->raise();
   outputWindow->activateWindow();
}  

//------------------------------------------------------------------------------
//
// display plot settings window
//
//------------------------------------------------------------------------------
   
void Window::displayPlot()
{  
   plotWindow->show();
   plotWindow->setMaximumSize(plotWindow->minimumWidth(),
                              plotWindow->minimumHeight());
   plotWindow->raise();
   plotWindow->activateWindow();
}  

//-----------------------------------------------------------------------------
//
// Files slots
//
//-----------------------------------------------------------------------------

void FilesTab::slotEditFilePeace1()
{
   QString tmp;
   tmp = QFileDialog::getOpenFileName(this, "", "", "CAA PEACE files (*.cef *.cdf)");
   *userFilePeace1 = tmp;
   editFilePeace1->setText(tmp);
}

void FilesTab::slotEditFilePeace2()
{
   QString tmp;
   tmp = QFileDialog::getOpenFileName(this, "", "", "CAA PEACE files (*.cef *.cdf)");
   *userFilePeace2 = tmp;
   editFilePeace2->setText(tmp);
}

void FilesTab::slotEditFilePotential()
{
   QString tmp;
   tmp = QFileDialog::getOpenFileName(this, "", "", "Spacecraft potential files (*.cef *.cdf)");
   *userFilePotential = tmp;
   editFilePotential->setText(tmp);
}

void FilesTab::slotEditFileAuxGSE()
{
   QString tmp;
   tmp = QFileDialog::getOpenFileName(this, "", "", "AUX SP files (*.cef *.cdf)");
   *userFileAuxGSE = tmp;
   editFileAuxGSE->setText(tmp);
}

void FilesTab::slotEditFileMagField()
{
   QString tmp;
   tmp = QFileDialog::getOpenFileName(this, "", "", "FGM files (*.cef *.cdf)");
   *userFileMagField = tmp;
   editFileMagField->setText(tmp);
}

//-----------------------------------------------------------------------------
//
// Files tab
//
//-----------------------------------------------------------------------------

FilesTab::FilesTab(QWidget *parent)
   : QWidget(parent)
{

//   setPalette(QPalette(QColor("#c0c0c0")));
//   setAutoFillBackground(true);

   QLabel *filePeace1 = new QLabel("PEACE sensor 1");
   QLabel *filePeace2 = new QLabel("PEACE sensor 2");
   QLabel *filePotential = new QLabel("Spacecraft potential");
   QLabel *fileAuxGSE = new QLabel("AUX SP");
   QLabel *fileMagField = new QLabel("Magnetic field");

   filePeace1->setToolTip("Specify a CDF or CEF file containing PEACE data");
   filePeace2->setToolTip("Specify a second CDF or CEF file containing PEACE data (optional)");
   filePotential->setToolTip("Specify a CDF or CEF file containing spacecraft potential data");
   fileAuxGSE->setToolTip("Specify a CDF or CEF file containing AUX GSE spin-axis angles (required for GSE, SR2 and ISR2)");
   fileMagField->setToolTip("Specify a CDF or CEF file containing FGM B-field data (required for B-field aligned moments");
      
   editFilePeace1 = new QLineEdit();
   editFilePeace1->setFixedWidth(FILE_WIDTH);
   editFilePeace1->setAlignment(Qt::AlignRight);
   editFilePeace2 = new QLineEdit();
   editFilePeace2->setFixedWidth(FILE_WIDTH);
   editFilePeace2->setAlignment(Qt::AlignRight);
   editFilePotential = new QLineEdit();
   editFilePotential->setFixedWidth(FILE_WIDTH);
   editFilePotential->setAlignment(Qt::AlignRight);
   editFileAuxGSE = new QLineEdit();
   editFileAuxGSE->setFixedWidth(FILE_WIDTH);
   editFileAuxGSE->setAlignment(Qt::AlignRight);
   editFileMagField = new QLineEdit();
   editFileMagField->setFixedWidth(FILE_WIDTH);
   editFileMagField->setAlignment(Qt::AlignRight);
   
   editFilePeace1->setToolTip("Specify a CDF or CEF file containing PEACE data");
   editFilePeace2->setToolTip("Specify a second CDF or CEF file containing PEACE data (optional)");
   editFilePotential->setToolTip("Specify a CDF or CEF file containing spacecraft potential data");
   editFileAuxGSE->setToolTip("Specify a CDF or CEF file containing AUX GSE spin-axis angles (required for GSE, SR2 and ISR2)");
   editFileMagField->setToolTip("Specify a CDF or CEF file containing FGM B-field data (required for B-field aligned moments");   

   QPushButton *selectEditFilePeace1 = new QPushButton("...");
   QPushButton *selectEditFilePeace2 = new QPushButton("...");
   QPushButton *selectEditFilePotential = new QPushButton("...");
   QPushButton *selectEditFileAuxGSE = new QPushButton("...");
   QPushButton *selectEditFileMagField = new QPushButton("...");

   selectEditFilePeace1->setToolTip("Browse file system to select a CDF or CEF file");
   selectEditFilePeace2->setToolTip("Browse file system to select a CDF or CEF file");
   selectEditFilePotential->setToolTip("Browse file system to select a CDF or CEF file");
   selectEditFileAuxGSE->setToolTip("Browse file system to select a CDF or CEF file");
   selectEditFileMagField->setToolTip("Browse file system to select a CDF or CEF file");

   selectEditFilePeace1->setFixedWidth(FILE_BUTTON_WIDTH);
   selectEditFilePeace2->setFixedWidth(FILE_BUTTON_WIDTH);
   selectEditFilePotential->setFixedWidth(FILE_BUTTON_WIDTH);
   selectEditFileAuxGSE->setFixedWidth(FILE_BUTTON_WIDTH);
   selectEditFileMagField->setFixedWidth(FILE_BUTTON_WIDTH);

   connect(selectEditFilePeace1, SIGNAL(clicked()),
           this, SLOT(slotEditFilePeace1()));

   connect(selectEditFilePeace2, SIGNAL(clicked()),
           this, SLOT(slotEditFilePeace2()));

   connect(selectEditFilePotential, SIGNAL(clicked()),
           this, SLOT(slotEditFilePotential()));

   connect(selectEditFileAuxGSE, SIGNAL(clicked()),
           this, SLOT(slotEditFileAuxGSE()));

   connect(selectEditFileMagField, SIGNAL(clicked()),
           this, SLOT(slotEditFileMagField())); 

   QGridLayout *mainLayout = new QGridLayout;
   mainLayout->addWidget(filePeace1, 0, 0);
   mainLayout->addWidget(filePeace2, 1, 0);
   mainLayout->addWidget(filePotential, 2, 0);
   mainLayout->addWidget(fileAuxGSE, 3, 0);
   mainLayout->addWidget(fileMagField, 4, 0);

   mainLayout->addWidget(editFilePeace1, 0, 1);
   mainLayout->addWidget(editFilePeace2, 1, 1);
   mainLayout->addWidget(editFilePotential, 2, 1);
   mainLayout->addWidget(editFileAuxGSE, 3, 1);
   mainLayout->addWidget(editFileMagField, 4, 1);

   mainLayout->addWidget(selectEditFilePeace1, 0, 2);
   mainLayout->addWidget(selectEditFilePeace2, 1, 2);
   mainLayout->addWidget(selectEditFilePotential, 2, 2);
   mainLayout->addWidget(selectEditFileAuxGSE, 3, 2);
   mainLayout->addWidget(selectEditFileMagField, 4, 2);

   userFilePeace1 = new QString("");
   userFilePeace2 = new QString("");
   userFilePotential = new QString("");
   userFileAuxGSE = new QString("");
   userFileMagField = new QString("");

   
  setLayout(mainLayout);   
   
}

//-----------------------------------------------------------------------------
//
// Data tab
//
//-----------------------------------------------------------------------------

DataTab::DataTab(QWidget *parent)
   : QWidget(parent)
{

//   setPalette(QPalette(QColor("#c0c0c0")));
//   setAutoFillBackground(true);

// begin date/time

   QGroupBox *startGroup = new QGroupBox(tr("Start"));

   QHBoxLayout *yearDoyLayoutSt = new QHBoxLayout;

   QLabel *stYearLabel = new QLabel(tr("Year"));
   stYearSpinBox = new QSpinBox;
   stYearSpinBox->setRange(1900, 2012);
   stYearSpinBox->setSingleStep(1);
   stYearSpinBox->setValue(2000);
   stYearSpinBox->setFrame(false);
        
   connect(stYearSpinBox, SIGNAL(valueChanged(int)), 
           this, SLOT(modifiedStartYear(int)));

   QLabel *stDoyLabel = new QLabel(tr("Day"));
   stDoySpinBox = new QSpinBox;
   stDoySpinBox->setAlignment(Qt::AlignRight);
   stDoySpinBox->setRange(1, 366);
   stDoySpinBox->setSingleStep(1);    
   stDoySpinBox->setValue(1);
   stDoySpinBox->setFrame(false);

   yearDoyLayoutSt->addWidget(stYearLabel);
   yearDoyLayoutSt->addWidget(stYearSpinBox);
   yearDoyLayoutSt->addStretch(1);
   yearDoyLayoutSt->addWidget(stDoyLabel);
   yearDoyLayoutSt->addWidget(stDoySpinBox);

   QHBoxLayout *dateLayoutSt = new QHBoxLayout;
      
   QLabel *stDateLabel = new QLabel(tr("Date"));
   stDateEdit = new QDateTimeEdit(QDate(2000,1,1));
   stDateEdit->setDisplayFormat("yyyy-MM-dd");
   stDateEdit->setDateRange(QDate(1900, 1, 1), QDate(2012, 12, 31));
   stDateEdit->setFrame(false);
        
   connect(stDoySpinBox, SIGNAL(valueChanged(int)), 
           this, SLOT(modifiedStartDoy(int)));

   dateLayoutSt->addWidget(stDateLabel);
   dateLayoutSt->addStretch(1);
   dateLayoutSt->addWidget(stDateEdit);

   QHBoxLayout *timeLayoutSt = new QHBoxLayout;

   QLabel *stTimeLabel = new QLabel(tr("Time"));
   stTimeEdit = new QDateTimeEdit(QTime(0,0,0,0));
   stTimeEdit->setDisplayFormat("hh:mm:ss");
   stTimeEdit->setFrame(false);

   connect(stDateEdit, SIGNAL(dateChanged(const QDate &)),
           this, SLOT(modifiedStartDate(const QDate &)));

   timeLayoutSt->addWidget(stTimeLabel);
   timeLayoutSt->addStretch(1);
   timeLayoutSt->addWidget(stTimeEdit);

   QVBoxLayout *startBoxLayout = new QVBoxLayout;
   startBoxLayout->addLayout(yearDoyLayoutSt);
   startBoxLayout->addLayout(dateLayoutSt);
   startBoxLayout->addLayout(timeLayoutSt);
   startGroup->setLayout(startBoxLayout);	

// copy begin date/time to end date/time

   copyTimeDate = new QPushButton("->");
   copyTimeDate->setAutoDefault(false);
   copyTimeDate->setToolTip("Copy start date/time to end date/time");
   copyTimeDate->setFixedWidth(30);
   connect(copyTimeDate, SIGNAL(clicked()), 
           this, SLOT(copyStartToStop()));

// end date/time

   QGroupBox *stopGroup = new QGroupBox(tr("End"));

   QHBoxLayout *yearDoyLayoutSp = new QHBoxLayout;

   QLabel *spYearLabel = new QLabel(tr("Year"));
   spYearSpinBox = new QSpinBox;
   spYearSpinBox->setRange(1900, 2012);
   spYearSpinBox->setSingleStep(1);
   spYearSpinBox->setValue(2000);
   spYearSpinBox->setFrame(false);

   connect(spYearSpinBox, SIGNAL(valueChanged(int)),      
           this, SLOT(modifiedStopYear(int)));

    QLabel *spDoyLabel = new QLabel(tr("Day"));
    spDoySpinBox = new QSpinBox;
    spDoySpinBox->setAlignment(Qt::AlignRight);
    spDoySpinBox->setRange(1, 366);
    spDoySpinBox->setSingleStep(1);
    spDoySpinBox->setValue(1);
    spDoySpinBox->setFrame(false);

    connect(spDoySpinBox, SIGNAL(valueChanged(int)),
            this, SLOT(modifiedStopDoy(int)));

    yearDoyLayoutSp->addWidget(spYearLabel);
    yearDoyLayoutSp->addWidget(spYearSpinBox);
    yearDoyLayoutSp->addStretch(1);
    yearDoyLayoutSp->addWidget(spDoyLabel);
    yearDoyLayoutSp->addWidget(spDoySpinBox);

    QHBoxLayout *dateLayoutSp = new QHBoxLayout;

    QLabel *spDateLabel = new QLabel(tr("Date"));
    spDateEdit = new QDateTimeEdit(QDate(2000,1,1));
    spDateEdit->setDisplayFormat("yyyy-MM-dd");
    spDateEdit->setDateRange(QDate(1900, 1, 1), QDate(2012, 12, 31));
    spDateEdit->setFrame(false);

    dateLayoutSp->addWidget(spDateLabel);
    dateLayoutSp->addWidget(spDateEdit);

    QHBoxLayout *timeLayoutSp = new QHBoxLayout;

    QLabel *spTimeLabel = new QLabel(tr("Time"));
    spTimeEdit = new QDateTimeEdit(QTime(23,59,59,0));
    spTimeEdit->setDisplayFormat("hh:mm:ss");
    spTimeEdit->setFrame(false);

    connect(spDateEdit, SIGNAL(dateChanged(const QDate &)),
            this, SLOT(modifiedStopDate(const QDate &)));

    timeLayoutSp->addWidget(spTimeLabel);
    timeLayoutSp->addWidget(spTimeEdit);

    QVBoxLayout *stopBoxLayout = new QVBoxLayout;
    stopBoxLayout->addLayout(yearDoyLayoutSp);
    stopBoxLayout->addLayout(dateLayoutSp);
    stopBoxLayout->addLayout(timeLayoutSp);
    stopGroup->setLayout(stopBoxLayout);
	
// data group

   QGroupBox *dataGroup = new QGroupBox("Data");
   
   dataScn = new QComboBox;
   dataScn->addItem("C1");
   dataScn->addItem("C2");
   dataScn->addItem("C3");
   dataScn->addItem("C4");
   dataScn->setFixedWidth(SELECTOR_WIDTH_4);
   dataScn->setToolTip("Set spacecraft number");
   
   dataPeace1 = new QComboBox;
   dataPeace1->setMaxCount(11);
   dataPeace1->addItem("3DRL");
   dataPeace1->addItem("3DXPL");
   dataPeace1->addItem("3DXL");
   dataPeace1->addItem("3DXPLARL");
   dataPeace1->addItem("3DXLARL");
   dataPeace1->addItem("3DRH");
   dataPeace1->addItem("3DXPH");
   dataPeace1->addItem("3DXH");
   dataPeace1->addItem("3DXPLARH");
   dataPeace1->addItem("3DXLARH");  
   dataPeace1->addItem("");
   dataPeace1->setFixedWidth(SELECTOR_WIDTH_4);
   dataPeace1->setToolTip("Set required product for PEACE sensor 1");
   
   dataPeace2 = new QComboBox;
   dataPeace2->setMaxCount(11);
   dataPeace2->addItem("3DRL");
   dataPeace2->addItem("3DXPL");
   dataPeace2->addItem("3DXL");
   dataPeace2->addItem("3DXPLARL");
   dataPeace2->addItem("3DXLARL");
   dataPeace2->addItem("3DRH");
   dataPeace2->addItem("3DXPH");
   dataPeace2->addItem("3DXH");
   dataPeace2->addItem("3DXPLARH");
   dataPeace2->addItem("3DXLARH");  
   dataPeace2->addItem("");
   dataPeace2->setCurrentIndex(10);
   dataPeace2->setFixedWidth(SELECTOR_WIDTH_4);
   dataPeace2->setToolTip("Set required product for PEACE sensor 2 (optional)");
   
   dataPotential = new QComboBox;
   dataPotential->addItem("EFW_L3_P");
   dataPotential->addItem("EFW_L2_P");
   dataPotential->addItem("");
   dataPotential->setFixedWidth(SELECTOR_WIDTH_4);
   dataPotential->setToolTip("Select required spacecraft potential data if necessary");
   
   dataAuxGSE = new QComboBox;
   dataAuxGSE->addItem("CL_SP_AUX");
   dataAuxGSE->addItem("");
   dataAuxGSE->setFixedWidth(SELECTOR_WIDTH_4);
   dataAuxGSE->setToolTip("Select AUX SP data if necessary");
   
   dataMagField = new QComboBox;
   dataMagField->addItem("FGM_SPIN");
   dataMagField->addItem("");
   dataMagField->setFixedWidth(SELECTOR_WIDTH_4);
   dataMagField->setToolTip("Select FGM B-field data if necessary");
   
   QLabel *scnLabel = new QLabel("Spacecraft");
   QLabel *filePeace1 = new QLabel("PEACE sensor 1");
   QLabel *filePeace2 = new QLabel("PEACE sensor 2");
   QLabel *filePotential = new QLabel("Spacecraft potential");
   QLabel *fileAuxGSE = new QLabel("AUX SP");
   QLabel *fileMagField = new QLabel("Magnetic field");
   
   QGridLayout *gridLayout = new QGridLayout;
   gridLayout->addWidget(scnLabel, 0, 0);
   gridLayout->addWidget(filePeace1, 1, 0);
   gridLayout->addWidget(filePeace2, 2, 0);
   gridLayout->addWidget(filePotential, 3, 0);
   gridLayout->addWidget(fileAuxGSE, 4, 0);
   gridLayout->addWidget(fileMagField, 5, 0);
   gridLayout->addWidget(dataScn, 0, 1);
   gridLayout->addWidget(dataPeace1, 1, 1);
   gridLayout->addWidget(dataPeace2, 2, 1);
   gridLayout->addWidget(dataPotential, 3, 1);
   gridLayout->addWidget(dataAuxGSE, 4, 1);
   gridLayout->addWidget(dataMagField, 5, 1); 
   
   dataGroup->setLayout(gridLayout);

// username and password

   QGroupBox *dataAuth = new QGroupBox("CAA login");
   dataAuth->setToolTip("To download data, please enter your CAA username and password");
   
   QLabel *labelU = new QLabel("Username");
   QLabel *labelP = new QLabel("Password");
   
   authUser = new QLineEdit;
   
   authPass = new QLineEdit;
   authPass->setEchoMode(QLineEdit::Password);
   
   QHBoxLayout *h1 = new QHBoxLayout;
   h1->addWidget(labelU);
   h1->addWidget(authUser);
   h1->addStretch(1);
   h1->addWidget(labelP);
   h1->addWidget(authPass);
   
   dataAuth->setLayout(h1);
   
   downloadData = new QPushButton("&Download");
   downloadData->setToolTip("Download selected data from the CAA");
   downloadData->setFixedWidth(SELECTOR_WIDTH_1);
//   downloadData->setEnabled(false);
   connect(downloadData, SIGNAL(clicked()), this, SLOT(slotDownloadData()));

   availableData = new QPushButton("&Available data");
   availableData->setToolTip("List all available data for the selected time interval");
   availableData->setFixedWidth(SELECTOR_WIDTH_1);
   availableData->setEnabled(false);
   connect(availableData, SIGNAL(clicked()), this, SLOT(slotAvailableData()));
   
   QHBoxLayout *getIt = new QHBoxLayout;
   getIt->addStretch(1);
//   getIt->addWidget(availableData);
   getIt->addWidget(downloadData);

// make the page

   QHBoxLayout *l1 = new QHBoxLayout;
   QVBoxLayout *l2 = new QVBoxLayout;
   
   l1->addWidget(startGroup);
   l1->addWidget(copyTimeDate);
   l1->addWidget(stopGroup);
   
   l2->addLayout(l1);
   l2->addStretch(1);
   l2->addWidget(dataAuth);
   l2->addLayout(getIt);

   QHBoxLayout *mainLayout = new QHBoxLayout;
   mainLayout->addLayout(l2);
   mainLayout->addStretch(1);
   mainLayout->addWidget(dataGroup);
   
   setLayout(mainLayout);

// HTTP download ability

   http = new QHttp(this);
   downloadTracker = -1; // keep track of what we're currently downloading

   qfileTemp = new QTemporaryFile("peacemoments-caa_tmp.XXXXXX");

   qfilePeace1 = new QString;
   qfilePeace2 = new QString;
   qfilePotential = new QString;
   qfileFGM = new QString;
   qfileGSE = new QString;

   filesDownloaded = new QStringList;
   
   //userDirOut = new QString;

}

//------------------------------------------------------------------------------
//
// user has changed start day-of-year - update start month & day
//
//------------------------------------------------------------------------------

void DataTab::modifiedStartDoy(int doy)
{
   int year, month, day;
      
   year  = stYearSpinBox->value();
   month = doytoday(year, doy, 0);
   day   = doytoday(year, doy, 1);
      
   stDateEdit->setDate(QDate(year, month, day));
      
}
   
//------------------------------------------------------------------------------
//
// user has changed stop day-of-year - update stop month & day
//
//------------------------------------------------------------------------------

void DataTab::modifiedStopDoy(int doy)
{
   int year,  month, day;
      
   year  = spYearSpinBox->value();
   month = doytoday(year, doy, 0);
   day   = doytoday(year, doy, 1);
   
   spDateEdit->setDate(QDate(year, month, day));
   
}   

//------------------------------------------------------------------------------
//
// user has changed start date - update start doy
//
//------------------------------------------------------------------------------

void DataTab::modifiedStartDate(const QDate &startDate)  
{
   int year, month, day, doy;
   QDate *startDated = new QDate;
   *startDated = (QDate)startDate;
   year  = startDated->year();
   month = startDated->month();
   day   = startDated->day(); 
   doy   = daytodoy(year, month, day);
   stDoySpinBox->setValue(doy);
   stYearSpinBox->setValue(year);
}

//------------------------------------------------------------------------------
//
// user has changed stop date - update stop doy
//
//------------------------------------------------------------------------------

void DataTab::modifiedStopDate(const QDate &stopDate)
{
   int year, month, day, doy;
   QDate *stopDated = new QDate;  
   *stopDated = (QDate)stopDate;  
   year  = stopDated->year();
   month = stopDated->month();
   day   = stopDated->day();
   doy   = daytodoy(year, month, day);
   spDoySpinBox->setValue(doy);
   spYearSpinBox->setValue(year);
}

//------------------------------------------------------------------------------
//
// user has changed start year - update doy limits & start month & day
//
//------------------------------------------------------------------------------

void DataTab::modifiedStartYear(int year)
{
   int maxDays;
   modifiedStartDoy(stDoySpinBox->value());  
   maxDays = maxDaysInYear(year);
   stDoySpinBox->setRange(1, maxDays);
}

//------------------------------------------------------------------------------
//
// user has changed stop year - update doy limits & stop month & day
//
//------------------------------------------------------------------------------

void DataTab::modifiedStopYear(int year)
{
   int maxDays;
   modifiedStopDoy(spDoySpinBox->value());   
   maxDays = maxDaysInYear(year);    
   spDoySpinBox->setRange(1, maxDays);
}

//------------------------------------------------------------------------------
//
// copy start date/time to stop date/time
//
//------------------------------------------------------------------------------

void DataTab::copyStartToStop()
{
   spDoySpinBox->setValue(stDoySpinBox->value());
   spYearSpinBox->setValue(stYearSpinBox->value());
   spDateEdit->setDate(stDateEdit->date());
   spTimeEdit->setTime(stTimeEdit->time());          
}

//------------------------------------------------------------------------------
//
// download data from the CAA
//
//------------------------------------------------------------------------------

void DataTab::slotDownloadData()
{

// check if a CAA username & password were specified

   if (authPass->text().size() < 2 || authUser->displayText().size() < 2)
   {
      QStringList errorMessage;
      errorMessage << "Authentication Error                    ";
      errorMessage << "You must set a CAA username and password";
      errorBox(errorMessage);
      return;
   }

// get the date/time range from the GUI

   QString timeRange = stDateEdit->text() + "T" + stTimeEdit->text() + "Z/" +
                       spDateEdit->text() + "T" + spTimeEdit->text() + "Z";    

// determine the request string to send to the CAA

   QString request;
   int iRet = downloadDataPrep(dataScn->currentText(),
                               dataPeace1->currentText(),
                               dataPeace2->currentText(),
                               dataPotential->currentText(),
                               dataAuxGSE->currentText(),
                               dataMagField->currentText(),
                               timeRange, authUser->displayText(), authPass->text(),
                               &request);

// only continue if everything okay

   if (iRet != _SUCCESS)
   {
      char messageStr[100];
      statusHandler(iRet, messageStr);
      QStringList messageList;
      messageList << QString(messageStr);
      errorBox(messageList);
      emit downloadAborted();
      return;
   }
   else
   {
      downloadTracker = 0; // indicate that we're downing the little HTML file first
      if (!qfileTemp->open())
      {
         emit downloadAborted();
         QStringList messageList;
         messageList << "Unable to write temporary file [html]";
         errorBox(messageList);
         return;
      }
      http->setHost("caa.estec.esa.int");
      http->get(request, qfileTemp);
   }
}

//------------------------------------------------------------------------------
//
// check data availability from CAA
//
//------------------------------------------------------------------------------

void DataTab::slotAvailableData()
{

// check if a CAA username & password were specified

   if (authPass->displayText().size() < 2 || authUser->displayText().size() < 2)
   {
      QStringList errorMessage;
      errorMessage << "Authentication Error                    ";
      errorMessage << "You must set a CAA username and password";
      errorBox(errorMessage);
      return;
   } 
}

//------------------------------------------------------------------------------
//
// slot to receive http download finished signal
//
//------------------------------------------------------------------------------

void Window::httpRequestFinished(int id, bool error)
{
   if (error) id = 0; // remove compiler warning message
}

//------------------------------------------------------------------------------
//
// slot to receive http download finished signal
//
//------------------------------------------------------------------------------

void Window::httpRequestStarted(int id)
{
   if (id == 0) id = 0; // remove compiler warning

   if (dataTabPage->downloadTracker == 0)
   {
      statusBar()->showMessage("Sent request to CAA, waiting for response...");
   }
   else if (dataTabPage->downloadTracker == 1)
   {
      statusBar()->showMessage("Downloading files from CAA...");
   }
}

//------------------------------------------------------------------------------
//
// slot to receive http download finished signal
//
//------------------------------------------------------------------------------

void Window::downloadHandler(bool error)
{

   if (error) error = true; // remove compiler warning

// firstly we download a little HTML file specifying the location of the
// actual .zip file containing the CAA data

   if (dataTabPage->downloadTracker == 0)
   {
      dataTabPage->qfileTemp->close();
      dataTabPage->downloadTracker = 1;
      dataTabPage->downloadActualData();
   }
   else if (dataTabPage->downloadTracker == 1)
   {
      dataTabPage->downloadFile->close();
      dataTabPage->downloadTracker = -1;
      statusBar()->showMessage("Download complete", 4000);
      dataTabPage->extractFiles();

// decide how to assign the extracted files (PEACE data, potential, B-field, AUX GSE) to
// the files GUI
      
      int numPea = 0;
      for (int i=0; i<dataTabPage->filesDownloaded->size(); i++)
      {
         int ft = checkForFileType(dataTabPage->filesDownloaded->at(i));
         if (ft == _LEEA || ft == _HEEA || ft == _LEEA_HEEA)
         {
            if (numPea == 0)
            {
               filesTabPage->editFilePeace1->setText(dataTabPage->filesDownloaded->at(i));
               numPea++;
            }
            else if (numPea == 1) filesTabPage->editFilePeace2->setText(dataTabPage->filesDownloaded->at(i));
         }
         else if (ft == _POTENTIAL_SPIN || ft == _POTENTIAL_HIGH)
         {
            filesTabPage->editFilePotential->setText(dataTabPage->filesDownloaded->at(i));
         }
         else if (ft == _BFIELD_SPIN || ft == _BFIELD_HIGH)
         {
            filesTabPage->editFileMagField->setText(dataTabPage->filesDownloaded->at(i));
         }
         else if (ft == _GSE_SPINAXIS_SP)
         {
            filesTabPage->editFileAuxGSE->setText(dataTabPage->filesDownloaded->at(i));
         }
      }
   }
}

//------------------------------------------------------------------------------
//
// data download has been aborted
//
//------------------------------------------------------------------------------

void Window::downloadAborted()
{
   dataTabPage->downloadTracker = -1; 
   statusBar()->showMessage("Download failed", 4000);
}

//------------------------------------------------------------------------------
//
// download data from the CAA
//
//------------------------------------------------------------------------------

void DataTab::downloadActualData()
{

// firstly, read the HTML file to determine the address

   if (!qfileTemp->open())
   {                
      QStringList messageList;
      messageList << "Unable to open temporary file [html]";
      errorBox(messageList);
      return;
   }

   QTextStream textStream(qfileTemp);
   
   QString newAddress = "";

   while (!textStream.atEnd())
   {
      QString line = textStream.readLine();
      if (line.contains("href"))
      {
         int iB = line.indexOf("\"", 0);
         int iE = line.indexOf("\"", iB+1);
         newAddress = line.mid(iB+1, iE-iB-1);
      }
   }
   qfileTemp->close();

// delete the temporary file

   delete qfileTemp;
   qfileTemp = new QTemporaryFile("peacemoments-caa_tmp.XXXXXX");

// extract the file name itself from the full address

   int iS = newAddress.indexOf("caa_download", 0);
   QString actualFile = newAddress.mid(iS-1, -1);

   if (actualFile.contains("CAA_Error"))
   {
      emit downloadAborted();
      QStringList messageList;
      messageList << "CAA error" << "Check your username and password";
      errorBox(messageList);
      return;
   }

// check for errors, including time-out errors

   if (iS == -1)
   {
      emit downloadAborted();
      QStringList messageList;
      messageList << "Unable to get a valid response from the CAA";
      errorBox(messageList);
      return;
   }

// make directory for storing data

   QString dataDirectoryTmp;
   if (userDirOut.isEmpty()) dataDirectoryTmp = QString("Data000");
   else dataDirectoryTmp = userDirOut + QString("/Data000");

// we now want to create a new, unique directory name

   dataDirectory = "NULL";
   QDir myDataDirectory;
   if (!myDataDirectory.exists(dataDirectoryTmp))
   {
      dataDirectory = dataDirectoryTmp;
   }
   else
   {
      int id = 0;
      while (myDataDirectory.exists(dataDirectoryTmp))
      {
         id++;
         if (id > 999) id = 0;
         if (userDirOut.isEmpty()) dataDirectoryTmp = QString("");
         else dataDirectoryTmp = userDirOut + QString("/");
         dataDirectoryTmp = dataDirectoryTmp + QString("Data%1").arg(id, 3, 10, QChar('0'));
      }
   }
   dataDirectory = dataDirectoryTmp;

// now make the directory

   bool retD = myDataDirectory.mkdir(dataDirectory);

// if we can't make the directory, give an error

   if (!retD)
   {
      emit downloadAborted();
      QStringList messageList;
      messageList << "Unable to make a data directory";
      errorBox(messageList);
      return;
   }

// delete the original QFile, make a new one for downloading the .zip file

   delete downloadFile;
   downloadFile = new QFile(dataDirectory + "/data.zip");

// try to open the file

   if (!downloadFile->open(QIODevice::WriteOnly))
   { 
      emit downloadAborted();
      QStringList messageList;
      messageList << "Unable to write temporary file [zip]";
      errorBox(messageList);
      return;
   }

// download the file

   http->setHost("caa.estec.esa.int");
   http->get(actualFile, downloadFile);

}

//------------------------------------------------------------------------------
// 
// extract CDF/CEF files from the unzipped directory
// 
//------------------------------------------------------------------------------
   
void DataTab::extractFiles()
{

// unzip the file

   decompress(downloadFile->fileName(), dataDirectory);

// search for CEF/CDF files in the extract directories

   filesDownloaded->clear();
   extract_files(dataDirectory, filesDownloaded);

// delete the .zip file

   downloadFile->remove();

}

//------------------------------------------------------------------------------
//
// error message
//
//------------------------------------------------------------------------------

void DataTab::errorBox(QStringList theMessage)
{  
   QMessageBox msgBox;
   msgBox.setIcon(QMessageBox::Critical);
   msgBox.setText(theMessage.at(0));
   if (theMessage.size() > 1)
   {  
      msgBox.setInformativeText(theMessage.at(1));
   }
   msgBox.exec();
}

//------------------------------------------------------------------------------
//
// make a spectrogram
//
//------------------------------------------------------------------------------

void Window::makeSpectrogram(QString saveFileName)
{	
	
   if (!spectrogramThread->isRunning())
   {	
	
// get info from window

      spectrogramThread->yMin = (plotWindow->spectroYmin->displayText()).toDouble();
      spectrogramThread->yMax = (plotWindow->spectroYmax->displayText()).toDouble();
      spectrogramThread->cMin = (plotWindow->spectroCmin->displayText()).toDouble();
      spectrogramThread->cMax = (plotWindow->spectroCmax->displayText()).toDouble();

      spectrogramThread->units = plotWindow->spectroUnitsCombo->currentText();
      spectrogramThread->spectroPolars = plotWindow->spectroAnodesCombo->currentIndex();
      spectrogramThread->spectroEnergy = plotWindow->spectroEnergyCombo->currentIndex();
      spectrogramThread->spectroSensors = plotWindow->spectroSensorCombo->currentIndex();
   
      spectrogramThread->plotHeight = myPML->plotHeightSpectro;
      spectrogramThread->plotWidth = myPML->plotWidthSpectro;
   
      spectrogramThread->filePotential = filesTabPage->editFilePotential->displayText();
      spectrogramThread->filePeace1 = filesTabPage->editFilePeace1->displayText();
      spectrogramThread->filePeace2 = filesTabPage->editFilePeace2->displayText();  	
      spectrogramThread->fileToSave = saveFileName;

      spectrogramThread->scpOff1 = potentialOffset1->displayText().toFloat();
      spectrogramThread->scpOff2 = potentialOffset2->displayText().toFloat();
      spectrogramThread->scpOff2units = potentialOffset2Units->currentIndex();

// generate the spectrogram

      spectrogramThread->start();
   
   }  
}

//------------------------------------------------------------------------------
//
// display the spectrogram
//
//------------------------------------------------------------------------------

void Window::displaySpectrogramPlotImage(const QImage &image)
{
   spectrogramWindow->plot = QPixmap::fromImage(image);

   spectrogramWindow->setSpectrogramWindowSize(myPML->plotWidthSpectro, myPML->plotHeightSpectro);
   spectrogramWindow->setMaximumSize(spectrogramWindow->minimumWidth(),
                                     spectrogramWindow->minimumHeight());
   spectrogramWindow->update();
   spectrogramWindow->show();   
   spectrogramWindow->raise();
   spectrogramWindow->activateWindow();
}

//------------------------------------------------------------------------------
//
// load user's default settings
//
//------------------------------------------------------------------------------

// note: before attempting to set each item, we check to make sure that the
// values are actually defined.

void Window::loadSettings()
{
   QSettings settings(QSettings::NativeFormat, QSettings::UserScope, "mssl-adl", "peacemoments_caa");

// output window

   if (settings.value("out-incl/n", 11).toInt() != 11)
   outputWindow->inclOutN->setChecked(settings.value("out-incl/n").toBool());
   if (settings.value("out-incl/v", 11).toInt() != 11)
   outputWindow->inclOutV->setChecked(settings.value("out-incl/v").toBool());
   if (settings.value("out-incl/vpara", 11).toInt() != 11)
   outputWindow->inclOutVpara->setChecked(settings.value("out-incl/vpara").toBool());
   if (settings.value("out-incl/vperp", 11).toInt() != 11)
   outputWindow->inclOutVperp->setChecked(settings.value("out-incl/vperp").toBool());
   if (settings.value("out-incl/q", 11).toInt() != 11)
   outputWindow->inclOutQ->setChecked(settings.value("out-incl/q").toBool());
   if (settings.value("out-incl/qpara", 11).toInt() != 11)
   outputWindow->inclOutQpara->setChecked(settings.value("out-incl/qpara").toBool());
   if (settings.value("out-incl/qperp", 11).toInt() != 11)
   outputWindow->inclOutQperp->setChecked(settings.value("out-incl/qperp").toBool());
   if (settings.value("out-incl/h", 11).toInt() != 11)
   outputWindow->inclOutH->setChecked(settings.value("out-incl/h").toBool());
   if (settings.value("out-incl/hpara", 11).toInt() != 11)
   outputWindow->inclOutHpara->setChecked(settings.value("out-incl/hpara").toBool());
   if (settings.value("out-incl/hperp", 11).toInt() != 11)
   outputWindow->inclOutHperp->setChecked(settings.value("out-incl/hperp").toBool());
   if (settings.value("out-incl/p", 11).toInt() != 11)
   outputWindow->inclOutP->setChecked(settings.value("out-incl/p").toBool());
   if (settings.value("out-incl/t", 11).toInt() != 11)
   outputWindow->inclOutT->setChecked(settings.value("out-incl/t").toBool());
   if (settings.value("out-incl/tpara", 11).toInt() != 11)
   outputWindow->inclOutTpara->setChecked(settings.value("out-incl/tpara").toBool());
   if (settings.value("out-incl/tperp", 11).toInt() != 11)
   outputWindow->inclOutTperp->setChecked(settings.value("out-incl/tperp").toBool());
   if (settings.value("out-incl/j", 11).toInt() != 11)
   outputWindow->inclOutJ->setChecked(settings.value("out-incl/j").toBool());
   if (settings.value("out-incl/jpara", 11).toInt() != 11)
   outputWindow->inclOutJpara->setChecked(settings.value("out-incl/jpara").toBool());
   if (settings.value("out-incl/jperp", 11).toInt() != 11)
   outputWindow->inclOutJperp->setChecked(settings.value("out-incl/jperp").toBool());
   if (settings.value("out-incl/b", 11).toInt() != 11)
   outputWindow->inclOutB->setChecked(settings.value("out-incl/b").toBool());
   if (settings.value("out-incl/scp", 11).toInt() != 11)
   outputWindow->inclOutScP->setChecked(settings.value("out-incl/scp").toBool());
   if (settings.value("out-incl/ce", 11).toInt() != 11)
   outputWindow->inclOutCE->setChecked(settings.value("out-incl/ce").toBool());
   if (settings.value("out-filt/quality", 11).toInt() != 11)
   outputWindow->filterQUA->setChecked(settings.value("out-filt/quality").toBool());
   if (settings.value("out-filt/whi", 11).toInt() != 11)
   outputWindow->filterWHI->setChecked(settings.value("out-filt/whi").toBool());
   if (settings.value("out-filt/efw", 11).toInt() != 11)
   outputWindow->filterEFW->setChecked(settings.value("out-filt/efw").toBool());
   if (settings.value("out-filt/counts", 11).toInt() != 11)
   outputWindow->filterStats->setChecked(settings.value("out-filt/counts").toBool());
   if (settings.value("out-filt/datas", 11).toInt() != 11)
   outputWindow->filterDataStream->setChecked(settings.value("out-filt/datas").toBool());

// plot window

   if (settings.value("plot-mom/on-off", 11).toInt() != 11)
   plotWindow->plotSettingsMoments->setChecked(settings.value("plot-mom/on-off").toBool());
   if (settings.value("plot-mom/spectro", 11).toInt() != 11)
   plotWindow->iPlotS->setChecked(settings.value("plot-mom/spectro").toBool());
   if (settings.value("plot-mom/n", 11).toInt() != 11)
   plotWindow->iPlotN->setChecked(settings.value("plot-mom/n").toBool());
   if (settings.value("plot-mom/v", 11).toInt() != 11)
   plotWindow->iPlotV->setChecked(settings.value("plot-mom/v").toBool());
   if (settings.value("plot-mom/q", 11).toInt() != 11)
   plotWindow->iPlotQ->setChecked(settings.value("plot-mom/q").toBool());
   if (settings.value("plot-mom/h", 11).toInt() != 11)
   plotWindow->iPlotH->setChecked(settings.value("plot-mom/h").toBool());
   if (settings.value("plot-mom/vpara", 11).toInt() != 11)
   plotWindow->iPlotVpara->setChecked(settings.value("plot-mom/vpara").toBool());
   if (settings.value("plot-mom/vperp", 11).toInt() != 11)
   plotWindow->iPlotVperp->setChecked(settings.value("plot-mom/vperp").toBool());
   if (settings.value("plot-mom/qpara", 11).toInt() != 11)
   plotWindow->iPlotQpara->setChecked(settings.value("plot-mom/qpara").toBool());
   if (settings.value("plot-mom/qperp", 11).toInt() != 11)
   plotWindow->iPlotQperp->setChecked(settings.value("plot-mom/qperp").toBool());
   if (settings.value("plot-mom/hpara", 11).toInt() != 11)
   plotWindow->iPlotHpara->setChecked(settings.value("plot-mom/hpara").toBool());
   if (settings.value("plot-mom/hperp", 11).toInt() != 11)
   plotWindow->iPlotHperp->setChecked(settings.value("plot-mom/hperp").toBool());
   if (settings.value("plot-mom/t", 11).toInt() != 11)
   plotWindow->iPlotT->setChecked(settings.value("plot-mom/t").toBool());
   if (settings.value("plot-mom/tpara", 11).toInt() != 11)
   plotWindow->iPlotTpara->setChecked(settings.value("plot-mom/tpara").toBool());
   if (settings.value("plot-mom/tperp", 11).toInt() != 11)
   plotWindow->iPlotTperp->setChecked(settings.value("plot-mom/tperp").toBool());
   if (settings.value("plot-mom/j", 11).toInt() != 11)
   plotWindow->iPlotJ->setChecked(settings.value("plot-mom/j").toBool());
   if (settings.value("plot-mom/jpara", 11).toInt() != 11)
   plotWindow->iPlotJpara->setChecked(settings.value("plot-mom/jpara").toBool());
   if (settings.value("plot-mom/jperp", 11).toInt() != 11)
   plotWindow->iPlotJperp->setChecked(settings.value("plot-mom/jperp").toBool());
   if (settings.value("plot-mom/b", 11).toInt() != 11)
   plotWindow->iPlotB->setChecked(settings.value("plot-mom/b").toBool());
   if (settings.value("plot-mom/scp", 11).toInt() != 11)
   plotWindow->iPlotScP->setChecked(settings.value("plot-mom/scp").toBool());

}

//------------------------------------------------------------------------------ 
//
// slot to adjust user output directory variable in data tab GUI
//
//------------------------------------------------------------------------------

void Window::slotModifiedOutputDir(QString in)
{
   dataTabPage->userDirOut = in;
}

