/*******************************************************************************
 **
 ** $Id: qsiplot.h 174 2008-12-12 20:54:03Z AndrewLahiff $
 **
 ****************************************************************************** 
 ** 
 ** Copyright (C) 2007-2008 Andrew Lahiff.
 ** 
 ** This file is part of plotcf.
 **
 ** plotcf is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 ** 
 ** plotcf is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 ** 
 ** You should have received a copy of the GNU General Public License
 ** along with plotcf.  If not, see <http://www.gnu.org/licenses/>.
 ** 
 *******************************************************************************/

#ifndef QSIPLOT_H_
#define QSIPLOT_H_

#include <QMainWindow>
#include <QImage>
#include <QColor>
#include <QPainter>
#include <QRectF>
#include <QTime>
#include <QTextDocument>
#include <QPrinter>

#include <math.h>

#define Scalar 0
#define TimeSeries 1
#define Spectrogram 2
#define Sauvaud 3

#define Linear 0
#define Log 1

#define pTrue 1
#define pFalse 0

#define Off 0
#define TopLeft 1
#define TopRight 2
#define BottomLeft 3
#define BottomRight 4
#define Outside 5

#define MaxAllowedValue 1.0e+30
#define FillValue -3.4e+38

#define symbolWidth 7
#define symbolHeight 7

#define PLOT_LINE 0
#define PLOT_LINE_DASHED 7
#define PLOT_LINE_DOTTED 8
#define PLOT_LINE_DASH_DOT 9
#define PLOT_LINE_DASH_DOT_DOT 10
#define PLOT_CIRCLE_FILLED 1
#define PLOT_CIRCLE 2
#define PLOT_SQUARE_FILLED 3
#define PLOT_SQUARE 4
#define PLOT_PLUS 5
#define PLOT_CROSS 6

#define ALIGN_LEFT 0
#define ALIGN_RIGHT 1
#define ALIGN_CENTRE 2

#define QS_PNG 0
#define QS_JPG 1
#define QS_TIFF 2
#define QS_BMP 3
#define QS_PS 4

#define QS_ROW 0
#define QS_COL 1

class QsiPlot
{
   public:
      QsiPlot(int, int, QString, int = -1, bool = true);
      ~QsiPlot();
      void setBackgroundColor(QColor);
      void setLayout(int, int, QString, QString = "");
      void setPlotType(int, int ,int);
      void setScaleTypeY(int, int, int);
      void setScaleTypeX(int, int, int);
      void drawFrame(int, int, bool = true);
      void setPlotAxisLimits(int, int, double, double, double, double);
      void drawCurve(int, int, QColor, int, double *, double *, QString, int = 0);
      void setLabelX(int, int, QString);
      void setLabelY(int, int, QString);
      void setNumTicksX(int, int, double, int);
      void setNumTicksY(int, int, int, int);
      void showValuesX(int, int, int);
      void setPlotLineWidth(int);
      void setCurveLineWidth(int);
      double getValueFromPositionY(int, int, double);
      double getValueFromPositionX(int, int, double);
      void setTimeStringFormat(QString);
      void saveToFile(QString);
      void setGlobalFontLabels(QFont);
      void setGlobalFontTicks(QFont);
      void setGlobalFontTitle(QFont);
      void setGlobalFontLegends(QFont);
      void setGlobalFontColourBox(QFont);
      void setLegendLocation(int);
      void showLegend(int, int, int, int);
      void getAutoScaleY(double, double, int *, int *, double *, double *, int);
      void getAutoScaleX(double *, int, double, double, double *, int *, double *, double *);
      void drawSpectrogramSlice(int, int, double, double, double, double, double, double, double);
      void drawSauvaudSlice(int, int, double, double, double, double, double, double, double, double);
      void drawColourBox(int, int, double, double, QString, int = -1);
      void drawPlotTitle(int, int, QString);
      void setSauvaudEnergyTable(int, int, int, float *, float *, float *);
      void drawSauvaudEnergyLines(int, int);
      void setGrid(bool, bool);
      void drawLineY(int, int, QColor, int, double);
      QColor colourBarDefinition(double);
      void setColourBarDefinition(int);

      int initialYear;
      int initialDoy;
      QDate initialDate;

      int iPlotWidth, iPlotHeight;

      QImage *plot;

   private:
      int imageWidthPlot, imageHeightPlot;
      int numRows, numCols, numTotal;
      double widthSubPlot, heightSubPlot;
      double lMarginDef, rMarginDef, tMarginDef, bMarginDef, colourBarWidth;
      double axisLimits[20][2][2];      // actual plot x- & y-axis limits
      double axisLimitsTicks[20][2][2]; // plot x- & y-axis limits for ticks only
      double plotWidths[20];
      double plotHeights[20];
      double plotStartsX[20];
      double plotStartsY[20];
      int showValueX[20], plotType[20], scaleTypeY[20], scaleTypeX[20];
      int iPlotTickMinor, iPlotTickMajor, iPlotLineWidth, iCurveLineWidth;
      QString timeStringFormat;
      int titleHeight;
      bool layoutDone;
      int allValuesY[20];
      double heightOfValuesX;
      double SafeLog(double);
      void determineYaxisTicks(double, double, double *, double *, double *, int);
      void determineXaxisTicks(double, double, double *, int *);
      QFont fontLabels, fontTicks, fontTitle, fontLegends, fontColourBox, fontPanelTitle,
            fontTicksColour, fontTicksX, fontSubTitle;
      int legendLocation;
      void toTime(int, int *, int *, int *, int *);
      QString toTimeString(int);
      QSizeF drawRichText(QString, QFont, int, QPainter *, int = 0);
      QColor myOwnColor;
      QString myOwnName;
      QString pageTitle_;

      QImage *symbolToPlot;

      int numEnergyBinSauvaud;
      float energyLowerSauvaud[60];
      float energyUpperSauvaud[60];
      float energyCentreValueSauvaud[60];

      double gapBetweenAxisAndTick;

      bool xGrid, yGrid;

      int colourBarDef;
      
      int indexPosition(int, int);
      int indexPositionReturn(int, int, int);

      bool useGUI;

};

#endif /*QSIPLOT_H_*/
