/*******************************************************************************
 **
 ** $Id: readcdf.cpp 176 2009-04-17 23:39:24Z AndrewLahiff $
 **
 ****************************************************************************** 
 ** 
 ** Copyright (C) 2007-2008 Andrew Lahiff.
 ** 
 ** This file is part of ReadFile.
 **
 ** ReadFile is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 ** 
 ** ReadFile is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 ** 
 ** You should have received a copy of the GNU General Public License
 ** along with ReadFile.  If not, see <http://www.gnu.org/licenses/>.
 ** 
 *******************************************************************************/

#include "readfile.h"
#include <iostream>
using namespace std;

const int maxReadCharLength = 512;

//=============================================================================
//
// ReadCdf class
//
//=============================================================================

//-----------------------------------------------------------------------------
//
// constructor
//
//-----------------------------------------------------------------------------

ReadCdf::ReadCdf()
{
   iCurrentDataRecord = 0;
}

//-----------------------------------------------------------------------------
//
// set filename and open file
//
//-----------------------------------------------------------------------------

int ReadCdf::open(QString userFileName)
{
   status = CDFopenCDF((userFileName.toStdString()).c_str(), &id);
   if (status != CDF_OK) return(_READ_FAILED);
   else return(_READ_SUCCESS);
}

//-----------------------------------------------------------------------------
//
// get global meta data
//                        
//-----------------------------------------------------------------------------

QString ReadCdf::getGlobalMetaData(char* name)
{
   char strName[maxReadCharLength];
   long attrNum, numElements, datatype;

   attrNum = CDFgetAttrNum(id, name);
   if (attrNum < CDF_OK)
   {  
      QString tmp;
      tmp.clear();
      return tmp;
   }
   else
   {
      status = CDFinquireAttrgEntry(id, attrNum, 0, &datatype, &numElements);
      status = CDFgetAttrgEntry(id, attrNum, 0, strName);
      strName[numElements] = '\0';
      return QString(strName);
   }
}

//-----------------------------------------------------------------------------
//
// get variable meta data
//
//-----------------------------------------------------------------------------

QString ReadCdf::getVariableMetaData(int variable, char* name)
{
   QString Qstr;
   char strTmp[maxReadCharLength];
   long attrNum, numElements, datatype;

   attrNum = CDFgetAttrNum(id, name);
   status = CDFinquireAttrzEntry(id, attrNum, variable, &datatype, &numElements);

   if (status == CDF_OK)
   {
      if (datatype == CDF_CHAR)
      {
         status = CDFgetAttrzEntry (id, attrNum, variable, strTmp);
         strTmp[numElements] = '\0';
         Qstr = QString(strTmp);
      }
      else if (datatype == CDF_FLOAT || datatype == CDF_REAL4)
      {
         float* ftmp = new float[numElements];
         status = CDFgetAttrzEntry (id, attrNum, variable, (float *)ftmp);
         Qstr = QString("%1").arg(ftmp[0]);
         delete [] ftmp;
      }
      else if (datatype == CDF_DOUBLE || datatype == CDF_REAL8 || datatype == CDF_EPOCH)
      {
         double* dtmp = new double[numElements];
         status = CDFgetAttrzEntry (id, attrNum, variable, (double *)dtmp);
         Qstr = QString("%1").arg(dtmp[0]);
         delete [] dtmp;
      }
      else if (datatype == CDF_INT2)
      {
         short int* itmp = new short int[numElements];
         status = CDFgetAttrzEntry (id, attrNum, variable, (short int*)itmp);
         Qstr = QString("%1").arg(itmp[0]);
         delete [] itmp;
      }
      else if (datatype == CDF_INT4)
      {
         int* itmp = new int[numElements];
         status = CDFgetAttrzEntry (id, attrNum, variable, (int *)itmp);
         Qstr = QString("%1").arg(itmp[0]);
         delete [] itmp;
      }
      return Qstr;
   }
   else
   {                            
      QString tmp;
      tmp.clear();
      return tmp;
   }  
}

//-----------------------------------------------------------------------------
//
// read CDF header
//
//-----------------------------------------------------------------------------

int ReadCdf::readHeader()
{
   long numDims, encoding, majority, maxrRec, numrVars, maxzRec, numzVars,
        numAttrs, dimSizes[4];
   long attrNum, numElements, datatype, numElems, recVary, dimVarys[4];
   char strName[maxReadCharLength];

// initialize global attributes

   varGlobalData.initialize();

// determine number of variables

   status = CDFinquireCDF(id, 
                          &numDims, dimSizes,
                          &encoding, &majority,
                          &maxrRec, &numrVars,
                          &maxzRec, &numzVars, &numAttrs);

   numVariables = numzVars;
   maxRecordsZ = maxzRec;

// get required global meta data; note that CDF is case sensitive, but different CAA
// teams use different cases

   // DATASET_ID

   varGlobalData.DATASET_ID = getGlobalMetaData("DATASET_ID");
   if (varGlobalData.DATASET_ID.isEmpty())
   {
      varGlobalData.DATASET_ID = getGlobalMetaData("Dataset_ID");
   }

   // DATASET_TITLE

   varGlobalData.DATASET_TITLE = getGlobalMetaData("DATASET_TITLE");
   if (varGlobalData.DATASET_TITLE.isEmpty())
   {
      varGlobalData.DATASET_TITLE = getGlobalMetaData("Dataset_title");
   }

   // LOGICAL_SOURCE

   varGlobalData.LOGICAL_SOURCE = getGlobalMetaData("LOGICAL_SOURCE");
   if (varGlobalData.LOGICAL_SOURCE.isEmpty()) 
   {    
      varGlobalData.LOGICAL_SOURCE = getGlobalMetaData("Logical_source");
   }

  // EXPERIMENT

   varGlobalData.EXPERIMENT = getGlobalMetaData("EXPERIMENT");
   if (varGlobalData.EXPERIMENT.isEmpty()) 
   {    
      varGlobalData.EXPERIMENT = getGlobalMetaData("Experiment");
   }

   // OBSERVATORY

   varGlobalData.OBSERVATORY = getGlobalMetaData("OBSERVATORY");
   if (varGlobalData.OBSERVATORY.isEmpty())
   {
      varGlobalData.OBSERVATORY = getGlobalMetaData("Observatory");
   }

   // MISSION

   varGlobalData.MISSION = getGlobalMetaData("MISSION");
   if (varGlobalData.MISSION.isEmpty())
   {
      varGlobalData.MISSION = getGlobalMetaData("Mission");
   }

   // INSTRUMENT_NAME

   varGlobalData.INSTRUMENT_NAME = getGlobalMetaData("INSTRUMENT_NAME");
   if (varGlobalData.INSTRUMENT_NAME.isEmpty())
   {
      varGlobalData.INSTRUMENT_NAME = getGlobalMetaData("Instrument_name");
   }

   // DATA_TYPE

   varGlobalData.DATA_TYPE = getGlobalMetaData("DATA_TYPE");
   if (varGlobalData.DATA_TYPE.isEmpty())
   {
      varGlobalData.DATA_TYPE = getGlobalMetaData("Data_type");
   }

// allocate storage for variable meta data

   varMetaData.resize(numVariables);
   varDataType.resize(numVariables);

// get information about each variable

   for (int i=0; i<numVariables; i++)
   {

      // initialize all meta data for the current variable

      varMetaData[i].initialize();

      // data type

      status = CDFgetzVarDataType(id, i, &datatype);
      if (status != CDF_OK) varDataType.at(i) = UNKNOWN;
      else
      {
         if (datatype == CDF_INT4)
         {
            varDataType.at(i) = INT;
         }
         else if (datatype == CDF_REAL4 || datatype == CDF_FLOAT)
         {
            varDataType.at(i) = FLOAT;
         }
         else if (datatype == CDF_REAL8 || datatype == CDF_DOUBLE || datatype == CDF_EPOCH)
         {
            varDataType.at(i) = DOUBLE;
         }
         else if (datatype == CDF_INT2)
         {
            varDataType.at(i) = INT2;
         }
         else if (datatype == CDF_UINT1)
         {
            varDataType.at(i) = UINT1;
         }
         else if (datatype == CDF_UINT2)
         {
            varDataType.at(i) = UINT2;
         }
         else if (datatype == CDF_UINT4)
         {
            varDataType.at(i) = UINT;
         }
         else if (datatype == CDF_CHAR)
         {
            varDataType.at(i) = CHAR;
         }
         else varDataType.at(i) = UNKNOWN;
      }

      // ENTITY

      varMetaData[i].ENTITY = getVariableMetaData(i, "ENTITY");

      // UNITS

      varMetaData[i].UNITS = getVariableMetaData(i, "UNITS");

      // PARAMETER_TYPE

      varMetaData[i].PARAMETER_TYPE = getVariableMetaData(i, "PARAMETER_TYPE");

      // FIELDNAM

      varMetaData[i].FIELDNAM = getVariableMetaData(i, "FIELDNAM");

      // LABLAXIS

      varMetaData[i].LABLAXIS = getVariableMetaData(i, "LABLAXIS");

      // SCALETYP

      varMetaData[i].SCALETYP = getVariableMetaData(i, "SCALETYP");

      // SCALEMIN

      varMetaData[i].SCALEMIN = getVariableMetaData(i, "SCALEMIN");

      // SCALEMAX

      varMetaData[i].SCALEMAX = getVariableMetaData(i, "SCALEMAX");

      // VALUE_TYPE

      varMetaData[i].VALUE_TYPE = getVariableMetaData(i, "VALUE_TYPE");

      // VAR_TYPE

      varMetaData[i].VAR_TYPE = getVariableMetaData(i, "VAR_TYPE");
         
      // DEPEND_0

      varMetaData[i].DEPEND_0 = getVariableMetaData(i, "DEPEND_0");

      // DEPEND_1

      varMetaData[i].DEPEND_1 = getVariableMetaData(i, "DEPEND_1");

      // DEPEND_2

      varMetaData[i].DEPEND_2 = getVariableMetaData(i, "DEPEND_2");

      // DEPEND_3

      varMetaData[i].DEPEND_3 = getVariableMetaData(i, "DEPEND_3");

      // LABEL_1

      varMetaData[i].LABEL_1 = getVariableMetaData(i, "LABEL_1");

      // LABEL_2

      varMetaData[i].LABEL_2 = getVariableMetaData(i, "LABEL_2");

      // LABEL_3

      varMetaData[i].LABEL_3 = getVariableMetaData(i, "LABEL_3");

      // REPRESENTATION_1

      varMetaData[i].REPRESENTATION_1 = getVariableMetaData(i, "REPRESENTATION_1");

      // REPRESENTATION_2

      varMetaData[i].REPRESENTATION_2 = getVariableMetaData(i, "REPRESENTATION_2");

      // REPRESENTATION_3

      varMetaData[i].REPRESENTATION_3 = getVariableMetaData(i, "REPRESENTATION_3");


// DELTA_PLUS

      attrNum = CDFgetAttrNum(id, "DELTA_PLUS");
      status = CDFinquireAttrzEntry(id, attrNum, i,
                                    &datatype, &numElements);

      if (status == CDF_OK)
      {

// DELTA_PLUS can either be vector or single value, or a reference to
// another variable. We need to handle each case.

         if (datatype == CDF_CHAR)
         {
            char *strTmp = new char[numElements+1];
            status = CDFgetAttrzEntry (id, attrNum, i, (char *)strTmp);
            strTmp[numElements] = '\0';
            varMetaData[i].DELTA_PLUS = QString(strTmp);
            delete [] strTmp;
         }
         else if (datatype == CDF_FLOAT)
         {
            float *fTmp = new float[numElements];
            status = CDFgetAttrzEntry (id, attrNum, i, (float *)fTmp);
            varMetaData[i].DELTA_PLUS = QString("%1").arg(fTmp[0]);
            if (numElements > 1)
            {
               for (int h=1; h<numElements; h++)
               {
                  varMetaData[i].DELTA_PLUS = varMetaData[i].DELTA_PLUS + QString(",%1").arg(fTmp[h]);
               }
            }
            delete [] fTmp;
         }
      }
      else
      {
         varMetaData[i].DELTA_PLUS = QString("NULL");
      }

// DELTA_MINUS

      attrNum = CDFgetAttrNum(id, "DELTA_MINUS");
      status = CDFinquireAttrzEntry(id, attrNum, i,
                                    &datatype, &numElements);

      if (status == CDF_OK)
      {

// DELTA_MINUS can either be vector or single value, or a reference to
// another variable. We need to handle each case.

         if (datatype == CDF_CHAR)
         {
            char *strTmp = new char[numElements+1];
            status = CDFgetAttrzEntry (id, attrNum, i, (char *)strTmp);
            strTmp[numElements] = '\0';
            varMetaData[i].DELTA_MINUS = QString(strTmp);
            delete [] strTmp;
         }
         else if (datatype == CDF_FLOAT)
         {
            float *fTmp = new float[numElements];
            status = CDFgetAttrzEntry (id, attrNum, i, (float *)fTmp);
            varMetaData[i].DELTA_MINUS = QString("%1").arg(fTmp[0]);
            if (numElements > 1)
            {
               for (int h=1; h<numElements; h++)
               {
                  varMetaData[i].DELTA_MINUS = varMetaData[i].DELTA_MINUS + QString(",%1").arg(fTmp[h]);
               }
            }
            delete [] fTmp;
         }
      }
      else
      {
         varMetaData[i].DELTA_MINUS = QString("NULL");
      }                      

      // DISPLAYTYPE

      varMetaData[i].DISPLAYTYPE = getVariableMetaData(i, "DISPLAYTYPE");

      // SIZES

      varMetaData[i].SIZES = getVariableMetaData(i, "SIZES");

      // FILLVAL

      varMetaData[i].FILLVAL = getVariableMetaData(i, "FILLVAL");
   
      // PROPERTY

      varMetaData[i].PROPERTY = getVariableMetaData(i, "PROPERTY");

// name, dimensions, sizes

      status = CDFgetzVarNumDims(id, i, &numDims);
      status = CDFinquirezVar(id, i, strName, &datatype,
                              &numElems, &numDims, dimSizes, &recVary, dimVarys);

      if (status != CDF_OK)
      {
         std::vector<int> tmpVec;
         tmpVec.push_back(-1);
         varDataSizes.push_back(tmpVec);
      }
      else
      {
         varMetaData[i].NAME = QString(strName);

         std::vector<int> tmpVec;

         if (numDims > 0) 
         {
            tmpVec.resize(numDims);
            for (int j=0; j<numDims; ++j)
            {
               tmpVec.at(j) = dimSizes[j];
            }
         }
         else
         {
            // scalars are reported has having 0 dimensions, but we want to refer to a scalar as
            // having 1 dimension
            tmpVec.resize(1);
            tmpVec.at(0) = 1;
         }

         varDataSizes.push_back(tmpVec);
      }     
   }

// check if LABEL_* and/or REPRESENTATION_*  need adjustments, i.e. they are references
// to other variables which actually store the data. Rather than just returning the name of
// the other variable we want to return the actual data, so that the data doesn't
// have to do this themselves.

   for (int i=0; i<numVariables; ++i)
   {
      for (int j=0; j<numVariables; ++j)
      {
         if (varMetaData[i].LABEL_1 == varMetaData[j].NAME)
         {
            status = CDFgetzVarNumDims(id, j, &numDims);
            status = CDFinquirezVar(id, j, strName, &datatype,
                                    &numElems, &numDims, dimSizes, &recVary, dimVarys);

            int iTotal = 0;
            for (int ii=0; ii<numDims; ++ii)
               for (int jj=0; jj<dimSizes[ii]; ++jj)
                  iTotal++;

            char *charTmp = new char[numElems*iTotal+1];
            int status;
            status = CDFgetzVarRecordData(id, j, 0, charTmp);
            charTmp[numElems*iTotal] = '\0';
            QString qsTmp = charTmp;

            varMetaData[i].LABEL_1 = "";
            for (int ii=0; ii<iTotal; ++ii)
            {
               if (ii>0) varMetaData[i].LABEL_1 = varMetaData[i].LABEL_1 + ",";
               varMetaData[i].LABEL_1 = varMetaData[i].LABEL_1 + qsTmp.mid(ii*numElems, numElems);
            }
            delete [] charTmp;
         }
         if (varMetaData[i].LABEL_2 == varMetaData[j].NAME)
         {  
            status = CDFgetzVarNumDims(id, j, &numDims);
            status = CDFinquirezVar(id, j, strName, &datatype,
                                    &numElems, &numDims, dimSizes, &recVary, dimVarys);

            int iTotal = 0;
            for (int ii=0; ii<numDims; ++ii)
               for (int jj=0; jj<dimSizes[ii]; ++jj)
                  iTotal++;
            
            char *charTmp = new char[numElems*iTotal+1];
            int status;
            status = CDFgetzVarRecordData(id, j, 0, charTmp);
            charTmp[numElems*iTotal] = '\0';
            QString qsTmp = charTmp;
            
            varMetaData[i].LABEL_2 = "";
            for (int ii=0; ii<iTotal; ++ii)
            {
               if (ii>0) varMetaData[i].LABEL_2 = varMetaData[i].LABEL_2 + ",";
               varMetaData[i].LABEL_2 = varMetaData[i].LABEL_2 + qsTmp.mid(ii*numElems, numElems);
            }
            delete [] charTmp;
         }
         if (varMetaData[i].LABEL_3 == varMetaData[j].NAME)
         {  
            status = CDFgetzVarNumDims(id, j, &numDims);
            status = CDFinquirezVar(id, j, strName, &datatype,
                                    &numElems, &numDims, dimSizes, &recVary, dimVarys);
                                    
            int iTotal = 0;
            for (int ii=0; ii<numDims; ++ii)
               for (int jj=0; jj<dimSizes[ii]; ++jj)
                  iTotal++;
                  
            char *charTmp = new char[numElems*iTotal+1];
            int status;
            status = CDFgetzVarRecordData(id, j, 0, charTmp);
            charTmp[numElems*iTotal] = '\0';
            QString qsTmp = charTmp; 
            
            varMetaData[i].LABEL_3 = "";
            for (int ii=0; ii<iTotal; ++ii)
            {
               if (ii>0) varMetaData[i].LABEL_3 = varMetaData[i].LABEL_3 + ",";
               varMetaData[i].LABEL_3 = varMetaData[i].LABEL_3 + qsTmp.mid(ii*numElems, numElems);
            }  
            delete [] charTmp;
         }  
         if (varMetaData[i].REPRESENTATION_1 == varMetaData[j].NAME)
         {
            status = CDFgetzVarNumDims(id, j, &numDims);
            status = CDFinquirezVar(id, j, strName, &datatype,
                                    &numElems, &numDims, dimSizes, &recVary, dimVarys);
                                    
            int iTotal = 0;
            for (int ii=0; ii<numDims; ++ii)
               for (int jj=0; jj<dimSizes[ii]; ++jj)
                  iTotal++;
                  
            char *charTmp = new char[numElems*iTotal+1];
            int status;
            status = CDFgetzVarRecordData(id, j, 0, charTmp);
            charTmp[numElems*iTotal] = '\0';
            QString qsTmp = charTmp; 
            
            varMetaData[i].REPRESENTATION_1 = "";
            for (int ii=0; ii<iTotal; ++ii)
            {
               if (ii>0) varMetaData[i].REPRESENTATION_1 = varMetaData[i].REPRESENTATION_1 + ",";
               varMetaData[i].REPRESENTATION_1 = varMetaData[i].REPRESENTATION_1 + qsTmp.mid(ii*numElems, numElems);
            }  
            delete [] charTmp;
         }  
         if (varMetaData[i].REPRESENTATION_2 == varMetaData[j].NAME)
         {
            status = CDFgetzVarNumDims(id, j, &numDims);
            status = CDFinquirezVar(id, j, strName, &datatype,
                                    &numElems, &numDims, dimSizes, &recVary, dimVarys);

            int iTotal = 0;
            for (int ii=0; ii<numDims; ++ii)
               for (int jj=0; jj<dimSizes[ii]; ++jj)
                  iTotal++;
                  
            char *charTmp = new char[numElems*iTotal+1];
            int status;
            status = CDFgetzVarRecordData(id, j, 0, charTmp);
            charTmp[numElems*iTotal] = '\0';
            QString qsTmp = charTmp;

            varMetaData[i].REPRESENTATION_2 = "";
            for (int ii=0; ii<iTotal; ++ii)
            {
               if (ii>0) varMetaData[i].REPRESENTATION_2 = varMetaData[i].REPRESENTATION_2 + ",";
               varMetaData[i].REPRESENTATION_2 = varMetaData[i].REPRESENTATION_2 + qsTmp.mid(ii*numElems, numElems);
            }
            delete [] charTmp;
         }
         if (varMetaData[i].REPRESENTATION_3 == varMetaData[j].NAME)
         {
            status = CDFgetzVarNumDims(id, j, &numDims);
            status = CDFinquirezVar(id, j, strName, &datatype,
                                    &numElems, &numDims, dimSizes, &recVary, dimVarys);

            int iTotal = 0;
            for (int ii=0; ii<numDims; ++ii)
               for (int jj=0; jj<dimSizes[ii]; ++jj)
                  iTotal++;
                  
            char *charTmp = new char[numElems*iTotal+1];
            int status;
            status = CDFgetzVarRecordData(id, j, 0, charTmp);
            charTmp[numElems*iTotal] = '\0';
            QString qsTmp = charTmp;

            varMetaData[i].REPRESENTATION_3 = "";
            for (int ii=0; ii<iTotal; ++ii)
            {
               if (ii>0) varMetaData[i].REPRESENTATION_3 = varMetaData[i].REPRESENTATION_3 + ",";
               varMetaData[i].REPRESENTATION_3 = varMetaData[i].REPRESENTATION_3 + qsTmp.mid(ii*numElems, numElems);
            }
            delete [] charTmp;
         }
      }
   }

// determine the begin/end times of the file

   iCurrentDataRecord = 0;
   getEpochValue(0, &timeFileBegin);
   iCurrentDataRecord = maxRecordsZ - 1;
   getEpochValue(0, &timeFileEnd);
   iCurrentDataRecord = 0;

   return _READ_SUCCESS;
}

//-----------------------------------------------------------------------------
//
// return number of variables
//
//-----------------------------------------------------------------------------

int ReadCdf::getNumVariables()
{
   return numVariables;
}

//-----------------------------------------------------------------------------
//
// return number of dimensions of variable
//
//-----------------------------------------------------------------------------

int ReadCdf::getVariableDimension(int index)
{
   if (index >= 0 && index < numVariables)
   {
      return varDataSizes.at(index).size();
   }
   else return _READ_FAILED;
}

//-----------------------------------------------------------------------------
//
// return size of a dimension of a variable
//
//-----------------------------------------------------------------------------

int ReadCdf::getVariableSize(int index, int inIndex)
{
   if (index >= 0 && index < numVariables)
   {
      if (inIndex >= 0 && inIndex < static_cast<int>(varDataSizes.at(index).size()))
      {
         return varDataSizes.at(index).at(inIndex);
      }
      else return _READ_FAILED;
   }
   else return _READ_FAILED;
}

//-----------------------------------------------------------------------------
//
// return variable data type
//
//-----------------------------------------------------------------------------

DataType ReadCdf::getVariableDataType(int index)
{
   if (index >= 0 && index < numVariables)
   {
      return varDataType.at(index);
   }
   else return UNKNOWN;
}

//-----------------------------------------------------------------------------
//
// get begin date/time
//
//-----------------------------------------------------------------------------

QString ReadCdf::getFileBeginDateTime()
{
   return timeFileBegin;
}

//-----------------------------------------------------------------------------
//
// get end date/time
//
//-----------------------------------------------------------------------------

QString ReadCdf::getFileEndDateTime()
{
   return timeFileEnd;
}

//-----------------------------------------------------------------------------
//
// read a data record
//
//-----------------------------------------------------------------------------

int ReadCdf::readRecord(bool unused)
{
   if (unused) {} // remove annoying compiler warning
   iCurrentDataRecord++;
   if (iCurrentDataRecord < maxRecordsZ) return _READ_SUCCESS;
   else return _READ_FAILED;
}

//-----------------------------------------------------------------------------
//
// get a data value from the current record (short integer version)
//
//-----------------------------------------------------------------------------

int ReadCdf::getDataValue(int index, short int *ptr)
{
   if (index >= 0 && index < numVariables)
   {
      int status = CDFgetzVarRecordData(id, index, iCurrentDataRecord, ptr);
      if (status == CDF_OK) return(_READ_SUCCESS);
      else return(_READ_FAILED);
   }
   else return(_READ_FAILED);
}

//-----------------------------------------------------------------------------
//
// get a data value from the current record (integer version)
//
//-----------------------------------------------------------------------------

int ReadCdf::getDataValue(int index, int *ptr)
{
   if (index >= 0 && index < numVariables)
   {
      int status = CDFgetzVarRecordData(id, index, iCurrentDataRecord, ptr);
      if (status == CDF_OK) return(_READ_SUCCESS);
      else return(_READ_FAILED);
   }
   else return(_READ_FAILED);
}

//-----------------------------------------------------------------------------
//    
// get a data value from the current record (float version)
//    
//-----------------------------------------------------------------------------

int ReadCdf::getDataValue(int index, float *ptr)
{     
   if (index >= 0 && index < numVariables)
   {
      int status = CDFgetzVarRecordData(id, index, iCurrentDataRecord, ptr);
      if (status == CDF_OK) return(_READ_SUCCESS);
      else return(_READ_FAILED);
   }
   else return(_READ_FAILED);
}

//-----------------------------------------------------------------------------
//    
// get a data value from the current record (double version)
//    
//-----------------------------------------------------------------------------

int ReadCdf::getDataValue(int index, double *ptr)
{     
   if (index >= 0 && index < numVariables)
   {
      int status = CDFgetzVarRecordData(id, index, iCurrentDataRecord, ptr);
      if (status == CDF_OK) return(_READ_SUCCESS);
      else return(_READ_FAILED);
   }
   else return(_READ_FAILED);
}

//-----------------------------------------------------------------------------
//    
// get a data value from the current record (string version)
//    
//-----------------------------------------------------------------------------

int ReadCdf::getDataValue(int index, QString *ptr)
{
   char c_ptr[50];
   if (index >= 0 && index < numVariables)
   {
      int status = CDFgetzVarRecordData(id, index, iCurrentDataRecord, c_ptr);
      if (status == CDF_OK)
      {
         *ptr = c_ptr;
         return(_READ_SUCCESS);
      }
      else return(_READ_FAILED);
   }
   else return(_READ_FAILED);
}

//-----------------------------------------------------------------------------
//    
// get a data value from the current record (EPOCH)
//    
//-----------------------------------------------------------------------------

int ReadCdf::getEpochValue(int index, QString *ptr)
{
   double cdfEpoch;
   if (index >= 0 && index < numVariables)
   {
      int status = CDFgetzVarRecordData(id, index, iCurrentDataRecord, &cdfEpoch);
      if (status == CDF_OK)
      {
         long YY, MM, DD, hh, mm, ss, ms;
	 char epString[25];
	 EPOCHbreakdown(cdfEpoch, &YY, &MM, &DD, &hh, &mm, &ss, &ms);
	 sprintf(epString, "%ld-%02ld-%02ldT%02ld:%02ld:%02ld.%03ldZ", YY, MM, DD, hh, mm, ss, ms);
	 *ptr = epString;
         return(_READ_SUCCESS);
      }  
      else return(_READ_FAILED);
   }
   else return(_READ_FAILED);
}

//-----------------------------------------------------------------------------
//
// close CDF file
//
//-----------------------------------------------------------------------------

void ReadCdf::close()
{
   status = CDFcloseCDF(id);
}

//-----------------------------------------------------------------------------
//
// get a global attribute
//
//-----------------------------------------------------------------------------

QString ReadCdf::getGlobalAttr(QString what)
{
   if (what == "MISSION")
   {
      return(varGlobalData.MISSION);
   }
   else if (what == "OBSERVATORY")
   {
      return(varGlobalData.OBSERVATORY);
   }
   else if (what == "EXPERIMENT")
   {
      return(varGlobalData.EXPERIMENT);
   }
   else if (what == "INSTRUMENT_NAME")
   {
      return(varGlobalData.INSTRUMENT_NAME);
   }
   else if (what == "DATASET_ID")
   {
      return(varGlobalData.DATASET_ID);
   }
   else if (what == "DATASET_TITLE")
   {
      return(varGlobalData.DATASET_TITLE);
   }
   else if (what == "LOGICAL_SOURCE")
   {
      return(varGlobalData.LOGICAL_SOURCE);
   }
   else if (what == "DATA_TYPE")
   {
      return(varGlobalData.DATA_TYPE);
   }
   else return("");
}

//-----------------------------------------------------------------------------
//
// get a variable attribute
//
//-----------------------------------------------------------------------------

QString ReadCdf::getVariableAttr(int index, QString what)
{
   if (what == "UNITS")
   {
      return(varMetaData[index].UNITS);
   }
   else if (what == "PARAMETER_TYPE")
   {
      return(varMetaData[index].PARAMETER_TYPE);
   }
   else if (what == "LABLAXIS")
   {
      return(varMetaData[index].LABLAXIS);	
   }
   else if (what == "NAME")
   {
      return(varMetaData[index].NAME);
   }
   else if (what == "PROPERTY")
   {
      return(varMetaData[index].PROPERTY);
   }
   else if (what == "FILLVAL")
   {
      return(varMetaData[index].FILLVAL);
   }
   else if (what == "SCALETYP")
   {
      return(varMetaData[index].SCALETYP);
   }
   else if (what == "SCALEMIN")
   {
      return(varMetaData[index].SCALEMIN);
   }
   else if (what == "SCALEMAX")
   {
      return(varMetaData[index].SCALEMAX);
   }      
   else if (what == "VALUE_TYPE")
   {
      return(varMetaData[index].VALUE_TYPE);
   }
   else if (what == "LABEL_1")
   {
      return(varMetaData[index].LABEL_1);
   }
   else if (what == "LABEL_2")
   {
      return(varMetaData[index].LABEL_2);
   }
   else if (what == "LABEL_3")
   {  
      return(varMetaData[index].LABEL_3);
   }
   else if (what == "DISPLAYTYPE")
   {
      return(varMetaData[index].DISPLAYTYPE);
   }
   else if (what == "DEPEND_0")
   {
      return(varMetaData[index].DEPEND_0);
   }
   else if (what == "DEPEND_1")
   {
      return(varMetaData[index].DEPEND_1);
   }
   else if (what == "DEPEND_2")
   {
      return(varMetaData[index].DEPEND_2);
   }
   else if (what == "DEPEND_3")
   {
      return(varMetaData[index].DEPEND_3);
   }
   else if (what == "REPRESENTATION_1")
   {
      return(varMetaData[index].REPRESENTATION_1);
   }
   else if (what == "REPRESENTATION_2")
   {
      return(varMetaData[index].REPRESENTATION_2);
   }
   else if (what == "REPRESENTATION_3")
   {
      return(varMetaData[index].REPRESENTATION_3);
   }
   else if (what == "DELTA_PLUS")
   {
      return(varMetaData[index].DELTA_PLUS);
   }
   else if (what == "DELTA_MINUS")
   {
      return(varMetaData[index].DELTA_MINUS);
   }
   else if (what == "SIZES")
   {
      return(varMetaData[index].SIZES);
   }
   else if (what == "FIELDNAM")
   {
      return(varMetaData[index].FIELDNAM);
   } 
   else if (what == "VAR_TYPE")
   {
      return(varMetaData[index].VAR_TYPE);
   }
   else if (what == "ENTITY")
   {        
      return(varMetaData[index].ENTITY);
   }
   else return("");
}

//-----------------------------------------------------------------------------
//
// get lower or upper deltas
//    
//-----------------------------------------------------------------------------

int ReadCdf::getDataValueDelta(int index, int which, float *ptr)
{
   if (which == _PLUS)
   {
      QString tmp = getVariableAttr(index, "DELTA_PLUS");
      int size = getVariableSize(index, 0);
      if (!tmp.contains(",") && tmp.contains("_"))
      {
         int varNumber = getVariableNumber(tmp);
         if (varNumber != _READ_FAILED)
         {
            return(getDataValue(varNumber, ptr));
         }
         else return(_READ_FAILED);
      }
      else
      {
         QStringList tmpList = tmp.split(",");
         for (int i=0; i < size; i++)
         {
            if (i < tmpList.size()) // sometimes just a single number (0.0) is given rather than multiple 0.0's
            {
               *(ptr+i) = tmpList.at(i).toFloat();
            }
            else
            {
               *(ptr+i) = tmpList.at(0).toFloat();
            }
         }
      }
   }
   else if (which == _MINUS)
   {
      QString tmp = getVariableAttr(index, "DELTA_MINUS");
      int size = getVariableSize(index, 0);
      if (!tmp.contains(",") && tmp.contains("_"))
      {
         int varNumber = getVariableNumber(tmp);
         if (varNumber != _READ_FAILED)
         {
            return(getDataValue(varNumber, ptr));
         }  
         else return(_READ_FAILED);
      }
      else
      {
         QStringList tmpList = tmp.split(",");
         for (int i=0; i < size; i++)
         {
            if (i < tmpList.size()) // sometimes just a single number (0.0) is given rather than multiple 0.0's
            {
               *(ptr+i) = tmpList.at(i).toFloat();
            }
            else
            {
               *(ptr+i) = tmpList.at(0).toFloat();
            }
         }
      }
   }
   else return(_READ_FAILED);
   return(_READ_SUCCESS);
}

//-----------------------------------------------------------------------------
//
// check if a QString is the name of a variable
//
//-----------------------------------------------------------------------------

int ReadCdf::getVariableNumber(QString name)
{
   int retVal = _READ_FAILED;
   for (int i=0; i<numVariables; i++)
   {
      if (name == getVariableAttr(i, "NAME")) retVal = i;
   }
   return(retVal);
}

