/*******************************************************************************
 **
 ** $Id: readcef.cpp 165 2008-11-30 01:17:02Z AndrewLahiff $
 **
 ****************************************************************************** 
 ** 
 ** Copyright (C) 2007-2008 Andrew Lahiff.
 ** 
 ** This file is part of ReadFile.
 **
 ** ReadFile is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 ** 
 ** ReadFile is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 ** 
 ** You should have received a copy of the GNU General Public License
 ** along with ReadFile.  If not, see <http://www.gnu.org/licenses/>.
 ** 
 *******************************************************************************/

#include "readfile.h"
#include <iostream>
using namespace std;

//=============================================================================
//
// ReadCef class
//
//=============================================================================

QStringList safeSplit(QString str, QString sep)
{
   str.remove(QRegExp("\\s+"));
   QStringList var = str.split(sep, QString::SkipEmptyParts);
   return(var);
}

//-----------------------------------------------------------------------------
//
// constructor
//
//-----------------------------------------------------------------------------

ReadCef::ReadCef()
{
   numVariables = 0;
   inHeaderOrData = 0;
   dontUpdate = -1;
}

//-----------------------------------------------------------------------------
//
// destructor
//
//-----------------------------------------------------------------------------

ReadCef::~ReadCef()
{

}

//-----------------------------------------------------------------------------
//
// set filename and open file
//
//-----------------------------------------------------------------------------

int ReadCef::open(QString userFileName)
{
   file.setFileName(userFileName);
   if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) return _READ_FAILED;
   textStream.setDevice(&file);

   return _READ_SUCCESS;
}

//-----------------------------------------------------------------------------
//
// read CEF header
//
//-----------------------------------------------------------------------------

int ReadCef::readHeader()
{

// leave immediately if we have already read the header

   if (inHeaderOrData != 0) return(_READ_FAILED);
   
// variable to indicate if inside a variable's metadata

   bool inVarMetaData = false;

// read the header; stop when we get to the data or the end of the file

   while (!textStream.atEnd())
   {
      QString line = textStream.readLine();      

       // check for beginning of the data

      if (line.contains("DATA_UNTIL", Qt::CaseInsensitive) ||
          line.contains("START_DATA", Qt::CaseInsensitive))
      {
          inHeaderOrData = 1;
          determinePositionOffsets();
          return _READ_SUCCESS;
      } 

      // replace any tabs with spaces
      QString lineNew = line;
      lineNew.replace(QRegExp("\\t"), QString(" "));

      // split the line at spaces
      QStringList lineListS = lineNew.split(QRegExp("\\s+"), QString::SkipEmptyParts);

      if (lineListS.size())
      {

         // Check for variable meta data

         if (lineListS.at(0).startsWith("START_VARIABLE", Qt::CaseInsensitive))
         {
            // Handle memory allocation
            varMetaData.resize(numVariables + 1);

            // Set the variable name
            varMetaData[numVariables].NAME = lineListS.at(2);

            // Initially assume there is no data in the header for this variable; later
            // we will change this if necessary
            varDataInHeader.push_back(false);

            // Indicate we are inside variable meta data
	    inVarMetaData = true;
	 }
	 else if (lineListS.at(0).startsWith("DATA", Qt::CaseInsensitive))
	 {
            varDataInHeader.at(numVariables) = true;
            varDataHeader[varMetaData[numVariables].NAME] = extractFromList(lineListS, 2);
	 }
	 else if (lineListS.at(0).startsWith("PROPERTY", Qt::CaseInsensitive))
	 {
            varMetaData[numVariables].PROPERTY = extractFromList(lineListS, 2);
	 }	
	 else if (lineListS.at(0).startsWith("UNITS", Qt::CaseInsensitive))
	 {
            varMetaData[numVariables].UNITS = extractFromList(lineListS, 2);
	 }	  
         else if (lineListS.at(0).startsWith("DELTA_PLUS", Qt::CaseInsensitive))
         {  
            varMetaData[numVariables].DELTA_PLUS = extractFromList(lineListS, 2);
         }               
         else if (lineListS.at(0).startsWith("DELTA_MINUS", Qt::CaseInsensitive))
         {
            varMetaData[numVariables].DELTA_MINUS = extractFromList(lineListS, 2);
         }
	 else if (lineListS.at(0).startsWith("LABLAXIS", Qt::CaseInsensitive))
	 {
	    varMetaData[numVariables].LABLAXIS = extractFromList(lineListS, 2);
	 }	  	 
         else if (lineListS.at(0).startsWith("FIELDNAM", Qt::CaseInsensitive))
         {  
            varMetaData[numVariables].FIELDNAM = extractFromList(lineListS, 2);
         }
         else if (lineListS.at(0).startsWith("ENTITY", Qt::CaseInsensitive))
         {
            varMetaData[numVariables].ENTITY = extractFromList(lineListS, 2);
         }
         else if (lineListS.at(0).startsWith("SCALETYP", Qt::CaseInsensitive))
         {
            varMetaData[numVariables].SCALETYP = extractFromList(lineListS, 2);
         } 
         else if (lineListS.at(0).startsWith("SCALEMIN", Qt::CaseInsensitive))
         {
            varMetaData[numVariables].SCALEMIN = extractFromList(lineListS, 2);
         }
         else if (lineListS.at(0).startsWith("SCALEMAX", Qt::CaseInsensitive))
         {
            varMetaData[numVariables].SCALEMAX = extractFromList(lineListS, 2);
         }
         else if (lineListS.at(0).startsWith("VALUE_TYPE", Qt::CaseInsensitive) && inVarMetaData)
         {
            varMetaData[numVariables].VALUE_TYPE = extractFromList(lineListS, 2);

            DataType dataType;
            if (varMetaData[numVariables].VALUE_TYPE == "INT")
            {
               dataType = INT;
            }
            else if (varMetaData[numVariables].VALUE_TYPE == "FLOAT")
            {
               dataType = FLOAT;
            }
            else if (varMetaData[numVariables].VALUE_TYPE == "DOUBLE")
            {
               dataType = DOUBLE;
            }
            else if (varMetaData[numVariables].VALUE_TYPE == "CHAR")
            {
               dataType = CHAR;
            }
            else if (varMetaData[numVariables].VALUE_TYPE == "ISO_TIME")
            {
               dataType = ISO_TIME;
            }
            else dataType = UNKNOWN;
            varDataType.push_back(dataType);
         }  
         else if (lineListS.at(0).startsWith("VAR_TYPE", Qt::CaseInsensitive))
         {  
            varMetaData[numVariables].VAR_TYPE = extractFromList(lineListS, 2);
         } 
         else if (lineListS.at(0).startsWith("DISPLAYTYPE", Qt::CaseInsensitive))
         {
            varMetaData[numVariables].DISPLAYTYPE = extractFromList(lineListS, 2);
         }  
         else if (lineListS.at(0).startsWith("LABEL_1", Qt::CaseInsensitive))
         {  
            varMetaData[numVariables].LABEL_1 = extractFromList(lineListS, 2);
         }
         else if (lineListS.at(0).startsWith("LABEL_2", Qt::CaseInsensitive))
         {  
            varMetaData[numVariables].LABEL_2 = extractFromList(lineListS, 2);
         }  
         else if (lineListS.at(0).startsWith("LABEL_3", Qt::CaseInsensitive))
         {
            varMetaData[numVariables].LABEL_3 = extractFromList(lineListS, 2);
         }
         else if (lineListS.at(0).startsWith("REPRESENTATION_1", Qt::CaseInsensitive))
         {  
            varMetaData[numVariables].REPRESENTATION_1 = extractFromList(lineListS, 2);
         }
         else if (lineListS.at(0).startsWith("REPRESENTATION_2", Qt::CaseInsensitive))
         {     
            varMetaData[numVariables].REPRESENTATION_2 = extractFromList(lineListS, 2);
         }  
         else if (lineListS.at(0).startsWith("REPRESENTATION_3", Qt::CaseInsensitive))
         {     
            varMetaData[numVariables].REPRESENTATION_3 = extractFromList(lineListS, 2);
         }  
         else if (lineListS.at(0).startsWith("FILLVAL", Qt::CaseInsensitive))
         {
            varMetaData[numVariables].FILLVAL = extractFromList(lineListS, 2);
         }
         else if (lineListS.at(0).startsWith("DEPEND_0", Qt::CaseInsensitive))
         {
            varMetaData[numVariables].DEPEND_0 = extractFromList(lineListS, 2);
         }
         else if (lineListS.at(0).startsWith("DEPEND_1", Qt::CaseInsensitive))
         {
            varMetaData[numVariables].DEPEND_1 = extractFromList(lineListS, 2);
         }
         else if (lineListS.at(0).startsWith("DEPEND_2", Qt::CaseInsensitive))
         {
            varMetaData[numVariables].DEPEND_2 = extractFromList(lineListS, 2);
         }
         else if (lineListS.at(0).startsWith("DEPEND_3", Qt::CaseInsensitive))
         {
            varMetaData[numVariables].DEPEND_3 = extractFromList(lineListS, 2);
         }
	 else if (lineListS.at(0).startsWith("END_VARIABLE", Qt::CaseInsensitive))
	 {
	    inVarMetaData = false;

            // If SIZES is not defined, there must only be 1 dimension

	    if (varMetaData[numVariables].SIZES.isEmpty())
	    {
	       std::vector<int> tmpVec;
               tmpVec.push_back(1);
               varDataSizes.push_back(tmpVec);
	    }

            // Increment the number of variables
            numVariables++;
	 }
	 else if (lineListS.at(0).startsWith("SIZES", Qt::CaseInsensitive))
	 {
            varMetaData[numVariables].SIZES = extractFromList(lineListS, 2);
  
            std::vector<int> tmpVec;
            if (!varMetaData[numVariables].SIZES.contains(","))
            {
               tmpVec.push_back(varMetaData[numVariables].SIZES.toInt());
            }
            else
            {
               QStringList tmpSplit = varMetaData[numVariables].SIZES.split(",");
               tmpVec.resize(tmpSplit.size());
               for (int i=0; i<tmpSplit.size(); ++i)
               {
                  tmpVec.at(i) = tmpSplit.at(i).toInt();
               }
            }
            varDataSizes.push_back(tmpVec);
	 }
	 else if (lineListS.at(0).startsWith("PARAMETER_TYPE", Qt::CaseInsensitive))
	 {
	    varMetaData[numVariables].PARAMETER_TYPE = extractFromList(lineListS, 2);
	 }
	 else if (lineListS.at(0) == "FILLVAL")
	 {
            varMetaData[numVariables].FILLVAL = extractFromList(lineListS, 2);
	 }
	 else if (lineListS.at(0).startsWith("END_OF_RECORD_MARKER", Qt::CaseInsensitive))
	 {
            endOfRecordMarker = extractFromList(lineListS, 2);
	 }

         // Check for global meta data

         if (QString::compare(lineListS.at(0), "START_META", Qt::CaseInsensitive) == 0 && lineListS.size() > 2)
         {
            if (QString::compare(lineListS.at(2), "DATASET_ID", Qt::CaseInsensitive) == 0)
            {
               varGlobalData.DATASET_ID = extractGlobalAttr(textStream);
            }
            else if (QString::compare(lineListS.at(2), "DATASET_TITLE", Qt::CaseInsensitive) == 0)
            {
               varGlobalData.DATASET_TITLE = extractGlobalAttr(textStream);
            }
            else if (QString::compare(lineListS.at(2), "LOGICAL_SOURCE", Qt::CaseInsensitive) == 0)
            {
               varGlobalData.LOGICAL_SOURCE = extractGlobalAttr(textStream);
            }
            else if (QString::compare(lineListS.at(2), "EXPERIMENT", Qt::CaseInsensitive) == 0)
            {  
               varGlobalData.EXPERIMENT = extractGlobalAttr(textStream);
            }
            else if (QString::compare(lineListS.at(2), "OBSERVATORY", Qt::CaseInsensitive) == 0)
            {  
               varGlobalData.OBSERVATORY = extractGlobalAttr(textStream);
            }
            else if (QString::compare(lineListS.at(2), "MISSION", Qt::CaseInsensitive) == 0)
            {  
               varGlobalData.MISSION = extractGlobalAttr(textStream);
            }
            else if (QString::compare(lineListS.at(2), "INSTRUMENT_NAME", Qt::CaseInsensitive) == 0)
            {  
               varGlobalData.INSTRUMENT_NAME = extractGlobalAttr(textStream);
            }
            else if (QString::compare(lineListS.at(2), "DATA_TYPE", Qt::CaseInsensitive) == 0)
            {  
               varGlobalData.DATA_TYPE = extractGlobalAttr(textStream);
            } 
            else if (QString::compare(lineListS.at(2), "FILE_TIME_SPAN", Qt::CaseInsensitive) == 0)
            {  
               QString tmp = extractGlobalAttr(textStream);
               QStringList tmpList = tmp.split("/");
               if (tmpList.size() == 2)
               {
                  timeFileBegin = tmpList.at(0);
                  timeFileEnd = tmpList.at(1);
               }
               else
               {
                  timeFileBegin = tmp;
               }
            }
         }
      }
   }
   inHeaderOrData = 1;
   determinePositionOffsets();
   return(_READ_SUCCESS);
}

//-----------------------------------------------------------------------------
//
// determine position offsets for each variable
//
//-----------------------------------------------------------------------------

void ReadCef::determinePositionOffsets(void)
{
   int iTmpPosition = 0;
   for (int i = 0; i < numVariables; i++)
   {
      varDataOffsets.push_back(iTmpPosition);
      int iCurrent = 1;
      for (int j = 0; j < getVariableDimension(i); j++)
      {
         iCurrent = iCurrent*getVariableSize(i, j);
      }
      if (!varDataInHeader.at(i)) iTmpPosition += iCurrent;
   }
   return;
}

//-----------------------------------------------------------------------------
//
// return number of variables
//
//-----------------------------------------------------------------------------

int ReadCef::getNumVariables()
{
   if (inHeaderOrData == 1)
   {
      return(numVariables);
   }
   else
   {
      return(-1);
   }
}

//-----------------------------------------------------------------------------
//
// return number of dimensions of variable
//
//-----------------------------------------------------------------------------

int ReadCef::getVariableDimension(int index)
{
   if (index >= 0 && index < numVariables)
   {  
      return varDataSizes.at(index).size();
   }
   else return _READ_FAILED;   
}

//-----------------------------------------------------------------------------
//
// return size of a dimension of a variable
//
//-----------------------------------------------------------------------------

int ReadCef::getVariableSize(int index, int inIndex)
{
   if (index >= 0 && index < numVariables)
   {
      if (inIndex >= 0 && inIndex < static_cast<int>(varDataSizes.at(index).size()))
      {
         return varDataSizes.at(index).at(inIndex);
      }
      else return _READ_FAILED;
   }
   else return _READ_FAILED;
}

//-----------------------------------------------------------------------------
//
// return variable type (e.g. integer, float, ...)
//
//-----------------------------------------------------------------------------

DataType ReadCef::getVariableDataType(int index)
{
   if (index >= 0 && index < numVariables)
   {
      return varDataType.at(index);
   }
   else return UNKNOWN;
}

//-----------------------------------------------------------------------------
//
// get begin date/time
//
//-----------------------------------------------------------------------------

QString ReadCef::getFileBeginDateTime()
{
   if (inHeaderOrData != 0)
   {
      return timeFileBegin;
   }
   else
   {
      return QString("");
   }
}

//-----------------------------------------------------------------------------
//
// get end date/time
//
//-----------------------------------------------------------------------------

QString ReadCef::getFileEndDateTime()
{
   if (inHeaderOrData != 0)
   {
      return timeFileEnd;
   }
   else
   {
      return QString("");
   }
}

//-----------------------------------------------------------------------------
//
// read a data record
//
//-----------------------------------------------------------------------------

int ReadCef::readRecord(bool updateNext)
{

// handle "don't update"
// - sometimes we need info from the 1st record before the main data reading
//   begins. This ensures we don't miss 1 record.

   if (!updateNext) dontUpdate = 0;
   else
   {
      if (dontUpdate == 0)
      {
         dontUpdate = -1;
         return _READ_SUCCESS;
      }
   }

// if we're in the header, return with an error

   if (inHeaderOrData == 0) return _READ_FAILED;

// clear the listing containing data for the current record

   currentData.clear();

// read in data for the current record

   bool atEndOfRecord = false;
   if (!textStream.atEnd())
   {
      while (!atEndOfRecord && !textStream.atEnd())
      {
         // read a single line

         QString line = textStream.readLine();

         // check if the current line is entirely a comment line; if it is, continue
         // on to the next line

         if (line.startsWith("!")) continue;

         // check if the current line is completely empty; if it is, continue 
         // on to the next line

         if (line.isEmpty()) continue;
         

         // add the current line's data to the data record storage QStringList,
         // ensuring that we ignore comments & remove the end of record marker if
         // necessary

         QStringList lineSplit = line.split(",", QString::SkipEmptyParts);
         for (int i=0; i<lineSplit.size(); ++i)
         {
            QString dataPiece = lineSplit.at(i);
            if (dataPiece == "!" || dataPiece == endOfRecordMarker) break;
            if (dataPiece.contains(endOfRecordMarker)) dataPiece.remove(endOfRecordMarker);
            dataPiece = dataPiece.trimmed(); // remove white space from beginning & end
            if (dataPiece.size() > 0) currentData << dataPiece;
         }
       
         // check if the current line contains the end of record marker

         if (line.contains(endOfRecordMarker)) atEndOfRecord = true;
      }
   }
   else return _READ_FAILED;

   // if we didn't find the end of record marker, something is wrong so return an error

   if (atEndOfRecord) return _READ_SUCCESS;
   else return _READ_FAILED;
}

//-----------------------------------------------------------------------------
//
// get a data value from the current record (short int version)
//
//-----------------------------------------------------------------------------

int ReadCef::getDataValue(int index, short int *ptr)
{
   if (index >= 0 && index < numVariables && currentData.size() > 0)
   {
      if (!varDataInHeader.at(index))
      {
         int k = 0;
         int l = varDataOffsets.at(index);

         int getSize = 1;
         for (int i = 0; i < getVariableDimension(index); i++)
         {
            getSize = getSize*getVariableSize(index, i);
         }
         for (int j = 0; j < getSize; j++)
         {
            *(ptr + k) = (currentData.at(l + k)).toInt();
            k++;
         }
      }
      else
      {
         QString tmp = varDataHeader[varMetaData[index].NAME];
         if (tmp.contains(","))
	 {
	    QStringList dataSplit = tmp.split(",", QString::SkipEmptyParts);
	    for (int i = 0; i < dataSplit.size(); i++)
	    {
	       *(ptr + i) = (dataSplit.at(i)).toInt();
	    }
	 }
	 else
	 {	 
            *ptr = tmp.toInt();
	 }
      }
      return(_READ_SUCCESS);
   }
   else
   {
      return(_READ_FAILED);
   }
}


//-----------------------------------------------------------------------------
//
// get a data value from the current record (integer version)
//
//-----------------------------------------------------------------------------

int ReadCef::getDataValue(int index, int *ptr)
{
   if (index >= 0 && index < numVariables && currentData.size() > 0)
   {
      if (!varDataInHeader.at(index))
      {
         int k = 0;
         int l = varDataOffsets.at(index);

         int getSize = 1;
         for (int i = 0; i < getVariableDimension(index); i++)
         {
            getSize = getSize*getVariableSize(index, i);
         }
         for (int j = 0; j < getSize; j++)
         {
            *(ptr + k) = (currentData.at(l + k)).toInt();
            k++;
         }
      }
      else
      {
         QString tmp = varDataHeader[varMetaData[index].NAME];
         if (tmp.contains(","))
         {
            QStringList dataSplit = tmp.split(",", QString::SkipEmptyParts);
            for (int i = 0; i < dataSplit.size(); i++)
            {
               *(ptr + i) = (dataSplit.at(i)).toInt();
            }
         }
         else
         {
            *ptr = tmp.toInt();
         }
      }
      return(_READ_SUCCESS);
   }
   else
   {
      return(_READ_FAILED);
   }
}

//-----------------------------------------------------------------------------
//    
// get a data value from the current record (float version)
//    
//-----------------------------------------------------------------------------

int ReadCef::getDataValue(int index, float *ptr)
{     
   if (index >= 0 && index < numVariables && currentData.size() > 0)
   {
      if (!varDataInHeader.at(index))
      {
         int k = 0;
         int l = varDataOffsets.at(index);

         int getSize = 1;
         for (int i = 0; i < getVariableDimension(index); i++)
         {
            getSize = getSize*getVariableSize(index, i);
         }
         for (int j = 0; j < getSize; j++)
         {
            *(ptr + k) = (currentData.at(l + k)).toFloat();
            k++;
         }
      }
      else
      {
         QString tmp = varDataHeader[varMetaData[index].NAME];
         if (tmp.contains(","))
         {
            QStringList dataSplit = tmp.split(",", QString::SkipEmptyParts);
            for (int i = 0; i < dataSplit.size(); i++)
            {
               *(ptr + i) = (dataSplit.at(i)).toFloat();
            }
         }
         else
         {
            *ptr = tmp.toFloat();
         }
      }
      return(_READ_SUCCESS);
   }
   else
   {
      return(_READ_FAILED);
   }
}

//-----------------------------------------------------------------------------
//    
// get a data value from the current record (double version)
//    
//-----------------------------------------------------------------------------

int ReadCef::getDataValue(int index, double *ptr)
{     
   if (index >= 0 && index < numVariables && currentData.size() > 0)
   {
      if (!varDataInHeader.at(index))
      {
         int k = 0;
         int l = varDataOffsets.at(index);

         int getSize = 1;
         for (int i = 0; i < getVariableDimension(index); i++)
         {
            getSize = getSize*getVariableSize(index, i);
         }
         for (int j = 0; j < getSize; j++)
         {
            *(ptr + k) = (currentData.at(l + k)).toDouble();
	    k++;
         }
      }
      else
      {
         QString tmp = varDataHeader[varMetaData[index].NAME];
         if (tmp.contains(","))
         {
            QStringList dataSplit = tmp.split(",", QString::SkipEmptyParts);
            for (int i = 0; i < dataSplit.size(); i++)
            {
               *(ptr + i) = (dataSplit.at(i)).toDouble();
            }
         }
         else
         {
            *ptr = tmp.toDouble();
         }
      }
      return(_READ_SUCCESS);
   }
   else
   {
      return(_READ_FAILED);
   }
}

//-----------------------------------------------------------------------------
//    
// get a data value from the current record (string version)
//    
//-----------------------------------------------------------------------------

int ReadCef::getDataValue(int index, QString *ptr)
{
   if (index >= 0 && index < numVariables)
   {
      int l = varDataOffsets.at(index);
      if (l < currentData.size())
      {
         *ptr = currentData.at(l);
         return(_READ_SUCCESS);
      }
      else return(_READ_FAILED);
   }
   else
   {
      return(_READ_FAILED);
   }
}

//-----------------------------------------------------------------------------
//
// close CEF file
//
//-----------------------------------------------------------------------------

void ReadCef::close()
{
   file.close();
}

//-----------------------------------------------------------------------------
//
// get a global attribute
//
//-----------------------------------------------------------------------------

QString ReadCef::getGlobalAttr(QString what)
{
   if (what == "OBSERVATORY")
   {
      return varGlobalData.OBSERVATORY;
   }
   else if (what == "EXPERIMENT")
   {
      return varGlobalData.EXPERIMENT;
   }
   else if (what == "DATASET_ID")
   {
      return varGlobalData.DATASET_ID;
   }
   else if (what == "DATASET_TITLE")
   {
      return varGlobalData.DATASET_TITLE;
   }
   else if (what == "LOGICAL_SOURCE")
   {
      return varGlobalData.LOGICAL_SOURCE;
   }
   else if (what == "MISSION")
   {
      return varGlobalData.MISSION;
   }
   else if (what == "INSTRUMENT_NAME")
   {
      return varGlobalData.INSTRUMENT_NAME;
   }
   else if (what == "DATA_TYPE")
   {
      return varGlobalData.DATA_TYPE;
   }
   else return(QString(""));
}

//-----------------------------------------------------------------------------
//
// get a variable attribute
//
//-----------------------------------------------------------------------------

QString ReadCef::getVariableAttr(int index, QString what)
{
   if (what == "UNITS")
   {
      return(varMetaData[index].UNITS);
   }
   else if (what == "PARAMETER_TYPE")
   {
      return(varMetaData[index].PARAMETER_TYPE);
   }
   else if (what == "LABLAXIS")
   {
      return(varMetaData[index].LABLAXIS);	
   }
   else if (what == "NAME")
   {
      return(varMetaData[index].NAME);
   }
   else if (what == "PROPERTY")
   {
      return(varMetaData[index].PROPERTY);
   }
   else if (what == "FILLVAL")
   {
      return(varMetaData[index].FILLVAL);
   }
   else if (what == "SCALETYP")
   {
      return(varMetaData[index].SCALETYP);
   }
   else if (what == "SCALEMIN")
   {
      return(varMetaData[index].SCALEMIN);
   }
   else if (what == "SCALEMAX")
   {
      return(varMetaData[index].SCALEMAX);
   }   
   else if (what == "VALUE_TYPE")
   {
      return(varMetaData[index].VALUE_TYPE);
   } 
   else if (what == "LABEL_1")
   {
      return(varMetaData[index].LABEL_1);
   }
   else if (what == "LABEL_2")
   {
      return(varMetaData[index].LABEL_2);
   }
   else if (what == "LABEL_3")
   {  
      return(varMetaData[index].LABEL_3);
   }
   else if (what == "DISPLAYTYPE")
   {
      return(varMetaData[index].DISPLAYTYPE);
   }
   else if (what == "DEPEND_0")
   {
      return(varMetaData[index].DEPEND_0);
   }
   else if (what == "DEPEND_1")
   {
      return(varMetaData[index].DEPEND_1);
   }
   else if (what == "DEPEND_2")
   {
      return(varMetaData[index].DEPEND_2);
   }
   else if (what == "DEPEND_3")
   {
      return(varMetaData[index].DEPEND_3);
   }
   else if (what == "REPRESENTATION_1")
   {
      return(varMetaData[index].REPRESENTATION_1);
   }
   else if (what == "REPRESENTATION_2")
   {  
      return(varMetaData[index].REPRESENTATION_2);
   }
   else if (what == "REPRESENTATION_3")
   {  
      return(varMetaData[index].REPRESENTATION_3);
   }
   else if (what == "DELTA_PLUS")
   {
      return(varMetaData[index].DELTA_PLUS);
   }
   else if (what == "DELTA_MINUS")
   {
      return(varMetaData[index].DELTA_MINUS);
   }
   else if (what == "FIELDNAM")
   {
      return(varMetaData[index].FIELDNAM);
   }
   else if (what == "VAR_TYPE")
   {
      return(varMetaData[index].VAR_TYPE);
   }
   else if (what == "ENTITY")
   {
      return(varMetaData[index].ENTITY);
   }
   else return(QString(""));
}

//-----------------------------------------------------------------------------
//
// get lower or upper deltas
//    
//-----------------------------------------------------------------------------

int ReadCef::getDataValueDelta(int index, int which, float *ptr)
{
   if (which == _PLUS)
   {
      QString tmp = getVariableAttr(index, "DELTA_PLUS");
      int size = getVariableSize(index, 0);
      if (!tmp.contains(",") && tmp.contains("_"))
      {
         int varNumber = getVariableNumber(tmp);
         if (varNumber != _READ_FAILED)
         {
            return(getDataValue(varNumber, ptr));
         }  
         else return(_READ_FAILED);
      }  
      else
      {
         QStringList tmpList = tmp.split(",");
         for (int i=0; i < size; i++)
         {
            if (i < tmpList.size()) // sometimes just a single number (0.0) is given rather than multiple 0.0's
            {
               *(ptr+i) = tmpList.at(i).toFloat();
            }
            else
            {
               *(ptr+i) = tmpList.at(0).toFloat();
            }
         }
      }  
   }  
   else if (which == _MINUS)
   {
      QString tmp = getVariableAttr(index, "DELTA_MINUS");
      int size = getVariableSize(index, 0);
      if (!tmp.contains(",") && tmp.contains("_"))
      {
         int varNumber = getVariableNumber(tmp);
         if (varNumber != _READ_FAILED)
         {
            return(getDataValue(varNumber, ptr));
         }
         else return(_READ_FAILED);
      }
      else
      {
         QStringList tmpList = tmp.split(",");
         for (int i=0; i < size; i++)
         {
            if (i < tmpList.size()) // sometimes just a single number (0.0) is given rather than multiple 0.0's
            {
               *(ptr+i) = tmpList.at(i).toFloat();
            }
            else
            {
               *(ptr+i) = tmpList.at(0).toFloat();
            }
         }
      }
   }
   else return(_READ_FAILED);
   return(_READ_SUCCESS);
}

//-----------------------------------------------------------------------------
//
// check if a QString is the name of a variable
//
//-----------------------------------------------------------------------------

int ReadCef::getVariableNumber(QString name)
{
   int retVal = _READ_FAILED;
   for (int i=0; i<numVariables; i++)
   {
      if (name == getVariableAttr(i, "NAME")) retVal = i;
   }
   return(retVal);
}

//-----------------------------------------------------------------------------
//
// extract multiple parts from a QStringList & remove quotes if necessary
//
//-----------------------------------------------------------------------------

QString ReadCef::extractFromList(QStringList list, int begin, int end)
{
   int myEnd = end;
   if (end == -1) myEnd = list.size();

   QString tmp;
   for (int i=begin; i<myEnd; ++i)
   {
      if (i > begin) tmp = tmp + " ";
      tmp = tmp + list.at(i);
   }
   tmp.remove("\"");
   return tmp;
}

//-----------------------------------------------------------------------------
//
// extract a global attribute
//
//-----------------------------------------------------------------------------

QString ReadCef::extractGlobalAttr(QTextStream& textStream)
{
   bool completed = false;
   while (!completed && !textStream.atEnd())
   {
      QString line = textStream.readLine();
      line.replace(QRegExp("\\t"), QString(" "));
      QStringList lineListS = line.split(QRegExp("\\s+"), QString::SkipEmptyParts);
      if (lineListS.size() > 0)
      {
         if (QString::compare(lineListS.at(0), "ENTRY", Qt::CaseInsensitive) == 0)
         {
            QString tmp = extractFromList(lineListS, 2);
            return tmp;
         }
         else if (QString::compare(lineListS.at(0), "END_META", Qt::CaseInsensitive) == 0)
         {
            completed = true;
         }
      }
   }
   QString tmp;
   tmp.clear();
   return tmp;
}
